
/********************************************************************************
 *										*
 *				 M A C E K					*
 *		("Matroids Also Computed Efficiently" Kit)			*
 *										*
 * A set of tools and routines for computations with representable matroids.	*
 * Copyright (C) 2001,2002  Petr Hlineny.					*
 *										*
 *   This program is free software; you can redistribute it and/or modify	*
 *   it under the terms of the GNU General Public License as published by	*
 *   the Free Software Foundation; either version 2 of the License, or		*
 *   (at your option) any later version.					*
 *   You should have received a copy of the GNU General Public License		*
 *   along with this program; if not, write to the Free Software		*
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.	*
 * 										*
 * See the top-level "ABOUT" and "doc/ *" files for more information.		*
 *										*
 ********************************************************************************/















/**
 * A matrix is a structure holding an array (considered 2-dimensional) of pfield elements,
 * together with its dimensions and maximal dimensions.
 * The basic manipulating functions for matrices (creating, freeing, accessing, refering,
 * printing, etc) are defined in ematrix.c.
 * 
 * This file defines more-involved matrix functions, like computations (pivoting, determinants)
 * and various simple matroid-related tests (rank, dependency, etc).
 * (Complicated structural functions are defined in struct/ * ...)
 * 
 * Look for more information in ../include/ematrix.h.
 * 
**/




#include "macek.h"
#include "emx.h"








/******************	Matrix computations	*******************/
/******************************************************************/


#undef CURDLEV
#define CURDLEV		7

static int	inrecur = 0;	/* (used for some recursive debug tests) */



/**
 * Pivoting operation on the given element ro,so of the matrix e.
 * Remember that the matrix is supposed to be in a standard form, i.e. without
 * the identity part, and it remains in the form afterwards.
 * If the computation gets out of in pfield numbers, an error is issued.
 * A variant that does not print error messages if ch==0 is provided as well.
 * Return value 0 means success, and -1 failure of computation.
 * (The modified matrix is stored in the same e.)
**/

void	ematrix_pivot(ematrix *e, int ro, int co) {
	if (ematrix_pivot_check(1,e,ro,co)<0) {
		EMATDEBUG(0,e,"  !!\t");
		PROGERROR("pivoting %d,%d in %p failed!",ro,co,e);
	}
}

int	ematrix_pivot_check(int ch, ematrix *e, int ro, int co) {
	int	i,j, r;
	exp_t	xx;
	sign_t	gg;
	
	if (SIGNM(e,ro,co)==0) {
		if (ch>=0) { PROGERROR("cannot pivot on a 0 element at %d,%d",ro,co);  EMATDEBUG(0,e," !\t"); }
		return -1;
	}
	DEBUG(CURDLEV+2,"pivoting on %d,%d in a matrix %p\n",ro,co,e);
	for (i=0; i<ROWSM(e); i++)
		if (i!=ro) {
			r = pfield_mul_ch(ch,1,EXPM(e,i,co),SIGNM(e,i,co),
				-1,EXPM(e,ro,co),-SIGNM(e,ro,co),&(EXPM(e,i,co)),&(SIGNM(e,i,co)));
			if (r<0)  return -1;
		}
	for (i=0; i<ROWSM(e); i++)  for (j=0; j<COLSM(e); j++)
		if (i!=ro && j!=co) {
			r = pfield_mul_ch(ch,1,EXPM(e,i,co),SIGNM(e,i,co),
					1,EXPM(e,ro,j),SIGNM(e,ro,j),&xx,&gg);
			if (r>=0)  r = pfield_sum_ch(ch,1,EXPM(e,i,j),SIGNM(e,i,j),
					1,xx,gg,&(EXPM(e,i,j)),&(SIGNM(e,i,j)));
			if (r<0)  return -1;
		}
	pfield_setzeroexp(&(EXPM(e,ro,co)));  SIGNM(e,ro,co) = 1;
			/* the pivoting exchanges the row and column ro,co id's */
	i = COLSID(e,co);  COLSID(e,co) = ROWSID(e,ro);  ROWSID(e,ro) = i;
	return 0;
}
  


/**
 * Recursive computation of a determinant of the given matrix.
 * The computation creates a local copy of the matrix not to change the original.
 * The result is stored into given pointers x,g;  and 0, or -1 for an error, is returned.
 * (Errors include non-square matrix, out of in pfield numbers, or invalid entries.)
 * A variant is provided that does not print error messages for ch=0.
**/

int	ematrix_determinant_recur(int ch, ematrix *e, exp_t *x, sign_t *g) ;

int	ematrix_determinant_ch(int ch, ematrix *e, exp_t *x, sign_t *g) {
	int	ret;
	ematrix	*e1;
	exp_t	xx, xxt;
	sign_t	gg, ggt;
	
	pfield_setzeroexp(&xx);  gg = 0;
	if (ROWSM(e)!=COLSM(e)) {
		if (ch>0)  {PROGERROR("not computing determinant of a non-square matrix  %d x %d",ROWSM(e),COLSM(e));}
		ret = -1;
	} else {	/* here a local copy of the matrix e is created - faster for the right max dimensions */
		e1 = ematrix_copy_to(e,ROWSM(e),COLSM(e));
			/* the real computation happens next: */
		ret = ematrix_determinant_recur(ch,e1,&xx,&gg);
#ifndef FASTPROG
		if (ret>=0 && IFRANDDEBUGLESS(222) && ch>0) {	/* testing correct computation (only if supposed in pfield) */
			if (ROWSM(e1)==2)  ematrix_determinant_2x2(e1,0,1,0,1,&xxt,&ggt);
			else  if (ROWSM(e1)>2) {
				ematrix_swaprows(e1,1,ROWSM(e1)-1);  ematrix_swapcols(e1,0,COLSM(e1)-2);
				ematrix_transpose(e1);  ematrix_determinant_ch(-1,e1,&xxt,&ggt);
			}		/* (call with ch==-1 to prevent inf recursion) */
			if (ROWSM(e1)>=2)  if (!pfield_isequal(xx,gg,xxt,ggt)) {
				PROGERROR("incorrect determinant computation!");  EMATDEBUG(0,e1,"!\t");  DEBUG(0," - det = %s ,  detch = %s\n",pfield_printvalue(20,xx,gg),pfield_printvalue(20,xxt,ggt)); }
		}	/* (e1 is modified here !!!) */
#endif
		xxt = xx;  ggt = gg;	/* (just to prevent "unused" warning) */
		dispose_ematrix(e1);
	}
	if (x)  *x = xx;  if (g)  *g = gg;
	if (ch>0 && ret<0)  {PROGERROR("computing determinant of %p [%dx%d] failed!",e,ROWSM(e),COLSM(e));}
	return ret;
}

int	ematrix_determinant_recur(int ch, ematrix *e, exp_t *x, sign_t *g) {
	int	i,j,  ret = 0;
	exp_t	xx, xm,xr;
	sign_t	gg, gm,gr;
	
	pfield_setzeroexp(&xx);  gg = 0;
	i = ROWSM(e)-1;  j = 0;
	if (i<=0)  {
		xx = EXPM(e,0,0);  gg = SIGNM(e,0,0);
	} else {
		for (j=COLSM(e)-1; j>=0; j--)
			if (SIGNM(e,i,j)!=0)  break;
	}
		/**
		 * How we compute the determinant by standard Gauss elimination:
		 * We find a nonzero entry i,j in the last row (if it exists) above,
		 * and then we pivot on that entry i,j.
		 * In this way the (original) column j becomes a unit line, not
		 * affecting the determinant value, and so the lines i,j are removed.
		 * The smaller determinant is then computed by recursion.
		 * Finally, we must multiply the recurrent value by the original
		 * entry at i,j since pivoting divided the determinant by this entry,
		 * and we must also add a proper sign by (i+j)%2.
		**/
	if (i>0 && j>=0) {
		xx = xm = EXPM(e,i,j);
		gg = gm = SIGNM(e,i,j) * ((i+j)%2?-1:1);
		ret = ematrix_pivot_check(ch,e,i,j);
		if (ret>=0) {
			ematrix_remove(e,i,j);
			ret = ematrix_determinant_recur(ch,e,&xr,&gr);
			if (ret>=0)
				ret = pfield_mul_ch(ch,1,xm,gm,1,xr,gr,&xx,&gg);
		}
	}
	*x = xx;  *g = gg;
	return ret;
}


/**
 * This is a simple computation of a 2 x 2 subdeterminant on r1,r2 x c1,c2 lines of e.
 * It does not check much for correctness, so it should be used carefully.
**/

int	ematrix_determinant_2x2ext(int ch, ematrix *e, int r1, int r2, int c1, int c2,
								exp_t *x, sign_t *g) {
	exp_t	x1,x2,xx;
	sign_t	g1,g2,gg;
	int	ret = 0;
	
	ret = pfield_mul_ch(ch,1,EXPM(e,r1,c1),SIGNM(e,r1,c1),
					1,EXPM(e,r2,c2),SIGNM(e,r2,c2),&x1,&g1);
	if (ret>=0)  ret = pfield_mul_ch(ch,1,EXPM(e,r1,c2),SIGNM(e,r1,c2),
					1,EXPM(e,r2,c1),SIGNM(e,r2,c1),&x2,&g2);
	if (ret>=0)  ret = pfield_sum_ch(ch,1,x1,g1,1,x2,-g2,&xx,&gg);
	
	if (x)  *x = xx;  if (g)  *g = gg;
	if (ret<0 && ch>0)  {PROGERROR("computing 2x2 determinant %d,%d x %d,%d of %p failed! %s %s; %s %s",r1,r2,c1,c2,e,
				pfield_pvalue(8,EXPM(e,r1,c1),SIGNM(e,r1,c1)),pfield_pvalue(8,EXPM(e,r1,c2),SIGNM(e,r1,c2)),pfield_pvalue(8,EXPM(e,r2,c1),SIGNM(e,r2,c1)),pfield_pvalue(8,EXPM(e,r2,c2),SIGNM(e,r2,c2)));
				EMATDEBUG(0,e," !D!\t");}
	return ret;
}



/**
 * This function multiplies a submatrix on rows r1,...,r2-1 and columns c1,...,c2-1 of
 * the given matrix e by x,g.
 * (The modified matrix is stored in the same e.)
**/

void	ematrix_multiply_rc(ematrix *e, int r1, int r2, int c1, int c2, int sx, exp_t x, sign_t g) {
	int	i,j, ret=0;
	
	for (i=r1; i<r2; i++)  for (j=c1; j<c2; j++) {
		ret = pfield_mul(1,EXPM(e,i,j),SIGNM(e,i,j),sx,x,g,
				&EXPM(e,i,j),&SIGNM(e,i,j)) <0?-1:ret;
	}
	if (ret<0)  {PROGERROR("multiplying %d_%d x %d_%d submatrix of %p failed!",r1,r2,c1,c2,e);}
}



/**
 * This function computes a new matrix obtained from em by deleting (dc==1) or contracting
 * (dc==0) the element of the given id (not line-number!).
 * If necessary, the element for removing is first pivoted in the matrix.
 * The new matrix is returned as a copy.
**/

ematrix*	ematrix_removematid_ext(ematrix *em, int dc, int id) {
	ematrix		*ee;
	int		i,j,k;
	
	ee = ematrix_copy(em);
	if (!dc)  ematrix_transpose(ee);
	
	for (i=k=0; i<ROWSM(ee); i++)  if (ROWSID(ee,i)==id) {
		if (k)  {PROGERROR("Duplicated id's in lines %d, %d!",k,i+1);}
		k = i+1;
		for (j=0; j<COLSM(ee) && SIGNM(ee,i,j)==0; j++) ;
		if (j>=COLSM(ee))  {PROGERROR("Cannot remove zero line!");}
		else  ematrix_pivot(ee,i,j);
	}
	for (j=k=0; j<COLSM(ee); j++)  if (COLSID(ee,j)==id) {
		if (k)  {PROGERROR("Duplicated id's in lines %d, %d!",k,j+1);}
		k = j+1;
		ematrix_remove_col(ee,j);
	}
	if (!dc)  ematrix_transpose(ee);
	return ee;
}













/******************	Matrix pfield testing	*******************/
/******************************************************************/


#undef CURDLEV
#define CURDLEV		7



/**
 * The next function tests whether the given matrix is in the current pfield, i.e.
 * whether all of its subdeterminants (of any size) are defined in the pfield numbers.
 * The computation is recursive by brute force - quite slow... (But how to do better?)
 * If klr,klc>0, then only subdeterminants containing the last row or column are considered.
 * 
 * If ch>0, then an error message is printed for a non-pfield matrix,
 * and the minimal bad matrix is printed to the output (not debug).
 * More specifically, ALL subdeterminants are computed for ch==2, and subdeterminants
 * are computed until the smallest one (size 2?) is found for ch==1.
 * So if ch<2, the test (if) fails rather quickly, but positive answer is always slow.
 * For fast (and almost accurate) positive test, there is a randomized variant
 * of the test for ch==-7.
 * 
 * It returns 0 for a pfield matrix, and -1 otherwise.
 * If one works in a pfield with total sum, then nothing is computed and 0 is returned.
**/

#define	MXINPF	28
static ematrix	*estmin = NULL;		/* used for storing minimal bad determinant... */
static int	sizmin;
int     ematrix_inpfield_core(int ch, ematrix *em, int klr, int klc) ;

int	ematrix_inpfield_ext(int ch, ematrix *em, int klr, int klc) {
	int	ret = 0, i,ii,j,jj;
	ematrix	*e;
	
	if (!pfispartial) {
		DEBUG(CURDLEV+1,"NOT checking for pfield since the sum is total, not partial\n");
		return 0;
	}
	e = ematrix_copy_to(em,ROWSM(em),COLSM(em));
	estmin = NULL;  sizmin = 9999;
	klr = klr>0 ?1:0;  klc = klc>0? 1:0;
	
	DEBUG(CURDLEV+1,"checking for pfield matrix [%dx%d] %p (with last %d,%d) %s...\n",ROWSM(e),COLSM(e),e,klr,klc,ch==-7?"randomized":"");
	ret = ematrix_inpfield_core(ch,e,klr,klc);
	
	if (ch<=-7 && ret>=0) {		/* when only randomized test was performed above, test 2x2 */
		for (i=0; i<ROWSM(e)-1; i++)  for (j=0; j<COLSM(e)-1; j++)
		  for (ii=klr? ROWSM(e)-1:i+1; ii<ROWSM(e); ii++)
		    for (jj=klc? COLSM(e)-1:j+1; jj<COLSM(e); jj++) {
				ret = ematrix_determinant_2x2check(e,i,ii,j,jj)<0? -1:ret;
				if (ret<0)  i=j=ii=jj = 9999;
		}
	}
	DEBUG(CURDLEV+1," ...%s checking for pfield found %d\n",ch==-7?"randomized":"",ret);
	if (ch>0 && ret<0) {
		USERERROR("Minimal non-%s submatrix of size %d found here:",pfield_curname(),sizmin);
		if (estmin)  EMATOUTPUT(estmin,":\t");	/* static variables are used to keep the smallest one found... */
		if (estmin)  dispose_ematrix(estmin);
	}
#ifndef FASTPROG
	if (ret>=0 && IFRANDDEBUGLESS(444) && ch>-6 && klr==0 && klc==0) {
		for (ii=jj=0; ii<4; ii++) {	/* extra testing with pivoted matrix when not random */
			i = RANDOM()%ROWSM(e);  j = RANDOM()%COLSM(e);
			if (SIGNM(e,i,j)!=0)  jj = ematrix_pivot_check(0,e,i,j)<0?-1:jj;
		}
		ematrix_transpose(e);
		jj = ematrix_inpfield_ext(-6,e,0,0)<0?-1:jj;	/* (ch==-6 to prevent inf recursion) */
		if (jj<0)  {PROGERROR("incorrect computation inpfield, ret=%d",ret);}
	}		/* (e is modified here !!!) */
#endif
	dispose_ematrix(e);
	return ret;
}

int	ematrix_inpfield_core(int ch, ematrix *em, int klr, int klc) {
	int	i,k, dr,dc, r,ret=0;
	int	delrx[MXINPF+1],*delr, delcx[MXINPF+1],*delc;
	char	c=' ';
	ematrix	*e, *edelx[MXINPF+1],**edel;
	
	if (ROWSM(em)<=1 || COLSM(em)<=1)  return 0;
	if (ROWSM(em)>=MXINPF || COLSM(em)>=MXINPF)  {PROGERROREXIT("Too large matrix to test in pfields %d,%d >%d.",ROWSM(em),COLSM(em),MXINPF);}
		/**
		 * How the brute force recursive computation works:
		 * First we check the current matrix determinant if it is a square.
		 * Then we remove all possible subsets of lines of the matrix
		 * that result in a square submatrix in backtracking.
		 * At each level we create a local copy of the current matrix.
		 * While the matrix is not yet square, only rows or only columns are removed.
		 * 
		 * delr[k],delc[k] keep the row/column to be deleted for the next step,
		 * edel[k-1] is the current (-smaller) copy of a submatrix,
		**/
	delr = delrx+1;  delc = delcx+1;  delr[-1] = delc[-1] = 0;
	edel = edelx+1;  edel[-1] = em;
	for (i=0; i<MXINPF; i++)  edel[i] = NULL;
	k = 0;  delr[0] = delc[0] = 9999;	/* (starting conditions for the first iteration) */
	ret = 0;
	while (k>=0) {
		e = edel[k-1];
#ifndef FASTPROG
		SDEBUG(CURDLEV+2,"%c",c);  c = '\n';
		DEBUG(CURDLEV+2,"%*s - calling [%dx%d] at level k=%d,  delr=%d, delc=%d",2*k,"",ROWSM(e),COLSM(e),k,delr[k]<9999?delr[k]:9,delc[k]<9999?delc[k]:9);
		if (ROWSM(e)<=1 || COLSM(e)<=1 || (k>=ROWSM(em)-1 && k>=COLSM(em)-1))
			{PROGERROR("something wrong in computing, should not get here (k=%d)",k); EMATDEBUG(0,e,"! ");}
#endif
		if (ROWSM(e)==COLSM(e)) {
			ret = (r = ematrix_determinant_ch(DECRCH(ch),e,NULL,NULL))<0? -1:ret;
			if (r<0 && ch>0 && ROWSM(e)<sizmin) {
				if (estmin)  dispose_ematrix(estmin);
				estmin = ematrix_copy(e);	/* a smaller non-pfield matrix is recorded */
				sizmin = ROWSM(e);
			}		/* cycling until the smallest is found for ch>0, forever if ch>1 */
			if (ret<0 && (ch<=0 || ch*sizmin<=2))  break;
		}
		if (ROWSM(e)==2 && COLSM(e)==2) {
			--k;  continue;		/* smaller matrices are not tested - no sense */
		}
		if (delr[k-1]>=ROWSM(e)-klr || delc[k-1]>=COLSM(e)-klc) {
			--k;  continue;		/* no more lines to delete after the previous one */
		}
			/**
			 * The exhausted search may be randomized by using, instead of ++,
			 * random increments of delc,delr (by a fraction of total rows,cols).
			 * A bad submatrix would be picked very quickly then, and a test in the
			 * good case would not take long and probably would be correct as well.
			**/
		delc[k]++;				/* increase deletion column */
		if (ch<=-7 && COLSM(em)>5)  delc[k] += RANDOM()%(2+COLSM(em)/8);
		if (delc[k]>=COLSM(e)-klc || ROWSM(e)>COLSM(e)) {
			delc[k] = delc[k-1];
			if (delr[k]>ROWSM(e)) {		/* (first iteration at this level) */
				delr[k] = delr[k-1];
			} else {
				delr[k]++;		/* increase deletion row */
				if (ch<=-7 && ROWSM(em)>5)  delr[k] += RANDOM()%(2+ROWSM(em)/8);
				if (delr[k]>=ROWSM(e)-klr || ROWSM(e)<COLSM(e)) {
					--k;  continue;	/* no more choices here for rows or columns */
				}
			}
		}
		dc = ROWSM(e)<=COLSM(e)? delc[k]:-1;  dr = ROWSM(e)>=COLSM(e)? delr[k]:-1;
#ifndef FASTPROG
		SDEBUG(CURDLEV+3,"\t -> dr=%d, dc=%d\n",dr,dc);
#endif
		if (edel[k])  dispose_ematrix(edel[k]);
		e = ematrix_copy(e);  ematrix_remove(e,dr,dc);
		edel[k] = e;			/* the smaller matrix for the next step */
		++k;
		delr[k] = delc[k] = 9999;
	}
	for (i=0; i<MXINPF; i++)  if (edel[i])  dispose_ematrix(edel[i]);
	c = ' ';
	return ret;
}



/**
 * This function is called to "import" the matrix ee into the current pfield
 * by the current selected translation (see pfield_translation_ext() in pfield.c).
 * Importing means translating the exponents of the entries of ee into powers
 * of the images of generators, as defined by the translation.
 * 
 * The imported entries are stored back to the given matrix ee.
 * The return value is >=0 for success, and <0 for failure.
**/

int	ematrix_import_ext(int ch, ematrix *ee) {
	int	i,j,r,g;
	
	for (i=r=0; i<ROWSM(ee) && r>=0; i++)
		for (j=0; j<COLSM(ee) && r>=0; j++) {
			g = SIGNM(ee,i,j);
			r = pfield_translation_ch(ch,&EXPM(ee,i,j),&SIGNM(ee,i,j));
			if (g && !SIGNM(ee,i,j)) { r = -1;  if (ch>=0) USERERROR("Pfield translation to zero 0."); }
		}
	if (ch>0 && r<0)  {PROGERROR("Translation of the matrix %p failed!",ee);}
	return r; 
}

















/******************	Some matrix tests	*******************/
/******************************************************************/


#undef CURDLEV
#define CURDLEV		7


/**
 * This function tests whether the two lines s1,s2 of the matrices e1,e2 are parallel up to
 * the excluded two rows no1,no2 (use -1 to not exclude them).
 * The return value is either -1 for not parallel, or r>=0 if s1,s2 are parallel up to no1,no2.
 * (Handles also cases when one of the columns is 0 - then parallel as well.)
**/

int	ematrix_twoparallel_rc(ematrix *e1, int tr, int s1, ematrix *e2, int s2, int no1, int no2) {
	int	i,f,z, ret;
	exp_t	de, ae;
	sign_t	dg, ag;
	
	pfield_setzeroexp(&de);  dg = 0;
	if (!tr)  ematrix_transpose(e1);
	if (!tr && e1!=e2)  ematrix_transpose(e2);
	if (ROWSM(e1)!=ROWSM(e2))
		{PROGERROR("cannot test parallel lines that have different lengths %d!=%d",ROWSM(e1),ROWSM(e2));}
	
	ret = -1;  f = 0;		/* tests all entries of s1,s2 except excluded no1,no2 */
	for (i=0; i<ROWSM(e1); i++)  if (i!=no1 && i!=no2) {
		if (SIGNM(e1,i,s1)==0 && SIGNM(e2,i,s2)==0)  continue;
					/* computes the fraction - OK to "divide by 0" which results in 0 */
		pfield_mul(1,EXPM(e1,i,s1),SIGNM(e1,i,s1),-1,EXPM(e2,i,s2),SIGNM(e2,i,s2), &ae,&ag);
		z = (SIGNM(e2,i,s2)==0? -1:1);		/* (indicates division by 0: z==-1) */
		if (f==0) {
			de = ae;  dg = ag;  f = z;	/* remembers the difference between the first nonzeros */
		} else {
			if (f!=z || !pfield_isequal(ae,ag,de,dg))  break;
		}			/* (must distinguish when s2 is a 0-multiple of s1 as well) */
	}
	if (i>=ROWSM(e1))  ret = 0;	/* if the above cycle was not broken for j, then parallel */
	if (!tr)  ematrix_transpose(e1);
	if (!tr && e1!=e2)  ematrix_transpose(e2);
	
#ifndef FASTPROG			/* an extra test for symmetric result... */
	if (IFRANDDEBUGLESS(225) && !inrecur) {
		inrecur = 1;		/* (to avoid inf recursion) */
		if ((ret<0)!=(ematrix_twoparallel_rc(e2,tr,s2,e1,s1,no2,no1)<0))
			{PROGERROR("Non-symmetric twoparallel computation");}
		inrecur = 0; }
#endif
	return ret;
}


/**
 * This function tests whether the line so of the matrix e is all-zero up to
 * the excluded three rows no1,no2,no3 (use -1 to not exclude them).
 * If nz==1, then the excluded rows must be nonzero to pass.
 * The return value is either -1 for other nonzero, or 1 for all but no1,no2 zero entries.
**/

int	ematrix_linezero_ext(ematrix *e, int tr, int so, int no1, int no2, int no3, int nz) {
	int	i,ret;
	
	if (!tr)  ematrix_transpose(e);
	for (i=0; i<ROWSM(e); i++)
		if (i!=no1 && i!=no2 && i!=no3 && SIGNM(e,i,so)!=0)  break;
	ret = (i<ROWSM(e)? -1:1);
	if (nz>0) if ((no1>=0? SIGNM(e,no1,so)==0:0) || (no2>=0? SIGNM(e,no2,so)==0:0) ||
		(no3>=0? SIGNM(e,no3,so)==0:0))  ret = -1;
	if (!tr)  ematrix_transpose(e);
	return ret;
}


/**
 * This function computes whether the given matrix e has a column parallel to the given
 * column (tr==1) so of a matrix eo, or similarly for the rows (tr==0).
 * If e==eo, then the column so is not tested against itself (of course).
 * The parameter ch is not used now.
 * The return value is either -1 for no parallel, or r>=0 if a column r is parallel to so.
**/

int	ematrix_lineparallel_ext(int ch, ematrix *e, int tr, ematrix *eo, int so) {
	int	j, ret;
	
	if (!tr)  ematrix_transpose(e);
	if (!tr && e!=eo)  ematrix_transpose(eo);
	junk = ch;  ret = -1;
	for (j=0; j<COLSM(e) && ret<0; j++)
		if (j!=so || eo!=e) {
			ret = ematrix_twoparallel_rc(e,1,j,eo,so,-1,-1);
			if (ret>=0)  ret = j;
		}
	if (!tr)  ematrix_transpose(e);
	if (!tr && e!=eo)  ematrix_transpose(eo);
	return ret;
}


/**
 * This function computes whether the given matrix e has two parallel columns or rows.
 * The return value is either -1 for no parallel, or r=0 if two rows are parallel,
 * or r=1 if two columns are parallel, or r=2 if both happen.
**/

int	ematrix_parallel_any(ematrix *e) {
	int	i,t, ret;
	
	ret = -1;
	for (t=0; t<2; t++) {
		ematrix_transpose(e);
		for (i=0; i<COLSM(e) && ret<0; i++)
			if (ematrix_parallel_col(e,i)>i)  ret = t+ (ret<0? 0:t);
	}
	return ret;
}


/**
 * This function returns the rank of the given matrix e (as an integer).
 * If the matrix is not representable and this is found here, then the result is -1.
 * (The matrix e is not modified here.)
**/

int	ematrix_matrank_ext(ematrix *e) {
	int	i,j,c, rk;
	ematrix	*ee;
	
	ee = ematrix_copy(e);
	rk = 0;
	while (ROWSM(ee)>0 && COLSM(ee)>0 && rk>=0) {
		j = COLSM(ee)-1;
		for (i=ROWSM(ee)-1; i>=0 && SIGNM(ee,i,j)==0; i--) ;
		if (i<0) {		/* zero column is ignored */
			ematrix_remove_col(ee,j);
		} else {
			rk++;		/* nonzero column is pivoted and the row-column removed */
			c = ematrix_pivot_check(0,ee,i,j);
			ematrix_remove(ee,i,j);
			if (c<0)  rk = -1;	/* for the case that the matrix is not representable */
		}
	}
	dispose_ematrix(ee);
	return rk;
}


/**
 * This function returns whether the three lines so1,so2,so3 (row/col by tr) of the matrix e
 * are dependent (1 for dep, -1 for not).
 * If the matrix is not representable, then the return value is like for dependent.
 * (This is for testing triangles and triads in the matrix, but only without parallel/serial elements.)
**/

int	ematrix_triax_rc(ematrix *e, int tr, int so1, int so2, int so3) {
	ematrix	*re;
	int	r;
	
	if (!tr)  ematrix_transpose(e);
	re = ematrix_refer(e,0,ROWSM(e),-1,-1);
	ematrix_refadd_col(re,so1);
	ematrix_refadd_col(re,so2);
	ematrix_refadd_col(re,so3);
	r = ematrix_matrank(re);
	dispose_ematrix(re);
	if (!tr)  ematrix_transpose(e);
#ifndef FASTPROG
	if (IFRANDDEBUGLESS(555) && !inrecur) {
		re = ISREFMAT(e)? ematrix_copy(e):e;  inrecur = 1;	/* (to avoid inf recursion) */
		if ((r<3)!=(ematrix_triax_rc(re,tr,so3,so1,so2)>0))  if (ematrix_inpfield(re)>=0)
			{PROGERROR("Wrong (refering) istriax computation or non-symmetric, tr=%d, so1=%d, so2=%d, so3=%d, ret=%d",tr,so1,so2,so3,r); EMATDEBUG(0,re,"!\t");}
		if (ISREFMAT(e))  dispose_ematrix(re);  inrecur = 0;
	}
#endif
	return r<3? 1:-1;
}














/******************	Closures and Separations	*******************/
/**************************************************************************/


#undef CURDLEV
#define CURDLEV		7


/**
 * This function returns the (co-)closure of the given submatrix re in the matrix e.
 * "Closure" means all matroid elements (rows and columns of e) that belong to the subspace
 * spanned by the subset corresponding to the lines of re.
 * "Co-closure" is the closure in the dual matroid (transposed matrix).
 * Closure is computed for tr==1, and co-closure for tr==0.
 * The matrix must be properly representable.
 * The return value is a new refering matrix (to e) containing all lines of re and of the
 * closure.
**/

ematrix*	ematrix_closure_ext(ematrix *e, ematrix *re, int tr) {
	ematrix		*ex, *ee,*ee2,*re2, *rre;
	int		i,ii,j, r0,r1;
	
	if (ISREFMAT(e) || !ISREFMAT(re) || REFMAT(re)!=e)  {PROGERROREXIT("call only for a refering matrix re -> e");}
	if (ISTRANSPM(e)!=ISTRANSPM(re))  {PROGERROR("call only for matrices e,re of the same transpose");}
	if (!tr) {
		ematrix_transpose(e);  ematrix_transpose(re);
	}
	rre = ematrix_refer_all(re);
	ex = ematrix_refextract_xrow(e,re);
	r0 = ematrix_matrank(ex);
		/**
		 * A column of e belongs to the closure of re iff it is dependent on the submatrix
		 * of e on the columns as re and on the rows not in re (matrix ex here).
		 * So we add corresponding part of a column to the matrix ex and compare the rank.
		**/
	for (j=0; j<COLSM(e); j++) {
		for (ii=0; ii<COLSM(ex) && GETREFMCOL(ex,ii)!=j; ii++) ;
		if (ii<COLSM(ex))  continue;		/* (column already in re) */
		ematrix_drefadd_col(ex,j);
		r1 = ematrix_matrank(ex);
		ematrix_remove_col(ex,COLSM(ex)-1);
		if (r1<0 || r0<0)  {PROGERROR("the matrix must be representable here");}
		if (r1<r0)  {PROGERROR("wrong rank computation of added column");}
		if (r1>r0)  continue;
		ematrix_drefadd_col(rre,j);
	}
		/**
		 * We then find all rows not in re but belonging to the closure.
		 * We actually compute with a row like with an extra column parallel to that row
		 * (so we are appending unit columns to a copy of the matrix ex).
		 * The rest is the same as previously.
		**/
	ee = ematrix_copy(ex);
	ematrix_resetid(ee);		/* (must prevent duplicate line id's for append) */
	ee = ematrix_append_col(ee);
	for (i=0; i<ROWSM(e); i++) {
		for (ii=0; ii<ROWSM(ex) && GETREFMROW(ex,ii)!=i; ii++) ;
		if (ii>=ROWSM(ex))  continue;		/* (row already in re - not in ex) */
		for (j=0; j<ROWSM(ee); j++)  SETEXSIGMZERO(ee,j,COLSM(ee)-1);
		SIGNM(ee,ii,COLSM(ee)-1) = 1;		/* must index rows according to ex ! */
		r1 = ematrix_matrank(ee);
		if (r1<0 || r0<0)  {PROGERROR("the matrix must be representable here");}
		if (r1<r0)  {PROGERROR("wrong rank computation of added row");}
		if (r1>r0)  continue;
		ematrix_drefadd_row(rre,i);
	}
	dispose_ematrix(ee);
	re2 = ee2 = NULL;
#ifndef FASTPROG		/* some extra paranoic tests - double closure and pivoted closure */
	if ((IFRANDDEBUGLESS(222)) && !inrecur) {
		inrecur = 1;		/* (to avoid inf recursion) */
		ee = ematrix_closure_ext(e,rre,1);
		if (!ematrix_isequal(ee,rre))  {PROGERROR("wrong double-closure computation");}
		dispose_ematrix(ee);
		ee2 = ematrix_copy(e);	/* trying to pivot the matrix not touching re-lines */
		re2 = ematrix_refer_all(re);  ematrix_rerefer(re2,ee2);
		i = ROWSM(rre)>ROWSM(re)? GETREFMROW(rre,ROWSM(rre)-1): RANDOM()%ROWSM(ee2);
		j = RANDOM()%COLSM(ee2);
		for (ii=0; ii<ROWSM(re2) && GETREFMROW(re2,ii)!=i; ii++) ;
		if (ii<ROWSM(re2))  i = -1;
		for (ii=0; ii<COLSM(re2) && GETREFMCOL(re2,ii)!=j; ii++) ;
		if (ii<COLSM(re2))  j = -1;
		if (i>=0 && j>=0 && SIGNM(ee2,i,j)!=0)  ematrix_pivot(ee2,i,j);
		ee = ematrix_closure_ext(ee2,re2,1);
		if (ROWSM(ee)+COLSM(ee)!=ROWSM(rre)+COLSM(rre)) {
			PROGERROR("wrong closure computation after pivoting");
			EMATDEBUG(0,e,"e\t"); EMATDEBUG(0,rre,"!r\t"); EMATDEBUG(0,ee2,"e2\t"); EMATDEBUG(0,ee,"!n\t"); }
		dispose_ematrix(ee);
		dispose_ematrix(re2);  dispose_ematrix(ee2);
		inrecur = 0;
	}
#endif
	dispose_ematrix(ex);
	if (!tr) {
		ematrix_transpose(e);  ematrix_transpose(re);
		ematrix_transpose(rre);	/* (the output matrix must be transposed back as well!) */
	}
	return rre;		/* the closure is returned as a new refering matrix */
}



/**
 * This function computes the connectivity of the given submatrix re in e,                                       
 * i.e. the separation value for the separation (re,e-re) minus one.
 * We use the formula for (A,B) as lambda(A) = r(A)+r(B)-r(M).
 * So the value of this function is 0 for 1-separation, 1 for exact 2-separation, etc.
 * 
**/

int	ematrix_whatsep(ematrix *e, ematrix *re) {
	ematrix	*rxr,*rxc;
	int	sp;
	
	if (ISREFMAT(e) || !ISREFMAT(re) || REFMAT(re)!=e)  {PROGERROREXIT("call only for a refering matrix re -> e");}
	if (ISTRANSPM(e)!=ISTRANSPM(re))  {PROGERROR("call only for matrices e,re of the same transpose");}
	
	rxr = ematrix_refextract_xrow(e,re);
	rxc = ematrix_refextract_xcol(e,re);
	sp = ematrix_matrank(rxr)+ematrix_matrank(rxc);
	
	DEBUG(CURDLEV+1,"Computed separation value %d for:\n",sp);
	EMATDEBUGS(CURDLEV+1,rxr,"\t\t\t"); EMATDEBUGS(CURDLEV+1,rxc,"\t\t\t");
	dispose_ematrix(rxr);  dispose_ematrix(rxc);
	return sp;
}












































