
/********************************************************************************
 *										*
 *				 M A C E K					*
 *		("Matroids Also Computed Efficiently" Kit)			*
 *										*
 * A set of tools and routines for computations with representable matroids.	*
 * Copyright (C) 2001,2002  Petr Hlineny.					*
 *										*
 *   This program is free software; you can redistribute it and/or modify	*
 *   it under the terms of the GNU General Public License as published by	*
 *   the Free Software Foundation; either version 2 of the License, or		*
 *   (at your option) any later version.					*
 *   You should have received a copy of the GNU General Public License		*
 *   along with this program; if not, write to the Free Software		*
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.	*
 * 										*
 * See the top-level "ABOUT" and "doc/ *" files for more information.		*
 *										*
 ********************************************************************************/














/**
 * Some theory about generating matroids for start...
 * The master description of the generating process is in ../include/gener.h.
 * 
 * This part handles extending the given matrix by one line.
 * (A job that may seem simple, but this is not so for partial fields with infinitely
 * many elements; then the choice for extension elements gets quite tricky...)
 * All possible (connected) extensions are generated; in the partial case even many
 * non-properly represented ones.
 * See gener.c for the additional filtering applied to the generated extensions.
 * 
 * 
**/


#include "macek.h"
#include "gen.h"











/******************	Generating matrix extensions	*******************/
/**************************************************************************/


#define	CURDLEV		6		/* (use <=DEBUGLEV+2 to perform extra checks) */



/**
 * This function generates column extensions to the matrix ex.
 * All distinct extension lines are generated here, normalized to unit-leading form.
 * When extending over a partial field, then also some pfield-checks of the matrix
 * are used to restrict the number of generated choices.
 * (But the extension is not guarranteed to be properly pfield-represented!)
 * The function works only for connected(!) matrices in the partial case.
 * 
 * The function returns a NULL-terminated list (alist) of pointers to the new matrices.
 * (Remember to free the list and the matrices later.)
**/

#define	MXEX	(NUM_LONG_BITS-3)		/* (must fit into long bits) */


ematrix**	gener_matextens_ext(int ch, ematrix *ex) {
	long		*ppat;
	int		i,k,r, rcon[MXEX], selsg[MXEX],selex[MXEX], nex=0;
	ematrix		*ee, *edet, *rre[MXEX], **eout;
	
#ifndef FASTPROG
	if (ch<-1)  {PROGERROR("Do not call with ch==%d <-1",ch);}
	DEBUG(CURDLEV-1,"Going to generate extensions of a %dx%d matrix %p ...\n",ROWSM(ex),COLSM(ex),ex);
	if (ch>1)  EMATDEBUGS(CURDLEV-0,ex,"  ex\t");
#endif
	eout = new_alist(32);		/* the return list is empty now */
	ee = ematrix_copy_to(ex,ROWSM(ex),COLSM(ex)+1);
	ee = ematrix_append_col(ee);	/* to get exactly as needed maximal dimensions - faster */
	if (ROWSM(ee)>=MXEX || COLSM(ee)>=MXEX)	/* to avoid long type overflow... (not practical anyway) */
		{PROGERROREXIT("not designed to deal with %d,%d > %d lines in extension",ROWSM(ee),COLSM(ee),MXEX);}
	for (i=0; i<MXEX; i++)  rre[i] = NULL;	/* supplementary refering matrices for repr pre-check */
	
	if (pfispartial) {		/* getting "connected order" and supplementary matrix for generating */
		gener_matextens_conorder(ee,rcon,rre,&edet,&ppat);
	} else {
		ppat = NULL;  edet = NULL;	/* (not using "connected order" for a full field) */
		for (i=0; i<MXEX; i++)  rcon[i] = i;
	}
	k = 0;
		/**
		 * We construct the k-th entries of the added line for k=0,1,2,... here:
		 * We use functions pfield_get[pa]value() to guess possibilities
		 * for the entry inside gener_matextens_cycled().
		 * The matrix ee is used here over and over as the working storage.
		 * The new entries are stored into ee automatically;
		 * selsg[],selex[] keep the possibilities for our choices
		 * (value -1 means the initial choice).
		**/
	selsg[0] = selex[0] = -1;
	while (k>=0) {
		DEBUG(CURDLEV+4,"%*s calling gener_matextens_cycled() at level k=%d\n",2*k,"",k);
		r = gener_matextens_cycled(k,ee,rcon,rre,edet,ppat,selex,selsg);
		if (r<0) {
			k += r;  continue;	/* going back in backtracking (r<0 here!) */
		} else if (r==0) {
			continue;		/* nothing useful for now, cycling again */
		} else if (r>0 && k<ROWSM(ee)-1) {
			++k;  selsg[k] = selex[k] = -1;
			continue;		/* going forward in backtracking - new cycle at level k+1 */
		}
		++nex;				/* when completed one extension... (r>0 here) */
#ifndef FASTPROG
		if (CURDLEV+0<=DEBUGLEV && ch>0) {
			DEBUG(CURDLEV+0,"> extension #%d found: ",nex);
			for (i=0; i<ROWSM(ee); i++)  SDEBUG(CURDLEV+0," %s,",pfield_pvalue(6,EXPM(ee,i,COLSM(ee)-1),SIGNM(ee,i,COLSM(ee)-1)));
			SDEBUG(CURDLEV+0,"\n"); }
		if (ch>1)  EMATDEBUGS(CURDLEV+1,ee,"\t\t`\t");
#endif
		eout = alist_append(eout,ematrix_copy(ee));
	}
	for (i=0; i<MXEX; i++)  if (rre[i])	/* (used for supplementary refering matrices in repr pre-check) */
		dispose_ematrix(rre[i]);
#ifndef FASTPROG
	DEBUG(CURDLEV-1,"  ... done, generated %d %s extensions of matrix %p\n",nex,pfield_curname(),ex);
	ematrix_remove_col(ee,COLSM(ee)-1);
	if (!ematrix_isequal(ee,ex))  {PROGERROR("The matrix ee was damaged here inside!");}
#endif
	dispose_ematrix(ee);			/* (used in generating extensions) */
	if (ppat)  FREE(ppat);			/* (used to store "connected order" information) */
	if (edet)  dispose_ematrix(edet);
	return eout;		/* the output array of matrices is returned, terminated by NULL */
}





/******************	Supplements for extensions	*******************/
/**************************************************************************/


/**
 * Here we compute a "connected order" and "2x2determinant connections" for generating
 * extensions of the matrix ee in above.
 * This function is reasonable only for a (really) partial field.
 * We store the connected order to the array rcon[] (must be allocated before),
 * and we store the "2x2determinant connections" to the matrix *edet and array *ppatx
 * (which are allocated here).
 * Moreover, we store to the array rrce[] refering matrices that refer to the rows of ee
 * in order given by rcon[] - for later use in a representability pre-check.
 * 
 * The rows of *edet correspond to the rows of ee, but in the connected order given in rcon[].
 * Each column of *edet corresponds to an (unordered) pair of rows of *edet.
 * Each entry of *ppatx corresponds to an (unordered) pair of rows as well, and the bits
 * of each entry of *ppatx are indexed by the rows of *edet.
 * 
 * The practical meaning of edet and ppatx is the following:
 * If r1,r2 are two rows of *edet, both entries l1,l2 of rows rcon[r1],rcon[r2] in the last column
 * in ee are nonzero, and other entries in the last column of ee corresponding to 1's in the
 * bitfield *ppatx[EERIX(r1,r2)] (bits remapped by rcon[]!) are zero, then the entries d1,d2 in
 * the column EERIX(r1,r2) of *edet at rows r1,r2 are nonzero and d1*l2-l1*d2 is equal to
 * a (+-) subdeterminat of the extended ee.
 * (This allows to guess possible values of l2 knowing l1 and d1,d2 in a partial field.)
 * The 1-bits of entries of *ppatx may also be viewed as the internal vertices of an "induced
 * path" joining the two rows r1,r2 ("induced" guaranties that d1,d2!=0).
 * Additionally, all 1-bits of *ppatx[EERIX(r1,r2)] are at smaller positions than max(r1,r2).
 * 
**/

#define	EERIX(e,r1,r2)	((r1)>(r2)? ((r1)+(r2)*ROWSM(e)):((r2)+(r1)*ROWSM(e)))
					/* (this must be symmetric in r1,r2 !) */

void	gener_matextens_conorder(ematrix *ee, int rcon[], ematrix *rrce[], ematrix **edet, long **ppatx) {
	int	i,j,k, n,nn,ll;
	long	zf,zp,zc, rozp[MXEX], *cpat,*ppat;
	ematrix	*re;
	exp_t	xx;
	sign_t	gg;
	
	DEBUG(CURDLEV-1,"Computing connected order and \"connecting\" matrix for %p...\n",ee);
	
	ll = EERIX(ee,ROWSM(ee)-1,ROWSM(ee)-1)+2;	/* (need an extra column for generated entries!) */
	*edet = new_ematrix(ROWSM(ee),ll,ROWSM(ee),ll);
	ppat = *ppatx = MMALLOC(ll*sizeof(ppat[0]));
	cpat = MMALLOC(ll*sizeof(cpat[0]));
		/**
		 * We first need to find a "connected order" of rows - so that each next row has
		 * a common nonzero column with one of the previous.
		 * (Matrix ee is supposed to be connected!)
		**/
	for (i=0; i<ROWSM(ee); i++)	/* records the zero-patterns of rows of ee (without the last column) */
		for (rozp[i]= j=0; j<COLSM(ee)-1; j++)
			if (SIGNM(ee,i,j)!=0)  rozp[i] |= (1l<<j);
	rcon[0] = 0;  k = 0;		/* finds a connected order of columns of ee, stored in rcon[] */
	zf = (1l<<rcon[0]);  zp = rozp[rcon[0]];
	while (++k<ROWSM(ee)) {
		for (j=0; j<ROWSM(ee); j++)  if ((zf&(1l<<j))==0 && (zp&rozp[j])!=0)  break;
		if (j>=ROWSM(ee))  break;	/* (should not happen for connected matrix...) */
		rcon[k] = j;   zf |= (1l<<j);  zp |= rozp[j];
	}
	if (zf!=(1l<<ROWSM(ee))-1) {	/* if the matrix was not connected, we are not able to finish */
		PROGERROR("Error computing - NOT connected order in the matrix %p",ee);  EMATDEBUG(0,ee," !\t");
		*edet = NULL;  *ppatx = NULL;  return;
	}
		/**
		 * We define a graph on the rows of edet where edges connect pairs of rows
		 * sharing a common nonzero column (we record one of such columns as well).
		 * For each pair of rows in this graph we find an induced path joining them
		 * and using only other rows before (=not below the lower one in the matrix).
		 * We record the internal rows of these paths in the bitfields ppat[], and
		 * the columns needed for the paths in the bitfields cpat[].
		 * (The paths need not be shortest, just induced.)
		**/
	for (i=0; i<ROWSM(ee); i++)  for (j=i; j<ROWSM(ee); j++)
		ppat[EERIX(ee,i,j)] = -1;
	for (k=-1; k<ROWSM(ee); k++) {	/* (we use classical "matrix" algorithm for computing a graph metric */
		ll = 0;
		for (i=0; i<ROWSM(ee); i++)  for (j=i+1; j<ROWSM(ee); j++) {
			if (i==k || j<=k)  continue;	/* (only rows before j may be used inside paths) */
			
			if (k==-1) {	/* the first run records edges between rows */
				zp = (rozp[rcon[i]]&rozp[rcon[j]]);
				for (nn=-1,n=0; n<COLSM(ee)-1; n++) {
					if ((zp&(1l<<n))!=0)  nn = n;
					if (nn>=0 && (n+i+j)%3==0)  break;	/* (not to get always the first nonzero) */
				}
				if (zp!=0) {	/* the path(=edge) uses column nn and n internal vertices */
					ppat[EERIX(ee,i,j)] = 0;  cpat[EERIX(ee,i,j)] = (1l<<nn);
				}
			} else {	/* the next runs look for better paths between rows */
				if (ppat[EERIX(ee,i,j)]<0)
				  if (ppat[EERIX(ee,i,k)]>=0 && ppat[EERIX(ee,k,j)]>=0) {
					cpat[EERIX(ee,i,j)] = (cpat[EERIX(ee,i,k)]|cpat[EERIX(ee,k,j)]);
					ppat[EERIX(ee,i,j)] = (ppat[EERIX(ee,i,k)]|ppat[EERIX(ee,k,j)]|(1l<<k));
			}	}
		}
		for (i=0; i<ROWSM(ee); i++)  for (j=i+1; j<ROWSM(ee); j++)
			if (ppat[EERIX(ee,i,j)]<0)  ll = 1;
		if (ll==0)  break;	/* when all paths (connections) are already found */
	}
	if (k>=ROWSM(ee))  {PROGERROREXIT("wrong computation of connected order, should not happen!");}
#ifndef FASTPROG
	DEBUG(CURDLEV-1," - ordering rows : \n\t\t");
	for (i=0; i<ROWSM(ee); i++)  SDEBUG(CURDLEV-1,"#%d: %d(%d),  ",i,rcon[i],ROWSID(ee,rcon[i]));
	SDEBUG(CURDLEV-1,"\n");
#endif
		/**
		 * Here we use the induced paths computed above to construct and evaluate
		 * subdeterminants of ee that consist of rows and columns of the paths.
		 * Since the paths are induced, the subdeterminants against i and against j
		 * have only one nonzero permutation each, so they are evaluated nonzero.
		**/
	for (i=0; i<ROWSM(ee); i++)  for (j=i+1; j<ROWSM(ee); j++) {
		zp = ppat[EERIX(ee,i,j)];  zc = cpat[EERIX(ee,i,j)];
		if (zp<0)  {PROGERROR("wrong computation of connected order, should not happen! i=%d, j=%d, zp=%lX<0",i,j,zp);}
		re = ematrix_refer_empty(ee);
		for (n=0; n<COLSM(ee)-1; n++)		/* add the columns of the path */
			if ((zc&(1l<<n))!=0)  ematrix_refadd_col(re,n);
		for (n=0; n<ROWSM(ee); n++)		/* add the internal rows of the path */
			if ((zp&(1l<<n))!=0)  ematrix_refadd_row(re,rcon[n]);
		ematrix_refadd_row(re,rcon[i]);
		ematrix_determinant(re,&xx,&gg);	/* the subdeterminant against rcon[j] */
		SETEXSIGM(*edet,i,EERIX(ee,i,j),xx,gg);
		if (gg==0)  {PROGERROREXIT("wrong computation of connected order, should not happen!!!");}
		ematrix_remove_row(re,ROWSM(re)-1);
		ematrix_refadd_row(re,rcon[j]);
		ematrix_determinant(re,&xx,&gg);	/* the subdeterminant against rcon[i] */
		SETEXSIGM(*edet,j,EERIX(ee,i,j),xx,gg);
		if (gg==0)  {PROGERROREXIT("wrong computation of connected order, should not happen!!!");}
		dispose_ematrix(re);
	}
		/**
		 * Here we prepare refering matrices that point to the rows of ee in the order
		 * rcon[], and that are later used to pre-check representability in pfield
		 * for the first several choice of entries in gener_matextens_cycled().
		 * (The current pointers to these matrices are all NULL.)
		**/
	if (GEN_USEEXTPREPF>0)  for (i=2; i<ROWSM(ee)-3; i++) {
		rrce[i] = ematrix_refer(ee,rcon[0],rcon[0]+1,0,COLSM(ee));
		for (n=1; n<=i; n++)  ematrix_refadd_row(rrce[i],rcon[n]);
	}
#ifndef FASTPROG
	DEBUG(CURDLEV+0," - 2x2determinant connections :\n");
	for (i=0; i<ROWSM(ee); i++)  for (j=i+1; j<ROWSM(ee); j++) {
		SDEBUG(CURDLEV+0,"\t\t [#%d (%d), #%d (%d)]:\t ppat=%lX cpat=%lX,  det1=%s  det2=%s\n",
			i,ROWSID(ee,rcon[i]), j,ROWSID(ee,rcon[j]),
			ppat[EERIX(ee,i,j)],cpat[EERIX(ee,i,j)],
			pfield_printvalue(16,EXPM(*edet,i,EERIX(ee,i,j)),SIGNM(*edet,i,EERIX(ee,i,j))),
			pfield_printvalue(16,EXPM(*edet,j,EERIX(ee,i,j)),SIGNM(*edet,j,EERIX(ee,i,j))));
	}
	DEBUG(CURDLEV+1," - rrce[] inpfield-pre-checking matrices :\n");
	for (i=0; i<ROWSM(ee); i++)  if (rrce[i])  EMATDEBUGS(CURDLEV+1,rrce[i],"\t\t");
	DEBUG(CURDLEV-1,"- Done with connected order and matrix.\n");
#endif
	FREE(cpat);	/* cpat[] is not needed further, but ppat[] must be kept! */
}



/**
 * This function implements one iteration of a generating cycle in gener_matextens_ext() above.
 * The iteration decides the entry on the row rcon[k] of the extended column of ee.
 * The meaning of edet and ppat is explained in gener_matextens_conorder() above.
 * The values selsg[k],selex[k] tell the last choice of sign and exponent for the entry.
 * (Values selsg[k],selex[k]==-1 start choosing for the entry, and there is a common
 * choice for the positive and negative entries.)
 * Function pfield_get[pa]value() is used to set the exponent values allowed.
 * Moreover, the function pre-checks representability of a submatrix to speed-up generating
 * the new entries if there are many rows - using refering matrices in rrce[] that are set-up
 * in gener_matextens_conorder() (like rcon[]).
 * 
 * The return value -1 means go back in backtracking (no more choices here), 0 means to try
 * a new iteration, and 1 means successful try - go forward or record the whole extension.
**/

/* #define EERIX(e,r1,r2) defined above... (using ee as the matrix) */


int	gener_matextens_cycled(int k, ematrix *ee, int rcon[], ematrix *rrce[], \
				ematrix *edet, long ppat[], int selsg[], int selex[]) {
	int	i,j,jf=0, r;
	long	zcp;
	exp_t	xx,xo;
	sign_t	gg,go;
	
	if (pfispartial && (!edet || !ppat))  {PROGERROREXIT("Must get edet when in partial field!");}
	zcp = 0;			/* (zero pattern of the new column by rcon[] order) */
	for (i=0; i<k; i++)  if (SIGNM(ee,rcon[i],COLSM(ee)-1)!=0)  zcp |= (1l<<i);
	if (selex[k]<=0) {
		++selsg[k];		/* the next value of the sign (trying +-1,0) */
		if (selsg[k]>1) {
			DEBUG(CURDLEV+3,"%*s calling \"cycled\" at k=%d:  ^back^\n",2*k,"",k);
			return -1;	/* choices are exhausted... */
		}
	}
#ifndef FASTPROG
	DEBUG(CURDLEV+3,"%*s calling \"cycled\" at k=%d: selsg=%d, selex=%d, zcp=%lX\n",2*k,"",k,1-selsg[k],selex[k],zcp);
	EMATDEBUGS(CURDLEV+4,ee,"\t\t\t\t\t");  if (edet) EMATDEBUGS(CURDLEV+4,edet,"");
#endif
		/**
		 * The current values of selsg[k] and selex[k] tell what is our next choice:
		 * selsg[k]==0 chooses a nonzero entry indexed by selex[k] (1...max) in
		 * pfield_get[pa]value(), and selsg[k]==1 chooses zero entry.
		 * The meaning of edet and ppat for a partial field is explained in above.
		 * The index values -1 mean the initial choice for selsg[k] and selex[k].
		 * The first nozero entry in the column is always 1.
		**/
	if (zcp==0 || selsg[k]==1) {	/* no exponent choice for zero or for the leading 1 */
		go = 1-selsg[k];  pfield_setzeroexp(&xo);
		r = 0;
	} else if (edet && ppat) {	/* choosing against "connecting" determinants - partial fields */
		go = 1-selsg[k];
		for (jf=0; jf<k; jf++)	/* finds a nonzero connected to k by an induced path with internal 0-rows */
			if ((zcp&(1l<<jf))!=0 && (zcp&ppat[EERIX(ee,jf,k)])==0)  break;
		if (jf>=k)  {PROGERROR("No induced path for the new entry, why ???"); return -1;}
					/* multiplies (and divides) the other three entries of a 2x2 det */
		pfield_mul(1,EXPM(edet,jf,COLSM(edet)-1),SIGNM(edet,jf,COLSM(edet)-1),
				1,EXPM(edet,k,EERIX(ee,jf,k)),SIGNM(edet,k,EERIX(ee,jf,k)),&xx,&gg);
		pfield_mul(-1,EXPM(edet,jf,EERIX(ee,jf,k)),-SIGNM(edet,jf,EERIX(ee,jf,k)),
				1,xx,gg,&xx,&gg);
		if (selex[k]<=0)	/* selects the unknown new entry *xo,*go */
			selex[k] = pfield_getpvalue_num(xx,gg);
		r = pfield_getpvalue(selex[k],xx,gg,&xo,&go);
	} else {
		go = 1-selsg[k];	/* choosing standalone values - finite fields */
		if (selex[k]<=0)  selex[k] = pfield_getavalue_num();
		r = pfield_getavalue(selex[k],&xo,&go);
	}
	--selex[k];			/* the next choice for the exponent on next call */
	if (r<0)  return 0;
	
		/**
		 * After choosing the extended entry, we try few more pfiel-oriented tests.
		 * We use the remaining (suitable) connecting determinants, and the
		 * refering submatrices from rrce[], for pfield-defined checks.
		 * If an undefined subdeterminant is found here, then this extension is dead.
		**/
	if (edet)  SETEXSIGM(edet,k,COLSM(edet)-1,xo,go);
	if (go!=0 && edet && ppat && pfispartial) {
		for (j=0; j<k && r>=0; j++) {	/* (internal vertices of the path must be disjoint from nonzeros of the extended column) */
			if ((zcp&(1l<<j))!=0 && (zcp&ppat[EERIX(ee,j,k)])==0)
				r = ematrix_determinant_2x2check(edet,j,k,EERIX(ee,j,k),COLSM(edet)-1);
			if (j<=jf && r<0)  {PROGERROR("Wrong setting of the new entry ???");}
		}		/* (the first choice (at row jf) is supposed to be set OK) */
	}
	if (r<0)  return 0;
	DEBUG(CURDLEV+2,"%*s    -> chosen at #%d; l.%d(%d) : %s (selex %d)\n",2*k,"",k,rcon[k],ROWSID(ee,rcon[k]),pfield_pvalue(16,xo,go),selex[k]+1);
	if (edet) {	/* records the chosen entry (and all previous ones if partial) */
		for (i=0; i<=k; i++)
			SETEXSIGM(ee,rcon[i],COLSM(ee)-1,EXPM(edet,i,COLSM(edet)-1),SIGNM(edet,i,COLSM(edet)-1));
	} else {
		SETEXSIGM(ee,rcon[k],COLSM(ee)-1,xo,go);
	}
	if (GEN_USEEXTPREPF>0 && pfispartial)  if (k>=1 && rrce[k]!=NULL) {
		if (!pfield_isequal(xo,go,EXPM(rrce[k],ROWSM(rrce[k])-1,COLSM(rrce[k])-1),SIGNM(rrce[k],ROWSM(rrce[k])-1,COLSM(rrce[k])-1)))
			{PROGERROR("Wrong rrce[%d] matrix for repr pre-check!",k);}
			  	/* (rrce[i] refers to the first i+1 rows of ee in order by rcon[]!) */
		if (ematrix_inpfield_randlast(rrce[k],1,1)<0) {
			DEBUG(CURDLEV+0,"> repr pre-check rejected at k=%d\n",k);
			EMATDEBUGS(CURDLEV+2,rrce[k],"\t\t\t-\t");
			return 0;	/* already a submatrix is not representable, then trying again */
		}
	}
	return 1;		/* successfull extension is returned */
}




































