
/********************************************************************************
 *										*
 *				 M A C E K					*
 *		("Matroids Also Computed Efficiently" Kit)			*
 *										*
 * A set of tools and routines for computations with representable matroids.	*
 * Copyright (C) 2001,2002  Petr Hlineny.					*
 *										*
 *   This program is free software; you can redistribute it and/or modify	*
 *   it under the terms of the GNU General Public License as published by	*
 *   the Free Software Foundation; either version 2 of the License, or		*
 *   (at your option) any later version.					*
 *   You should have received a copy of the GNU General Public License		*
 *   along with this program; if not, write to the Free Software		*
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.	*
 * 										*
 * See the top-level "ABOUT" and "doc/ *" files for more information.		*
 *										*
 ********************************************************************************/














/**
 * Some theory about generating matroids for start...
 * The master description of the generating process is in ../include/gener.h.
 * 
 * This file puts together all tests that must be performed at each (co)extension
 * step when generating an eliminatin sequence - "admissibility tests" and "canonical tests".
 * These tests classes are presented in the next two sections.
 * It is very important to follow the structure of these test classes when adding new
 * particaluar tests...
 * 
**/


#include "macek.h"
#include "gen.h"










/******************	Checking generated matrices	*******************/
/**************************************************************************/


#undef CURDLEV
#define	CURDLEV		6


/**
 * The following functions put together all tests that must be performed at each (co)extension
 * step when reconstructing an eliminatin sequence - "admissibility tests".
 * (These do not include canonical-order testing, which is written in the next section.)
 * Read more about the general theory in the file gener.h ...
 * 
 * In general, there are 5 test classes distinguished by what are the tested properties
 * invariant on - only line scales, also base-minor automorphisms, all sequences of a fixed
 * basis, all equivalences of the matrix, or specially the sequence signature bits only.
 * Remember that one call usually performs the test only for the last added line (except structural),
 * so one must call the tests repeatedly along the sequence to determine validity of the
 * whole elim sequence from scratch (this is implicite in normal generating process).
 * More precise description follows at the functions below.
 * 
 * The given value of lev determines the level of tests - 0 means to do all tests thoroughly,
 * while values >0 (1,2,3 for now) request faster but incomplete tests - to speed up rejection.
 * The value lev==0 must always perform a full test, even when this means a repetition of
 * some previous faster tests.
 * Specially, lev==-3 is used to request a test for test_elimseqm_ext(), which may work
 * slightly differently.
 * If the value tr==-1 is given, then the tests should not use information about the
 * last added line.
 * Keep in mind that the matrix may not be fully valid at levels lev>0, and that the matrix
 * may even not be fully representable at lev>1, so design checks carefully.
 * 
 * 
 * All sequence-dependent properties checked here must be hereditary(!) - that means if
 * something passed at some of the previous steps (of the sequence), then it must be OK
 * also at all subsequent steps (so that the whole sequence need not be re-checked).
 * Moreover, the sequence-dependent checks must be repeated along the sequence for all added
 * lines since they work only on the last line in one call.
 * 
 * The tests should not assume (in general) proper representability of the supplied
 * resulting matrix, since no full representability check is performed at lev>1 (would
 * take too long to complete).
 * However, from the highest-level structure test (genstep_structurecheck(), lev==3),
 * all 2x2 subdeterminants are defined, which is enough for our usual tests.
 * 
 * Additionally, it is possible to use pre-computed data for q in the checks - in calling
 * genstep_pretest_start() before generating starts, and genstep_pretest_stop() afterwards.
 * This call may use the matrix in q only without(!) the last added line.
 * (The validity of the checks must not depend on availability of precomputed data.)
 * See  static elimseqm *precheck_seq;  and  genstep_pretest_ext()...
**/


	/* standard start for subsequent functions - fills e from q if e==NULL: */

#define	GENER_COMMON(q,e,tr)	\
		timeel = TIME();	\
		if (!e) { e = ESMATRIX(q);  if (tr>=0)  tr = (ISTRANSPM(e)!=ESELIMSG(q)%2); }	\
		if (ROWSM(ESMATRIX(q))!=ROWSM(e) || COLSM(ESMATRIX(q))!=COLSM(e) ||	\
						ISTRANSPM(e)!=ISTRANSPM(ESMATRIX(q)))	\
			{PROGERROR("given different size/state matrix %p for seq %p",e,q);}	\
		if (tr>=0 && tr!=(ISTRANSPM(e)!=ESELIMSG(q)%2))	\
			{PROGERROREXIT("wrong transp %d~%d compared to the seq %p",tr,ISTRANSPM(e),q);}	\

static int	timeel,tiel;
#define	TIMEELAPSED()	if ((tiel=TIME()-timeel)>3)  DEBUG(CURDLEV-1,"\t  - time elapsed %d.\n",tiel)



/**
 * This function checks the matrix e (or q) line-properties - i.e. properties dependent on the
 * particular elimination sequence and order of matrix lines (including the base minor lines).
 * (Remember that all nonzero scalings of a matrix line are "the same" for matroids.)
 * A compulsory property to check is the standard scaling form of a line.
 * If set>0, then the line is multiplied to get to the standard scaling form instead.
 * If e is a refering matrix and set>1, then the refered matrix(!!!) by e is scaled instead of e.
 * (Generaly, set>0 should set the property otherwise checked here, but it may sometimes fail.)
 * 
 * The check is carried out on the last line (row if tr==0, column if tr==1) of the
 * matrix e (which is supposed to fit as the resulting matrix of the sequence q).
 * This means that to perform a full check of the whole sequence, one must call this function
 * repeatedly along the elimination sequence!
 * If e is given, then the matrix entries in q are never accessed, only the shape.
 * These checks must not assume that the matrix is properly represented over the pfield.
 * 
 * Do not try to put here any long test, since this function is called very often.
 * In fact, there should probably be no more tests than the unit-scaled line one.
 * The return value is 1 if the matrix passed the test, and -1 otherwise.
**/

int	genstep_linecheck_ext(int set, int lev, elimseqm *q, ematrix *e, int tr) {
	int	r = 1;
	
	if (tr<0)  {PROGERROR("Line-check must be given the last line, it does not work on the whole matrix!");}
	GENER_COMMON(q,e,tr);
	junk = lev;		/* (level not used here) */
	
		/* so far, we only check/set the unit-scale for the matrix line */
	r = elimchm_unitlastline(set,e,tr)>=0;
	
	DEBUG(CURDLEV+2*r,"Line-check (set %d) for q %p (e %p, tr %d) found %s\n",set,q,e,tr,r?"pass":"NO");
	TIMEELAPSED();
	return r? 1:-1;
}




/**
 * This function checks the matrix sequence-properties - i.e. properties dependent on the
 * particular elimination sequence and the added lines, but invariant on line scaling and
 * the base minor automorphisms.
 * (The major difference from genstep_linecheck_ext() is in allowing base-minor automoprhisms.)
 * A sequence-connectivity check should be carried out here.
 * 
 * The check is carried out on the last line (row if tr==0, column if tr==1) of the
 * matrix e (which is supposed to fit as the resulting matrix of the sequence q).
 * This means that to perform a full check of the whole sequence, one must call this function
 * repeatedly along the elimination sequence!
 * If e is given, then the resulting matrix entries in q are never accessed, only the shape.
 * These checks must not assume that the matrix is properly represented over the pfield for lev>1.
 * 
 * Do not try to put here any long tests or debug outputs, since this function is called very often
 * (especially during canonical checking).
 * It returns 1 if the matrix passed the test, and -1 otherwise.
**/

int	genstep_sequencecheck_ext(int lev, elimseqm *q, ematrix *e, int tr) {
	int	i,tm, r = 1;
	
	if (tr<0 && lev>-3)  {PROGERROR("Sequence check must be given the last line, it does not work on the whole matrix!");}
	GENER_COMMON(q,e,tr);		/* (calls at lev==-3 from test_elimseqm() may not set tr) */
	
		/* whether the last line is added so that no parallel/series pair occurs -> 3-connectivity */
	if (r && (lev==3 || lev<=0) && tr>=0) {
		r = elimchm_3connlastline(e,tr)>=0;
	}
	
		/* and whether there are no fans exceeding limit ESNOFAN(q) at this step of the sequence */
	if (r && (lev<=2 || lev<=0))  if (ESNOFAN(q)>0) {
		if (lev==2 && tr>=0) {		/* (may not yet be representable here) */
			r = struct_nofanlastline_nofull(e,tr,ESNOFAN(q))>=0;
		} else if (lev==1 && tr>=0) {	/* (full check of fans ending with the last line) */
			r = struct_nofanlastline(e,tr,ESNOFAN(q))>=0;
		}
		if (lev<=0) {			/* (checks all fans, not only the last line) */
			r = struct_nofanall(e,ESNOFAN(q))>=0;
		}
	}
		/* and whether there are no tails exceeding limit ESNOTAIL(q) at this step of the sequence */
	if (r && lev<=1 && lev>-3)  if (ESNOTAIL(q)>0) {
		tm = 0;  i = ESNOFAN(q)-ESNOTAIL(q);
		if (i>0)  tm = TMOD_FANONLY(-i);
		//********** handle tm in general.......
		if (lev==1 && tr>=0) {		/* (a full last-line test, needs representability!) */
			r = struct_notail3lastline(e,tr,ESNOTAIL(q),tm)>=0;
		}
		if (lev<=0) {
			r = struct_notail3all(e,ESNOTAIL(q),tm)>=0;
		}
		//************ distinguish different types of tails (related to fans ???)
	}
	
		/* finally, here we may call user-defined extra sequence checks: */
	if (r)  r = genstep_sequencecheck_incl(lev,q,e,tr)>0;
	
	DEBUG(CURDLEV+2*r,"%*sSequence-check at lev %d for q %p (e %p, tr %d) found %s\n",lev,"",lev,q,e,tr,r?"pass":"NO");
	TIMEELAPSED();
	return r? 1:-1;
}




/**
 * This function checks the elimination sequence signature nsq against the shape of q.
 * The check has absolutely nothing(!) to do with entries of the resulting matrix in q
 * (not even with the basis), only the matrix shape and other sequence attributes are accessed.
 * 
 * It returns 1 if the signature nsq passed the test, and -1 otherwise.
**/

int	genstep_seqsigncheck_ext(int lev, elimseqm *q, eseq_t nsq) {
	int	r = 1;
	
	junk = lev;		/* (level not used here) */
	
		/* we look whether the signature nsq has the right number of extensions and coextensions */
	r = elimchm_signbits_poss(q,nsq)>=0;
	
	DEBUG(CURDLEV+1+r,"Signature-check for q %p and nsq %lX found %s\n",q,nsq,r?"pass":"NO");
	return r? 1:-1;
}


/**
 * This function checks the matrix base-properties - i.e. properties dependent on the
 * chosen basis of the matrix e as if it was the resulting matrix of q.
 * The checks are invariant on the sequnce or on the order of lines in the matrix,
 * but they may depend on the sequence signature in q.
 * Nothing is implemented here so far, and if something will be implemented later, then
 * that would probably require to change extension-generating to consider all bases(!)...
 * 
 * Returns 1 if the matrix passed the test, and -1 otherwise.
**/

int	genstep_basischeck_ext(int lev, elimseqm *q, ematrix *e) {
	int	trx=0, r = 1;
	
	junk = lev;		/* (level not used here) */
	if (e)  trx = (ISTRANSPM(e)!=ESELIMSG(q)%2);
	GENER_COMMON(q,e,trx);
	
		/* nothign is tested here so far ... (what about representability here???) */
	
	DEBUG(CURDLEV+2*r,"Basis-check at lev %d for q %p (e %p) found %s\n",lev,q,e,r?"pass":"NO");
	TIMEELAPSED();
	return r? 1:-1;
}




/**
 * This function checks the structral properties of the matrix - i.e. properties not dependent
 * on matrix equivalence (pivoting, permuting, scaling) or the elimination sequence.
 * The checks must include full testing the matrix e in the partial field(!) at levels lev<=2.
 * Other possible tests include forbidden minors and similar stuff...
 * 
 * The matrix e is supposed to fit as the resulting matrix of the sequence q.
 * The check may use an information about the last added line (in tr) to the matrix, but it
 * should NOT entirely depend on such information - remember it is sequence-independent.
 * (For example, it is OK to use tr in a fast incomplete check at lev>0, but not at lev==0...)
 * The checks for lev>1 do not assume that the matrix is properly represented over the pfield.
 * For lev>0, we use randomized structure checks to speed-up rejection.
 * (Specially, lev==-3 is used to request a test for test_elimseqm_ext() - somehow restricted to be faster.)
 * 
 * It returns 1 if the matrix passed the test, and -1 otherwise.
**/

int	genstep_structurecheck_ext(int lev, elimseqm *q, ematrix *e, int tr) {
	int	tri, cht = -1, r = 1;
	
	if ((lev<=0 && tr>=0 && tr<=1) || (lev>0 && tr<0))  {PROGERROR("Wrong use of tr=%d at level %d",tr,lev);}
	GENER_COMMON(q,e,tr);
#ifndef FASTPROG
	if (CURDLEV-2<=DEBUGLEV)  cht = 0;	/* >0 do extra internal checks in the following tests... */
#endif
	
		/* depending on lev, we look whether the matrix e is in our pfield */
	if (r && lev==3 && tr>=0)
		r = elimchm_represented_randlast(q,e,tr)>=0;
	if (r && lev==2 && tr>=0)
		r = elimchm_represented_last(q,e,tr)>=0;
	if (r && lev<=0 && lev>-3)	/* (maybe not necessary, but we formally require a FULL matrix test for lev==0) */
		r = elimchm_represented(q,e)>=0;
	if (!r)  DEBUG(CURDLEV,"     - structure: NOT in pfield found.\n");
	if (!r && lev<=1)  {PROGERROR("We should find not-represented already at level >1 !");}
	TIMEELAPSED();
	
		/* for lev==3, we look whether the matrix e is a tight major against displayed minors from ESTIGHT */
	if (r && lev==3 && tr>=0)  if (ESTIGHT(q)) {
		r = genstep_pretest_tight(q,e,tr)>=0;	/* (no representability required here) */
	}
	
	/* further checks assume properly represented matrix e over the pfield ! */
	tri = ISTRANSPM(e);
	ematrix_transpose_set(e,0);	/* (we must look at the 0-transp state) */
	
		/* depending on lev, we look whether the matrix e has a forbidden minor from ESFORBID */
	if (r && lev<=1 && lev>-3)
	  if (ESFORBID(q))  if (lev<=0 || (GEN_RANDSTRUCTCH>0 && ROWSM(e)+COLSM(e)>=GEN_RANDSTRUCTCH)) {
		if (lev>0) {
			r = !struct_hasminorlist_rand(e,ESFORBID(q));	/* (only fast randomized test) */
		} else {
			r = !struct_hasminorlist_ch(cht,e,ESFORBID(q));
		}
	}

		/* depending on lev, we look whether the matrix e is a tight major for the list ESTIGHT */
	if (r && lev<=1 && lev>-3)
	  if (ESTIGHT(q))  if (lev<=0 || (GEN_RANDSTRUCTCH>0 && ROWSM(e)+COLSM(e)>=GEN_RANDSTRUCTCH)) {
		if (lev>0) {
			r = !struct_hasdelcontr_rand(e,ESTIGHT(q));	/* (only fast randomized test) */
		} else {
			r = !struct_hasdelcontr_ch(cht,e,ESTIGHT(q));
		}
	}
	ematrix_transpose_set(e,tri);
	
		/* finally, here we may call user-defined extra structure checks: */
	if (r)  r = genstep_structurecheck_incl(lev,q,e,tr)>0;
	
	DEBUG(CURDLEV+2*r,"%*sStructure-check at lev %d for q %p (e %p, tr %d) found %s\n",lev,"",lev,q,e,tr,r?"pass":"NO");
	TIMEELAPSED();
	return r? 1:-1;
}






/******************	Pre-tests for generated sequences	*******************/
/**********************************************************************************/


#undef CURDLEV
#define	CURDLEV		6


	/* the precomputed data are here - see genstep_pretest_start() for description: */
		
elimseqm	*precheck_seq = NULL;	/* mark the sequence - to check correct data */
ematrix		*precheck_ecopy = NULL;	/* a copy of the matrix before extension, transposed so that a column is to be added */

static long	*disptight;	/* the row patterns of displayed "tight" minors in the matrix to extend */
static int	disptlen = 0;

static emxaut	**lincanonaut = NULL;	/* automorphisms of the whole matrix without the added line */
static int	nolineauts = 0;		/* if there is no non-identical automorphism */



/**
 * This function is called to precompute some data structures for the extension generating process.
 * See precheck_seq and related variables above.
 * The data should not be essential to other computing, just use them to speedup some checks.
 * The function is called from gener_extensions_ext().
 * For wh>0, it prepares the data for the sequence q; and for wh<0, it clears the data.
 * It returns 1 for success or -1 on failure (but this is not much useful anyway).
 * 
 * The matrix in q is given already extended by the new line, so one must forget that line.
 * This function expects the added line be given as the last column of ESMATRIX(q) in the current
 * transpose state (otherwise an error is reported).
 * 
 * Row-patterns are precomputed for displayed "tight" minors in the matrix - patterns have
 * 1-bits at rows that support the displayed minors.
 * These patterns are used in genstep_structurecheck_ext() to see whether the extended matrix
 * is "displayed" as not a tight major.
 * 
 * Automorphisms for the matrix before extension are precomputed for use in gcanon_lineauts()
 * - a quick test that just permutes the rows and compares the last column for canonical.
 * 
 * See more in the functions...
**/

int	genstep_pretest_ext(elimseqm *q, int tr, int wh) {
	int		i,j,l, r = 1;
	ematrix		**ex[100],**x;
	
	if ((precheck_seq==NULL)!=(wh>0) || (precheck_seq==q)!=(wh<=0)) {
		PROGERROR("Wrong precheck call for q=%p, wh=%d",q,wh);  return -1;
	}
		/* calling possible third-side pre-test function for clearing */
	if (wh<=0)  genstep_pretest_incl(q,tr,wh);
	
		/* clearing previously precomputed data here */
	if (wh<=0) {
		if (disptlen>0) { FREE(disptight);  disptight = NULL; }
		disptlen = 0;
		if (lincanonaut)  alist_free(lincanonaut);	/* (better no recycling here...) */
		lincanonaut = NULL;
		if (precheck_ecopy)  dispose_ematrix(precheck_ecopy);
		precheck_ecopy = NULL;
		precheck_seq = NULL;
		DEBUG(CURDLEV-1,"Pre-computed data for seq %p cleared.\n",q);
		return 1;
	}
	
	if (!tr)  ematrix_transpose(ESMATRIX(q));
	if ((!ISTRANSPM(ESMATRIX(q)))!=(ESELIMSG(q)%2)) {	/* expect a column to be added! */
		PROGERROR("Wrong precheck call for matrix transpose");  return -1;
	}
		/* precomputing data here... */
	DEBUG(CURDLEV-1,"Pre-computing data for generating an extension of seq %p (%d x %d+1).\n",q,ROWSM(ESMATRIX(q)),COLSM(ESMATRIX(q))-1);
	precheck_seq = q;
	precheck_ecopy = ematrix_copy(ESMATRIX(q));
	ematrix_remove_col(precheck_ecopy,COLSM(precheck_ecopy)-1);
	lincanonaut = NULL;
	disptlen = 0;
	
		/* row-patterns for displayed "tight" minors in the matrix precheck_ecopy */
	if (ESTIGHT(q) && ROWSM(ESMATRIX(q))<NUM_LONG_BITS-4) {
		l = 0;
		DEBUG(CURDLEV,"Pre-computing row patterns for displayed \"tight\" minors ...\n");
		for (i=l=0; ESTIGHT(q)[i] && i<100; i++) {
			ex[i] = struct_dispminor(precheck_ecopy,ESTIGHT(q)[i]);
			l += alist_getlength(ex[i]);	/* (gets the total number of displayed minors in l) */
		}
		disptight = MMALLOC((l+2)*sizeof(disptight[0]));
		disptlen = l;  l = 0;
		for (i=0; ESTIGHT(q)[i] && i<100; i++)  if (ex[i]) {
			for (x=ex[i]; *x; x++) { /* computes the row pattern for each of the minors */
				disptight[l] = 0;
				for (j=0; j<ROWSM(*x); j++)
					disptight[l] |= (1l<<GETREFMROW(*x,j));
				l++;
				DEBUG(CURDLEV," - \"tight\" minor #%d (%d): row pattern  %lX\n",i,disptlen,disptight[l-1]);
			}
			dispose_alist_mats(ex[i]);	/* free all the minors */
		}
	}
		/* matrix automorphisms for fast canonical check in gcanon_lineauts() */
	if (1) {
		DEBUG(CURDLEV,"Pre-computing automorphisms for fast canonical pre-check ...\n");
		lincanonaut = struct_automorphisms(precheck_ecopy);
		nolineauts = ((i=alist_getlength(lincanonaut))<=1);	/* (only sense when there is a non-identical aut) */
		DEBUG(CURDLEV," - %d automorphisms found\n",i);
	}
	
		/* calling possible third-side pre-test function for starting */
	genstep_pretest_incl(q,tr,wh);
	
	if (!tr)  ematrix_transpose(ESMATRIX(q));
	return r? 1:-1;		/* (not much sense in the return value anyway) */
}


/**
 * This is a fast tight-major pre-test, based on data precomputed above.
 * For each occurence of a displayed minor from the "tight" list in the original matrix,
 * we remembered the row pattern of the minor.
 * If we now find that the non-zero pattern of the added column is disjoint from some
 * of these row patterns, then we know that the extended matrix is not a tight major;
 * any nonzero may be pivoted not changing the displayed minor from the "tight" list.
**/

int	genstep_pretest_tight(elimseqm *q, ematrix *e, int tr) {
	long	lzp;
	int	i,j, r=1;
	
	if (q==precheck_seq && disptlen>0) {
		if (tr<0)  {PROGERROR("Wrong use of tr=%d at level>0",tr);}
		if (!tr)  ematrix_transpose(e);
#ifndef FASTPROG
		if (IFRANDDEBUGLESS(72))  if (elimchm_compare_col(e,0,precheck_ecopy,0)!=0)
			{PROGERROR("the tested matrix is not the same as the precomputed one!");}
#endif
		for (i=lzp=0; i<ROWSM(e); i++)	/* nonzeros in the added line - in column now */
			if (SIGNM(e,i,COLSM(e)-1)!=0)  lzp |= (1l<<i);
		for (j=0; j<disptlen; j++)	/* test against rows of "tight" minors - must match some nonzero to be tight */
			if ((disptight[j]&lzp)==0)  r = 0;
		DEBUG(CURDLEV+2*r,"     - fast tight-major test for %p found %s\n",e,r?"pass":"NOT");
		if (!tr)  ematrix_transpose(e);
	}
	return r?1:-1;
}




















/******************	Checking canonical forms	*******************/
/**************************************************************************/



/**
 * The following functions put together all tests that must be performed at each (co)extension
 * step when reconstructing an elimination sequence in order to refuse repeated occurences of
 * an equivalence class of matrices starting with the (identically) same minor.
 * Briefly speaking, an extension sequence must be "minimal" by the canonical ordering among
 * all elimination sequences leading to equivalent (to e) matrices, to pass this test.
 * Read more about the related theory in the file ../include/gener.h ...
 * (Remember that the sequence signatures are compared first, and only then the matrix lines
 * in their natural order and possibly other resulting-matrix properties.)
 * 
 * It is vital that the functions here use exactly the same genstep checks to test admissibility
 * of a matrix as the generating process in gener.c - otherwise some extensions may be lost(!).
 * Besides those, we use the elimchm_signbitscomp() and elimchm_seqcompare() tests for finding
 * which sequence (matrix) is smaller, to reject sequences that are not canonically-minimal.
 * 
 * General scheme of the canonical testing follows:
 *  - cycle all admissible signatures that are not larger than that one of current sequence;
 *    - cycle all bases of the current matrix, check their admissibility;
 *      - cycle all displayed base-minors for the chosen basis (auts factored out);
 *        - cycle all orders of the added lines with respect to the chosen signature,
 *          check admissibility of the constructed sequence (no line-scales yet);
 *          - cycle base-minor automorphisms, apply them to the constructed matrix,
 *            set the displayed minor identically,
 *            set unit-scales for all added lines in the sequence,
 *            compare the constructed sequence with the current one.
 * To speed-up the computation, we add a pre-comparing check for subsequences during cycling
 * all orders of added lines if SUBSEQCOMP>0 is defined.
 * (For this to work, the ordering of sequences must be hereditary on subsequences!)
 * We also implement randomized pre-check if RANDCANON>0 - using just random subselection of bases.
 * 
 * No third-side additions are possible to the canonical testing, as it would not have much sense
 * and might introduce serious errors...
 * 
**/


#undef CURDLEV
#define	CURDLEV		6


static int	timeelc,tielc;
#define	GENER_COMMON_C(q,e,tr)	timeelc = TIME();  GENER_COMMON(q,e,tr)
#define	TIMEELAPSED_C()		if ((tielc=TIME()-timeelc)>2) DEBUG(CURDLEV-1,"\t  - time canonical elapsed %d.\n",tielc)



/**
 * This function checks whether the matrix e is minimal by the canonical ordering among
 * all equivalent (to e) matrices and all their sequences.
 * The matrix e is supposed to fit as the resulting matrix of the sequence q.
 * (However, if lev>0, then the matrix e may not yet be fully valid generated matrix;
 * and if lev>1, then e may not even be properly represented in the pfield.)
 * 
 * The given value of lev determines the level of tests - 0 means to do all tests thoroughly,
 * while values >0 (1,3 for now) request faster but incomplete tests - to speed up rejection.
 * The value lev==0 must always perform a full test, even when this means a repetition of
 * some previous faster tests.
 * The value of tr determines what was the last added line (0 row/column 1).
 * This may be used for fast preliminary tests, but the full canonical test lev==0 should be
 * caried out regardless of tr, taking the whole resulting matrix into an account.
 * 
 * Returns 1 if the matrix passed the test, and -1 otherwise.
**/

int	genstep_canonicalcheck_ext(int lev, elimseqm *q, ematrix *e, int tr) {
	int	r = 1;
	
	if ((lev<=0 && tr>=0 && tr<=1) || (lev>0 && tr<0))  {PROGERROR("Wrong use of tr=%d at level %d",tr,lev);}
	GENER_COMMON_C(q,e,tr);
	
		/* for lev==3 (highest), we look whether the last line is minimal for automorphisms */
	if (r && lev==3 && !nolineauts) {	/* (nolineauts pre-computed, otherwise 0) */
		r = gcanon_lineauts(q,e,tr)>=0;
	}
	
	/* further checks assume properly represented matrix e over the pfield ! */
	
		/* for lev==1, we check randomized equivalent sequences and matrices */
	if (r && lev==1 && GEN_RANDCANON>0 && ROWSM(e)+COLSM(e)>=GEN_RANDCANON) {
		r = gcanon_allequivm(q,e,1)>=0;
	}
	
		/* for lev==0, we check all equivalent sequences and matrices - a full test */
	if (r && lev<=0) {
		r = gcanon_allequivm(q,e,0)>=0;
	}
	DEBUG(CURDLEV-(lev<=1)+r,"%*sCanonical-check at lev %d for q %p (e %p, tr %d) found %s\n",lev,"",lev,q,e,tr,r?"pass":"NO");
	TIMEELAPSED_C();
	return r? 1:-1;
}





/**
 * This function performs a quick check of the added line against all automorphisms
 * of the previous part of the matrix e (these auts change nothing on the previous steps
 * of the sequence, so only the added line is considered, which is quite fast).
 * The check is carried out on the last line (row if tr==0, column if tr==1) of the matrix e.
 * (The matrix e is supposed to fit as the resulting matrix of the sequence q.)
 * It must not assume proper representability (in the pfield) of the matrix.
 * 
 * Returns 1 if the matrix passed the test, and -1 otherwise.
**/

int	gcanon_lineauts(elimseqm *q, ematrix *e, int tr) {
	int		i,lc, ru=1,r=1;
	emxaut		**aut, **a;
	ematrix		*ev;
	char		buf[100];
	elimseqm	*q2;	/* (used only for debug testing below) */
	
	if (tr<0)  {PROGERROR("Wrong use of tr=%d at level>0",tr);}
	if (!tr)  ematrix_transpose(e);
	ev = ematrix_copy(e);		/* uses "working copy" ev of the matrix e */
	lc = COLSM(ev)-1;  ematrix_remove_col(ev,lc);
	if (lincanonaut!=NULL) {
		aut = lincanonaut;	/* if the automorphisms are already precomputed, takes them */
	} else {
		aut = struct_automorphisms(ev);	/* computes scaled automorphisms of ev */
	}
#ifndef	FASTPROG
	if (IFRANDDEBUGLESS(222))  if (genstep_linecheck(q,e,1)<0)
		{PROGERROR("expect the added column to be unit-leading scaled!");}
	if (DEBUGLEV>=CURDLEV+3)  for (a=aut,i=0; *a; a++,i++) ;
	DEBUG(CURDLEV+3," - testing last line of the ext %p against %d auts\n",e,i);
	if (lincanonaut!=NULL && IFRANDDEBUGLESS(222))  if (!ematrix_isequal(ev,precheck_ecopy))
		{PROGERROR("the tested matrix is not the same as the precomputed one!");}
#endif
		/**
		 * We cycle through all automorphisms of ev without the added column,
		 * and we apply each of them to the added column (relocation and scaling).
		 * Then we look whether we get a "smaller" line after the automorphism,
		 * which would mean our matrix was not canonical, so it would be refused.
		**/
	ev = ematrix_append_col(ev);
	r = 1;
	for (a=aut; *a && r; a++) {
		for (i=0; i<ROWSM(ev); i++)
			COPYEXSIGM(ev,(*a)->pr[i],lc, e,i,lc);
		for (i=0; i<ROWSM(ev); i++)
			ematrix_multiply_entry(ev,i,lc, 1,(*a)->xr[i],(*a)->gr[i]);
			/* the following test is just a shortcut for the full test in elimchm_seqcompare() !!!*/
		ru = genstep_linecheck_set(q,ev,1)>=0;
		if (ru)  r = elimchm_compare_col(ev,lc,e,lc)>=0;
	}
	q2 = NULL;  buf[0] = 0;
#ifndef FASTPROG
	if (!r && ru && IFRANDDEBUGLESS(111)) {	/* for safety, we check whether the above shortcut test was correct */
		q2 = elimseqm_copyfor(q,ev);
		if (elimchm_seqcompare(q2,q)>=0)  {PROGERROR("Incorrect shortcut for sequence-compare test!");}
		FREE(q2);
		if (!struct_dispminor_ch(-3,ev,e))  {PROGERROR("The smaller matrix is not equivalent to the tested one!");}
	}
	if (!r && CURDLEV+1<=DEBUGLEV)  for (buf[0]=0,i=0; i<ROWSM(ev); i++) {
		pfield_printvalue_to(buf+strlen(buf),8,EXPM(ev,i,lc),SIGNM(ev,i,lc));
		strcat(buf,",");  if (strlen(buf)>60)  break; }
	if (!r)  DEBUG(CURDLEV+1,"    - smaller aut #%d found:  %s\n",(int)(a-aut),buf);
#endif
	if (lincanonaut==NULL && aut) {	/* frees all the automorphisms if not precomputed in lincanonaut */
		struct_automorphisms_recycle(ev,aut);  aut = NULL;
	}
	dispose_ematrix(ev);		/* frees the local matrix ev */
	if (!tr)  ematrix_transpose(e);
	return r? 1:-1;
}






/******************	Canonically-minimal sequence	*******************/
/**************************************************************************/


#undef CURDLEV
#define	CURDLEV		5		/* (OK with randomized debug tests if >=5) */



/**
 * This function performs a thorough check of the extended matrix e against all elimination
 * sequences of all equivalent (to e) resulting matrices, with respect to the (identically)
 * same base minor as in q.
 * More is contained in the general algorithm description at the top of this section.
 * The matrix e must be properly represented over the pfield for this function.
 * 
 * If ran>0, then only randomized pre-test is carried out - not complete(!!!).
 * (One may call ran>0 even when the matrix is not yet fully validated, so be careful...)
 * No attention is paid to the current transposition state or to the last added line
 * in the test - everything is tested against the 0-transp state of matrices.
 * It returns 1 if the matrix e passed the test, and -1 otherwise.
**/

static int	idfound = 0;		/* (used for debugging only) */


int	gcanon_allequivm(elimseqm *q, ematrix *e, int ran) {
	int		i,nb, al,bl, nlq=0,nbm=0, cht=-1, ret=1;
	ematrix		*ee,*esm, **bas,**b, **bmin,**m, ***bmlists;
	elimseqm	*qn;
	eseq_t		lq;
	emxaut		**aut;
	
	if (GEN_RANDCANON<=0 && ran>0)  {PROGERROR("Randomized test is not allowed now!");}
#ifndef FASTPROG
	if (CURDLEV-2<=DEBUGLEV)  cht = 1;	/* >1 do extra internal checks in the following tests... */
	if (ran>0 && cht>0)  cht = 0;	/* (0 cannot do extra checks when randomized - e is not yet fully valid!) */
	idfound = 0;		/* (used for debug - whether the identical sequence was tested) */
	DEBUG(CURDLEV+0,"Going to look for %s equivalent sequences (seq %p, mat %p) ...\n",ran?"random":"all",q,e);
	if (cht>=0)  ELIMQDEBUGS(CURDLEV+2,q,ran<=0?"\t\t|=\t":"\t\t|-\t");
	if (IFRANDDEBUGLESS(32))  test_elimseqm_notough(q);
#endif
	esm = ematrix_copy(ESMINOR(q));		/* (use a copy of ESMINOR(q) so that aut recycling works, no interference) */
	aut = struct_automorphisms(esm);	/* base-minor automorphisms - must be considered here */
	ee = ematrix_copy(e);
	qn = elimseqm_copyfor(q,ee);		/* local copies of the matrix and the sequence */
	ematrix_transpose_set(ee,0);		/* (we use 0-transposition here, also for bases) */
	if (ran>0)  bas = struct_allbases_rand(ee);	/* randomized basis-list for ran>0 */
	else  bas = struct_allbases_ch(DECRCH(cht),ee);
	nb = alist_getlength(bas);		/* space to store dispminor lists later */
	bmlists = MMALLOC((nb+2)*sizeof(bmlists[0]));
	for (i=0; i<=nb; i++)  bmlists[i] = NULL;
	b = bas;
			/**
			 * Here we cycle through all smaller or equal sequence signatures,
			 * through all admissible bases and all displayed base-minors.
			 * Then we call gcanon_allequivm_onedisp() to perform minimality
			 * test for each one displayed minor.
			 * (Recall that base-minor automorphisms are factored out for dispminors.)
			**/
	ret = 1;
	for (lq=ESMAXLSG(q)-1; lq>=0 && ret; lq--) {
		if (genstep_seqsigncheck(q,lq)<0)  continue;
		if (elimchm_signbitscomp(q,lq)<0)  continue;
		DEBUG(CURDLEV+1,"~~ considering seq signature %lX (length %d)...\n",lq,ESLENGTH(qn));
		nlq++;
		for (b=bas, i=0; *b && ret; b++, i++) {	/* cycling all bases and their displayed minors */
			qn = elimseqm_setfor(qn,*b,lq);
			if (genstep_basischeck(qn,*b)<0)  continue;
			if (bmlists[i]==NULL) {
				bmin = bmlists[i] = struct_dispminor_ch(DECRCH(cht),*b,ESMINOR(q));
			} else  bmin = bmlists[i];	/* (already computed list) */
			if (bmin) {
				for (m=bmin; *m && ret; m++)
					ret = gcanon_allequivm_onedisp(cht,q,qn,*b,*m,aut)>=0;
				nbm += m-bmin;
			}
		}
		//TIMEELAPSED_C();
	}
			/* final cleaning after the decision was made above... */
	al = alist_getlength(aut);  bl = alist_getlength(bas);
	struct_automorphisms_recycle(esm,aut);
	dispose_ematrix(esm);
	FREE(qn);  dispose_ematrix(ee);
	for (i=0; i<nb; i++)  if (bmlists[i]!=NULL)
		dispose_alist_mats(bmlists[i]);
	FREE(bmlists);
	if (ran==0) {	/* when dealing with full list of bases, we may "recycle" it for further use */
		struct_allbases_recycle(bas);
	} else {
		dispose_alist_mats(bas);
	}
#ifndef FASTPROG
	if (ret && idfound==0 && ran<=0)  {PROGERROR("Why the identical sequence was not checked ???");};
	DEBUG(CURDLEV-1+2*ret,"Testing seq %p for %s canonical found %s\n",q,ran?"random":"all",ret?"pass":"NO");
	DEBUG(CURDLEV+ret,"\t (- used %d smaller sigs, %d auts, %d bases, %d dispmins)\n",nlq,al,(int)(b-bas),nbm);
#endif
	return ret? 1:-1;
}



/**
 * This function examines all equivalent elim sequences that use the basis bs with the 
 * displayed base-minor mn.
 * Given is qo the original sequence, qn a new sequence with bs as the resulting matrix,
 * bs the resulting matrix of qn equivalent to the matrix of qo, mn is a matrix refering
 * to the base-minor in bs (in correct order, but not scaled), aut the list of automorphisms
 * of the base-minor.
 * 
 * It returns 1 if the matrix passed the test, and -1 otherwise (i.e. smaller seq found).
**/

int	gcanon_allequivm_onedisp(int cht, elimseqm *qo, elimseqm *qn,
					ematrix *bs, ematrix *mn, emxaut **aut) {
	ematrix		*rem, *esc;
	int		i,j,k, br, qnl, eqsg, ret=1;
	int		xc[MAXELIMLENGTH], sel[MAXELIMLENGTH];
	elimseqm	*qq, *qqs;
	emxaut		**a;
	
	if (!ISREFMAT(mn) || REFMAT(mn)!=bs || ROWSM(bs)!=ROWSM(ESMATRIX(qn)) || COLSM(bs)!=COLSM(ESMATRIX(qn)))
		{PROGERROR("Call only for a refering matrix  mn -> bs (bs ~ q-matrix) !");}
	if (ISTRANSPM(bs))  {PROGERROR("Call only for a 0-transp matrix bs !");}
	if (ESMINOR(qo)!=ESMINOR(qn))  {PROGERROR("Call only for the same base minor !");}
	if (ESLENGTH(qn)>MAXELIMLENGTH)  {PROGERROREXIT("cannot work with elim seq longer than %d > %d",ESLENGTH(qn),MAXELIMLENGTH);}
#ifndef	FASTPROG
	DEBUG(CURDLEV+2," # testing one basis %p for canonical, base minor %p\n",bs,mn);
	if (elimchm_signbitscomp(qo,ESELIMSG(qn))<0)  {PROGERROR("Do not call for a larger qn-signature (than qo) !");}
	if (cht>1) { EMATDEBUGS(CURDLEV+3,bs,"\t\t\t");  EMATDEBUGS(CURDLEV+3,mn,"\t\t\t>\t"); }
	if (cht>=0 && IFRANDDEBUGLESS(111))  if (!struct_isscaledequal(ESMINOR(qn),mn))
			{PROGERROR("The matrix mn is not (scaled-)equal to the minor in qn!");}
#endif	
		/**
		 * Here we consider all orders of lines of bs added to the base-minor mn.
		 * For that we use a refering matrix rem that starts from mn and grows to
		 * the whole bs.
		 * With each added line, we check for the sequence test.
		 * After all, we compare the newly created sequence with the original one.
		 * (We also pre-compare subsequences if they are too long to save time.)
		 * More is contained in the general algorithm description at the top of this section.
		**/
	qnl = ESLENGTH(qn);
	eqsg = (elimchm_signbitscomp(qo,ESELIMSG(qn))==0);
	qq = elimseqm_copyfor(qn,bs);	/* (used in finding the order of added lines) */
	qqs = elimseqm_copyfor(qn,bs);	/* (used when calling gcanon_allequivm_scal) */
	for (i=0; i<qnl; i++)  xc[qnl-1-i] = (ESELIMSG(qn)&(1l<<i))?1:0;
	rem = ematrix_refer_all(mn);
	ret = 1;
	k = 0;  sel[0] = -1;
	while (k>=0 && ret) {	/* cycle through all orders of added lines to the minor mn */
		j = ROWSM(rem)+COLSM(rem)-(ROWSM(mn)+COLSM(mn));
		while (--j>=k) {		/* (remove previously added extra lines...) */
			if (xc[j]==0)  ematrix_remove_row(rem,ROWSM(rem)-1);
			else  ematrix_remove_col(rem,COLSM(rem)-1);
		}
		if (++sel[k]>=(xc[k]?COLSM(bs):ROWSM(bs))) {
			--k;  continue;		/* (no more choice at this level) */
		}
		br = 0;		/* look whether the line sel[k] is already used in rem */
		if (xc[k]==0) {
			for (i=0; i<ROWSM(rem) && !br; i++)
				if (sel[k]==GETREFMROW(rem,i))  br = 1;
		} else {
			for (i=0; i<COLSM(rem) && !br; i++)
				if (sel[k]==GETREFMCOL(rem,i))  br = 1;
		}
		if (br)  continue;
				/* add the line sel[k] (indexed in the dest matrix!), and try it in the sequence check */
		if (xc[k]==0)  ematrix_drefadd_row(rem,sel[k]);
		else  ematrix_drefadd_col(rem,sel[k]);
		qq = elimseqm_setfor(qq,rem,(ESELIMSG(qn)>>(qnl-k-1)));
		printlev -= 1;	/* (to suppress too many debug messages from the sequence check here) */
		br = genstep_sequencecheck(qq,rem,xc[k])<0;
		printlev += 1;
		if (br)  continue;
		DEBUG(CURDLEV+3,"%*s . passed subsequence %d, last %s %d\n",2*k,"",k,xc[k]?"col":"row",sel[k]);
		if (ESLENGTH(qq)!=k+1)  {PROGERROR("Wrong qq-subsequence length %d!=%d.",ESLENGTH(qq),k+1);}
		
				/* pre-comparing sequences for the same signature (see below for description) */
		if (GEN_SUBSEQCOMP>0 && eqsg)  if (k<qnl-2) {	/* (use qnl-2 or qnl-3 ???) */
			DEBUG(CURDLEV+3,"   subsequence pre-compare for k=%d, qnl=%d\n",k,qnl);
			for (br=1, a=aut; *a && br; a++) {
				qqs = elimseqm_setfor(qqs,rem,ESELIMSG(qq));
				esc = gcanon_allequivm_scal(cht,rem,*a,xc,qqs);
				if (esc==NULL)  continue;
				qqs = elimseqm_setfor(qqs,esc,ESELIMSG(qq));
#ifndef FASTPROG
				if (cht>0 && IFRANDDEBUGLESS(444))	/* (some extra tests here...) */
					test_elimseqm_notough(qqs);
#endif
				br = elimchm_seqcompare_part(DECRCH(cht),qo,qqs,k+1)<0;
								/* (only a subsequence is compared) */
				dispose_ematrix(esc);  elimseqm_setfor(qqs,NULL,0);
			}	/* (only if some subsequence is <= the original subsequence, then there is a chance of refusing below) */
		}
		if (br)  DEBUG(CURDLEV+2,"  - no smaller sequence possible by a pre-compare at k=%d\n",k);
		if (br)  continue;
		
		if (k<qnl-1) {	/* go to the next level with an initial choice of line */
			sel[++k] = -1;  continue;
		}
			/**
			 * Above we have found lines sel[k] to add at level k, for all levels.
			 * Now the matrix rem refers the whole bs in chosen order.
			 * We call gcanon_allequivm_scal() to apply each of the automorphisms
			 * to rem, and to produce a properly scaled copy esc of it.
			 * (Properly scaled means that the base-minor is identically displayed
			 * and the added lines are unit-scaled.)
			 * Moreover, if the new sequence has the same signature as the original
			 * one, we compare the matrices lexicographically (see pre-comparing above).
			 * Finally, if the new sequence is smaller, then we refuse the original one.
			**/
		for (a=aut; *a && ret; a++) {
			qqs = elimseqm_setfor(qqs,rem,ESELIMSG(qn));
			esc = gcanon_allequivm_scal(cht,rem,*a,xc,qqs);
			if (esc==NULL)  continue;
			qqs = elimseqm_setfor(qqs,esc,ESELIMSG(qn));
			if (eqsg) {		/* for equeal signatures, we must compare matrices */
				br = elimchm_seqcompare_ch(DECRCH(cht),qo,qqs);
			} else  br = 1;
			if (br>0)  ret = 0;	/* if the new seq is found smaller, then reject */
#ifndef FASTPROG
			if (br==0)  idfound = 1;
			if (!ret) {
				DEBUG(CURDLEV-1,"  - a smaller sequence found for sig %lX (~%lX) and mat %p\n",ESELIMSG(qqs),ESELIMSG(qo),esc);
				if (cht>=0) { EMATDEBUGS(CURDLEV,esc,"\t\t^^\t");  EMATDEBUGS(CURDLEV+3,ESMATRIX(qo),"\t\t\to\t"); }
				if (cht>0 && IFRANDDEBUGLESS(cht>1?55:222))	/* some extra checks... */
					test_elimseqm(qqs);
				else  if (IFRANDDEBUGLESS(36))
					test_elimseqm_notough(qqs);	/* (matrix may not be yet fully valid) */
				ematrix_transpose_set(esc,ISTRANSPM(ESMATRIX(qo)));
				if (cht>0 && IFRANDDEBUGLESS(cht>1?22:444))
				  if (!struct_isequivalent(esc,ESMATRIX(qo)))
					{PROGERROR("The smaller matrix is not equivalent to the tested one!");}
			} else if (IFRANDDEBUGLESS(cht>1?88:1111)) {
				if (cht>0)  test_elimseqm_rand(qqs);	/* some extra checks even here... */
				else  test_elimseqm_notough(qqs);
			}
#endif
			dispose_ematrix(esc);  elimseqm_setfor(qqs,NULL,0);
		}
	}
	DEBUG(CURDLEV+2*ret,"> testing one basis %p for canonical found %s\n",bs,ret?"pass":"NO");
	dispose_ematrix(rem);
	FREE(qq);  FREE(qqs);
	return ret? 1:-1;
}


/**
 * This function produces a scale of the given matrix e after the automorphism au is applied,
 * so that the base-minor of qn is identically displayed in the scale and the added lines
 * are all unit-scaled.
 * xc[] gives the sequence signature separated each bit into one array element from highest bit.
 * 
 * Returns the properly scaled (+unit) matrix, or NULL if not possible to scale.
**/

ematrix*	gcanon_allequivm_scal(int cht, ematrix *e, emxaut *au, int xc[], elimseqm *qn) {
	int		i,j,k, r;
	ematrix		*emin, *es, *res;
	emxaut		*sc;
	elimseqm	*qq;
	
	emin = ESMINOR(qn);
	es = ematrix_copy(e);		/* copying the (permuted by au) entries of e to es */ 
	qq = elimseqm_copyfor(qn,es);	/* (needs a temp seq for scaling along the sequence) */
	for (i=0; i<ROWSM(es); i++)  for (j=0; j<COLSM(es); j++) {
		COPYEXSIGM(es,i<ROWSM(emin)?au->pr[i]:i,j<COLSM(emin)?au->pc[j]:j, e,i,j);
		if (i>=ROWSM(emin) && j>=COLSM(emin))  break;
	}
#ifndef FASTPROG			/* (we need to move line id's only for debug printing purposes) */
	if (CURDLEV+3<=DEBUGLEV) {
		for (i=0; i<ROWSM(emin); i++)  ROWSID(es,au->pr[i]) = ROWSID(e,i);
		for (j=0; j<COLSM(emin); j++)  COLSID(es,au->pc[j]) = COLSID(e,j);
		DEBUG(CURDLEV+3,"  calling scal for e=%p, perm ",e);
		for (i=0; i<ROWSM(emin); i++)  SDEBUG(CURDLEV+3,"r%d ",au->pr[i]);
		for (j=0; j<COLSM(emin); j++)  SDEBUG(CURDLEV+3,"c%d ",au->pc[j]);
		SDEBUG(CURDLEV+3,"\n");
		if (cht>1) { EMATDEBUGS(CURDLEV+4,e,"\t\t\t\te\t");  EMATDEBUGS(CURDLEV+4,es,"\t\t\t\tes\t"); }
		for (i=0; i<ROWSM(emin); i++)  if (au->pr[i]>=ROWSM(emin) || au->pr[i]<0)
			{PROGERROR("Wrong permutation given at  pr[%d] -> %d",i,au->pr[i]);}
		for (j=0; j<COLSM(emin); j++)  if (au->pc[j]>=COLSM(emin) || au->pc[j]<0)
			{PROGERROR("Wrong permutation given at  pc[%d] -> %d",j,au->pc[j]);}
	}
#endif
	res = ematrix_refer(es,0,ROWSM(emin),0,COLSM(emin));
	sc = new_emxaut(ROWSM(emin)+2,COLSM(emin)+2);
	r = struct_isscaledequal_scal(DECRCH(cht),res,emin,sc);	/* gets the necessary scale to sc */
	if (r) {
		for (i=0; i<ROWSM(emin); i++)	/* scale to get the minor identically displayed */
			ematrix_multiply_row(es,i, 1,sc->xr[i],sc->gr[i]);
		for (j=0; j<COLSM(emin); j++)
			ematrix_multiply_col(es,j, 1,sc->xc[j],sc->gc[j]);
#ifndef FASTPROG
		if (cht>=0 && IFRANDDEBUGLESS(44))  if (!ematrix_isequal(res,emin))  {PROGERROR("Not found the base minor, why???");}
#endif
		for (k=0, i=ROWSM(emin), j=COLSM(emin); r && (i<ROWSM(es) || j<COLSM(es)); k++) {
			if (xc[k]==0)  ematrix_refadd_row(res,i++);
			else  ematrix_refadd_col(res,j++);	/* scale to get unit-leading lines */
			qq = elimseqm_setfor(qq,res, (ESELIMSG(qn)>>(ESLENGTH(qn)-k-1)));
				/* (local qq copy is not really needed, but otherwise debug checks fail in genstep_linecheck_setref()) */
			r = genstep_linecheck_setref(qq,res,xc[k])>=0;
				/* (it should not happen that setref fails, but the test is here for safety) */
#ifndef FASTPROG
			DEBUG(CURDLEV+4,"   - setref i=%d, j=%d, k=%d, r=%d\n",i,j,k,r);
			if (cht>1)  EMATDEBUG(CURDLEV+4+(k<ESLENGTH(qn)-1),res,"\t\t\t\tres\t");
#endif
		}
		if (i>ROWSM(es) || j>COLSM(es) || k>ESLENGTH(qn))  {PROGERROR("Whole matrix was not unit-scaled, why??? i=%d>%d, j=%d>%d, k=%d>%d",i,ROWSM(es),j,COLSM(es),k,ESLENGTH(qn));}
	} else {
		PROGERROR("Cannot find scale of the base-minor in es, why???");	/* the matrices were scaled-equal! */
	}
	dispose_ematrix(res);
	FREE(qq);  FREE(sc);
	if (!r) { dispose_ematrix(es);  es = NULL; }
	if (!r)  {PROGERROR("Scale not found, really correct ???");}
		/* (this !r may happen only if genstep_linecheck_setref() fails somewhere, which is not possible now) */
	return es;
}

































/******************		*******************/



/**
 * 
**/




















