
/********************************************************************************
 *										*
 *				 M A C E K					*
 *		("Matroids Also Computed Efficiently" Kit)			*
 *										*
 * A set of tools and routines for computations with representable matroids.	*
 * Copyright (C) 2001,2002  Petr Hlineny.					*
 *										*
 *   This program is free software; you can redistribute it and/or modify	*
 *   it under the terms of the GNU General Public License as published by	*
 *   the Free Software Foundation; either version 2 of the License, or		*
 *   (at your option) any later version.					*
 *   You should have received a copy of the GNU General Public License		*
 *   along with this program; if not, write to the Free Software		*
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.	*
 * 										*
 * See the top-level "ABOUT" and "doc/ *" files for more information.		*
 *										*
 ********************************************************************************/













/**
 * Some theory for start... about structure of matroid representations:
 * 
 * This part contains fuctions for "structural work" with represented matroids.
 * (We do not attempt to work with matroids without a matrix representation - too complicated.)
 * Read ematrix.h to learn about our matrices, and pfield.h for the pfield arithmetic.
 * 
 * We consider matrix representations in the standard form - i.e. without the leading unit matrix.
 * The matrix lines are labeled with the matroid elements, here given initially
 * as 1...r and -1...-r*.
 * A particular matrix representation is determined by a choice of a basis in it
 * - forming the rows of the matrix.
 * The representation is invariant on scaling of matrix lines, and we may as well consider
 * it as invariant on line-permutations (provided we permute also the line-labels).
 * However, it is often good to stick with certain line-orders in a representation.
 * Sometimes we do not even care about the line-labels ("unlabeled" matroids).
 * In many functions we require the involved matroids to be connected (otherwise an error).
 * 
 * There is one BIG PROBLEM with matroid representations - caused by the fact that
 * non-equivalent representations exist for the same matroid.
 * Two matroid representations are "strongly equivalent" if one can be obtained from the other
 * one by pivoting (changing basis), row/column permutations, and nonzero scales.
 * (In this definition, we stick the line-labels with the lines, and we do not allow
 * pfield automorphisms.)
 * However, the nature of this program needs also a slightly weaker definition of an
 * "unlabeled (strong) equivalence", when a matrix strongly equivalent to the first matrix
 * is matroid-isomorphic to the second matrix.
 * (That means we forget the line-labels from the representations...)
 * 
 * For example, the matroid U24 has two strongly-non-equivalent representations over
 * GF(4), but these two are unlabeled-equivalent.
 * These problems must be taken seriously especially when looking for matroid minors
 * or matroid isomorphism.
 * Similar (but somehow different as well) problems exist when generating matrix
 * extensions, as described in gener.h ...
 * 
 * 
 * More description of the "structural" functions is given in the sections below...
 * 
**/



#ifndef STRUCT_H 
#define STRUCT_H 











/******************	Matrix automorphisms (straut.c)	*******************/
/**************************************************************************/



/**
 * An "automorphism" of a matrix A is a mapping (= row/column permutations and scales) that
 * maps the matrix A onto itself identically
 * - just a better name than "unlabeled-strongly-self-equivalent".
 * Automorphisms are useful when factoring out repeated occurences of the same
 * subrepresentation, etc.
 * .....
 * Do not mistake this term with automorphisms of the underlying matroid!
**/


/**
 * The following structure is supplied for description of a matrix automorphism.
 * A matrix automorphism consists of arbitrary permutations of rows and columns and of
 * nonzero scaling of rows and columns.
 * The associated row and column permutations are stored in pr[],pc[], resp.,
 * and the row and column scalings are stored in xr[],gr[] and xc[],gc[].
 * A function is given for allocating a new aut structure (dipose with "free()").
**/

typedef struct {
	
	short	*pr,*pc;	/* the row and column permutation (sends i,j to pr[i],pc[j]) */
	exp_t	*xr,*xc;	/* the arrays of exponents and signs of row/column multiplication */
	sign_t	*gr,*gc;
	int	nr,nc;
	long	sz;		/* (the size of this memory piece - for easy duplication) */
	
}	emxaut;

emxaut* new_emxaut(int r, int c) ;
/* (it is enough to "free(a)" to dispose an automorphism) */

void    emxaut_print(FILE *fout, emxaut *a, char *pref) ;

#ifndef	CACHEDAUTS
	/* cached (=recycled) automorphisms are used to speed up some repeated tests - use carefully */
#define	CACHEDAUTS	1
#endif


/**
 * This function computes all automorphisms of the given matrix, unscaled if un==0.
 * Read details in struct/straut.c.
 * The given matrix must be connected.
 * 
**/

emxaut**        struct_automorphisms_ext(int ch, ematrix *ee, int un) ;
#define	struct_automorphisms_ch(ch,e,un)	struct_automorphisms_ext(ch,e,un)
#define struct_automorphisms(e)			struct_automorphisms_ch(0,e,1)
#define struct_automorphisms_unscaled(e)	struct_automorphisms_ch(0,e,0)

emxaut**        struct_automorphisms_reuse(ematrix *e, int ll, emxaut **al) ;
	/* one may "recycle" a full non-altered list of automorphisms by this call: */
#define	struct_automorphisms_recycle(e,al)	struct_automorphisms_reuse(e,-1,al)
#define	struct_automorphisms_recfree()		struct_automorphisms_reuse(NULL,-1,NULL)










/******************	Bases and Minors (strmin.c)	*******************/
/**************************************************************************/


/**
 * The word "minor" is used in the matroidal sense, but that has little meaning in
 * matrices alone.
 * Thus we speak about a "displayed minor" which corresponds to a submatrix in
 * a suitable representation obtained by a choice of the basis (via pivoting and scaling).
 * When reporting displayed minors, we factor out all minor automorphisms (it is faster),
 * so in order to get really all displayed-minor mappings, one must to apply
 * the minor automorphisms later.
 * We shortly say a "dispminor".
 * 
 * In order to find whether a matroid has a given minor, we must search over all bases
 * of the matroid representation for a displayed minor similar to the searched minor.
 * Moreover, we either have to make sure that the searched minor appears (unlabeled-)
 * equivalent the given minor representation - like if the representation is unique,
 * or to give a list of all inequivalent representations of the minor in the search.
 * Notice that here we consider unlabeled(!) equivalence for the searched minor.
 * 
 * .............................
 * 
 * 
 * 
 * *********** another way of looking at minors - try all deletions/contractions
 * 	- good when the size difference is small,
 * 	  but it needs a very fast equivalence testing
 * 
 * *********** implement hash-codes based on matroidal properties
 * 	(small-rank flats for fast, count bases per elements/pairs, etc for slower)
 * 	- use different levels of hashes - matroidal, fixed basis, fixed rows...
 * 
 * 
**/


/**
 * The following functions produce (displayed) bases,
 * and look for displayed minors and matroid-minors in general.
 * Keep in mind that a minor is found only in the given representation class
 * if there are non-equivalent representations!!! - see above.
 * Read particular descriptions of the functions in struct/strmin.c.
 * And remember that the given matrices must be connected.
 * 
 * The functions work with matrix lists (ematrix**) that are variants of "alist"
 * from misc.c; these allows to handle collections of an arbitrary length.
 * These lists, as well as the matrices in them, should be freed by caller when
 * no longer used, by calling dispose_alist_mats(ls).
 * To save time, a list of all bases may be "recycled", i.e. stored for possible next call
 * for the same list (this must not be used for randomized bases or altered lists).
 * 
**/

ematrix**       struct_allbases_ext(int ch, ematrix *e) ;
#define	struct_allbases_ch(ch,e)	struct_allbases_ext(ch,e)
#define	struct_allbases(e)		struct_allbases_ch(0,e)
	/* special - ch==-7 returns only a random sublist of all bases - for quick pre-checks */
#define	struct_allbases_rand(e)	struct_allbases_ch(-7,e)

	/* one may "recycle" a full non-altered list of bases by calling estruct_allbases_recycle(): */
ematrix**       struct_allbases_reuse(ematrix *e, int ll, ematrix **bl) ;
#define	struct_allbases_recycle(bl)	struct_allbases_reuse(NULL,-1,bl)
#define	struct_allbases_recfree()	struct_allbases_reuse(NULL,-1,NULL)

#ifndef	CACHEDBASES
	/* cached (=recycled) bases are used to speed up some repeated tests - see struct_allbases_recycle() */
#define	CACHEDBASES	1
#endif


/**
 * These functions look for a dispminor or a minor in the given matrix.
 * The displayed minors are returned from struct_dispminor_ext() as a list of refering
 * (to e) submatrices such that they are equal to emin up to scales.
 * The functions struct_hasminorlist_ext() and struct_hasminor_ext() look for dispminors
 * in all bases generated for the given matrix e.
 * They return only whether a minor exists (1) or not (0).
**/

ematrix**       struct_dispminor_ext(int ch, ematrix *e, ematrix *emin) ;
#define	struct_dispminor_ch(ch,e,em)	struct_dispminor_ext(ch,e,em)
#define	struct_dispminor(e,em)		struct_dispminor_ch(0,e,em)
	/* special - ch>=-2 returns only 0/1 whether a dispminor exists, no real list is returned */
#define	struct_isdispminor(e,em)	(struct_dispminor_ch(-3,e,em)?1:0)
#define	struct_isdispminor_pr(e,em)	(struct_dispminor_ch(-2,e,em)?1:0)


int     struct_hasminorlist_ext(int ch, ematrix *e, ematrix **lmin) ;
#define	struct_hasminorlist_ch(ch,e,lmin)	struct_hasminorlist_ext(ch,e,lmin)
#define	struct_hasminorlist(e,lmin)		struct_hasminorlist_ch(0,e,lmin)
	/* special - ch==-7 tests only against a random sublist of bases - for quick pre-checks !!! */
#define	struct_hasminorlist_rand(e,lmin)	struct_hasminorlist_ch(-7,e,lmin)

int     struct_hasminor_ext(int ch, ematrix *e, ematrix *emin) ;
#define	struct_hasminor_ch(ch,e,emin)	struct_hasminor_ext(ch,e,emin)
#define	struct_hasminor(e,emin)		struct_hasminor_ch(0,e,emin)
#define	struct_hasminor_rand(e,emin)	struct_hasminor_ch(-7,e,emin)

	/* request to print the displayed minors when they are found in the computation */
void    struct_hasminor_printit_ext(FILE *fp, char *pref, int a) ;
#define	struct_hasminor_printit(fp,pr)	struct_hasminor_printit_ext(fp,pr,0)
#define	struct_hasminor_printall(fp,pr)	struct_hasminor_printit_ext(fp,pr,1)









/******************	Other structural... (struct.c)	*******************/
/**************************************************************************/


/**
 * The following functions test whether the two matrices are unlabeled (strongly) equivalent,
 * and whether they are equal up to line-scaling.
 * A dispminor-mapping between bases may be returned from struct_equivalent_ext() in bs,rbs.
 * Scales mapping between e1,e2 may be returned from struct_scaledequal_ext() in ao.
 * 
 * A "tight major" N of F is defined with respect to a (finite) family F of matroids.
 * N is a tight major of F if N has a minor from F and no element of N may be both contracted
 * or deleted keeping some minor isomorphic to one of F.
 * The function struct_hasdelcontr...() is provided for testing a tight major against
 * the list in tghl[] (in the opposite meaning of the return value).
 * 
**/

int     struct_equivalent_ext(int ch, ematrix *e1, ematrix *e2, ematrix **bs, ematrix **rbs) ;
#define	struct_isequivalent_ch(ch,e1,e2)	struct_equivalent_ext(ch,e1,e2,NULL,NULL)
#define	struct_isequivalent(e1,e2)		struct_isequivalent_ch(0,e1,e2)

int     struct_scaledequal_ext(int ch, ematrix *e1, ematrix *e2, emxaut *ao) ;
#define	struct_isscaledequal_ch(ch,e1,e2)	struct_scaledequal_ext(ch,e1,e2,NULL)
#define	struct_isscaledequal_scal(ch,e1,e2,ao)	struct_scaledequal_ext(ch,e1,e2,ao)
#define	struct_isscaledequal(e1,e2)		struct_isscaledequal_ch(0,e1,e2)


int     struct_hasdelcontr_ext(int ch, ematrix *em, ematrix **tghl, int *which) ;
#define	struct_hasdelcontr_ch(ch,em,tghl)	struct_hasdelcontr_ext(ch,em,tghl,NULL)
#define	struct_hasdelcontr(em,tghl)		struct_hasdelcontr_ch(0,em,tghl)
#define	struct_hasdelcontr_wh(em,tghl,wh)	struct_hasdelcontr_ext(0,em,tghl,wh)
	/* special - ch==-7 tests only against a random sublist of bases - for quick pre-checks !!! */
#define	struct_hasdelcontr_rand(em,tghl)	struct_hasdelcontr_ch(-7,em,tghl)










/******************	Handling 3-separations (strsep3.c)	*******************/
/**********************************************************************************/


/**
 * Here we provide some other "structural" functions dealing with 3-separations:
 *  
 * A fan of length d is a sequence of d element where consecutive triples switch between
 * being triangles and triads.
 * The function struct_hasfan..() looks whether the matrix e has a "fan" of the given length,
 * moreover ending with the given line lo if tr>=0 (row for tr==0, column for tr==1).
 * If one asks about fans ending with the given line, then the matroid must be 3-connected
 * after this line is removed(!), otherwise the answer is incorrect.
 * One may also ask for a fan only in the given submatrix (ec) of the whole matrix.
 * 
**/

int     struct_hasfan_ext(int ch, ematrix *e, ematrix *ec, int tr, int lo, int fan) ;
#define struct_hasfan_ch(ch,e,tr,lo,fan)	struct_hasfan_ext(ch,e,NULL,tr,lo,fan)
#define struct_hasfan_part(ch,e,ec,fan)		struct_hasfan_ext(ch,e,ec,-1,-1,fan)
#define struct_hasfan(e,tr,lo,fan)		struct_hasfan_ch(0,e,tr,lo,fan)
#define struct_hasfan_all(e,fan)		struct_hasfan_ch(0,e,-1,-1,fan)
#define struct_hasfan_max(e)			struct_hasfan_all(e,-1)

void    struct_hasfan_printit(FILE *fp, char *pref) ;	/* (request to print the fan) */

#define	struct_nofanall(e,fan)			(struct_hasfan_all(e,(fan))? -1:1)
#define	struct_nofanlastline(e,tr,fan)		(struct_hasfan(e,tr,-1,(fan))? -1:1)
	/* remember that when asking for "lastline" fans, the matrix must be 3-connected the step before! */
#define	struct_nofanlastcol(e,fan)		struct_nofanlastline(e,1,fan)
	/* special ch<0 -> not a full test since the matrix may not be representable, so cannot compute rank */
#define	struct_nofanlastline_nofull(e,tr,fan)	(struct_hasfan_ch(-1,e,tr,-1,fan)? -1:1)


/**
 * Next we compute a "full-closure" of a submatrix (like of a subset of elements of a matroid).
 * A full closure is defined by a sequence of closures and coclosures applied to the submatrix,
 * until this process stops (related to the pathwidth of a matroid).
 * 
 * The full-closure of a triangle or a triad in the matroid is called a "tail" of width 3.
 * It is a piece of the matroid that has path-width 3.
 * The function estruct_hastail3_ext() is provided to compute that, but
 * this function works only for 3-connected matroids!!!
 * (It is possible to specify in txl the list of trangles and triads to close in the search.)
 * Additionally, one may request certain modifications to the tail length for special tail types
 * (like a fan or a line) by the parameter tmod - see the macros TMOD_xxx... and the function.
 * 
**/

ematrix*        struct_fclosure_ext(int ch, ematrix *e, ematrix *re) ;
#define	struct_fclosure_ch(ch,e,re)	struct_fclosure_ext(ch,e,re)
#define	struct_fclosure(e,re)		struct_fclosure_ch(0,e,re)

int     struct_hastail3_ext(int ch, ematrix *e, int tr, int lo, ematrix **txl, int tail, int tmod) ;
#define struct_hastail3_ch(ch,e,tl,tm)		struct_hastail3_ext(ch,e,-1,-1,NULL,tl,tm)
#define struct_hastail3(e,tl,tm)		struct_hastail3_ch(0,e,tl,tm)

/**
 * The value of TMOD.. allows to request modifications to the tail size for particular tail
 * types (size up or down); the tail types must be self-dual.
 * Precisely speaking, when a sub-tail of the particular type is found inside the whole tail
 * (or if the whole tail is of that particular type for TMOD_xxxONLY flag),
 * then the value of TMOD_xxx modifier is ADDed to the length of the particular-type tail, 
 * and the resulting tail length is adjusted accordingly up.
 * For negative modifier value, if the whole tail length plus mod falls below the particular
 * type length, then the resulting tail length is adjusted accordingly down.
 * See more inside struct_hastail3_core() ...
**/
#define	TMOD_FAN(md)		(((md)>=0? ((md)&7):8+((-(md))&7))<<5)
#define	TMOD_FANONLY(md)	(TMOD_FAN(md)|(1<<9))
#define	ISTMOD_FAN(tm)		((((tm)>>5)&7)*(((tm)&(1<<8))?-1:1))
#define	ISTMOD_FANONLY(tm)	((tm)&(1<<9))

#define struct_notail3all(e,tl,tm)		(struct_hastail3(e,tl,tm)? -1:1)
#define struct_notail3lastline(e,tr,tl,tm)	struct_notail3all(e,tl,tm)
//********** last-line oriented checking is not yet implemented (would use txl as well...)




/**
 * We provide a simple greedy (but precise) algorithm to test for branchwidth 3 of a matroid.
 * 
 * We use the fact that any 3-separating set of size >=4 can be displayed in some
 * bw-3 decomposition, even when considering an already partitioned matroid.
 * On the other hand, if a bw-3 decomposition exists, then there always is 3-separating
 * subset of <=6 elements, or <=4 blocks if using bigger than 1-element parts.
 * Hence we just look for such possible 3-separating set.
 * If we find one, we make its union as a new block in the matroid.
 * If we succeed all the way to just two blocks at the end, we have a bw-3 decomposition.
 * Otherwise, the matroid has no such decomposition.
 * 
 * The return value is 1 for branch-width 3, and 0 for no one.
 * The function requires the matroid to be 3-connected !!!
 * 
**/


int     struct_hasbwidth3_ext(int ch, ematrix *ee) ;
#define	struct_hasbwidth3(e)	struct_hasbwidth3_ext(0,e)











#endif	/* (of #ifndef STRUCT_H) */





























