
/********************************************************************************
 *										*
 *				 M A C E K					*
 *		("Matroids Also Computed Efficiently" Kit)			*
 *										*
 * A set of tools and routines for computations with representable matroids.	*
 * Copyright (C) 2001,2002  Petr Hlineny.					*
 *										*
 *   This program is free software; you can redistribute it and/or modify	*
 *   it under the terms of the GNU General Public License as published by	*
 *   the Free Software Foundation; either version 2 of the License, or		*
 *   (at your option) any later version.					*
 *   You should have received a copy of the GNU General Public License		*
 *   along with this program; if not, write to the Free Software		*
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.	*
 * 										*
 * See the top-level "ABOUT" and "doc/ *" files for more information.		*
 *										*
 ********************************************************************************/















/**
 * Some theory for start... about matrix automorphisms:
 * (Read general description of "structural" functions on the top of ../include/struct.h.)
 * 
 * We want to produce all (unscaled or scaled) automorphisms of the given matrix
 * (or, in formal words, strongly self-equivalent mappings of the matrix).
 * To state that VERY clearly, an entry Eij at (i,j) is mapped by an automorphism to an entry
 * Eij*sr[pr[i]]*sc[pc[j]] at (pr[i],pc[j]), where pr,pc are the row and column permutations
 * and sr,sc are the row and column scales (indexed by the original matrix indices!).
 * 
 * There is a possibility of "recycling" the automorphisms, but this should be used with caution.
**/






#include "macek.h"
#include "str.h"












/******************	Automorphism structure	*******************/
/******************************************************************/


//#undef CURDLEV
#define CURDLEV		7		/* (need <=DEBUGLEV+2 to perform paranoic tests) */



/**
 * This is a supplementary function for use inside estruct_automorphisms().
 * It allocates and returns one structure for describing an automorphism of a matrix.
 * The whole structure uses one chunk of memory, so it is enough to "free(a)".
**/

emxaut*		new_emxaut(int r, int c) {
	emxaut	*a;
	int	k;
	
	k = MEMALIGNSIZE(sizeof(*a));
	k += MEMALIGNSIZE(r*sizeof(a->pr[0])) + MEMALIGNSIZE(c*sizeof(a->pc[0]));
	k += MEMALIGNSIZE(r*sizeof(a->xr[0])) + MEMALIGNSIZE(c*sizeof(a->xc[0]));
	k += MEMALIGNSIZE(r*sizeof(a->gr[0])) + MEMALIGNSIZE(c*sizeof(a->gc[0]));
	a = MMALLOC(k);
	a->sz = k;  a->nr = r;  a->nc = c;
	a->pr = (short*)MEMALIGNPOINT(a+1);
	a->pc = (short*)MEMALIGNPOINT(a->pr+r);
	a->xr = (exp_t*)MEMALIGNPOINT(a->pc+c);
	a->xc = (exp_t*)MEMALIGNPOINT(a->xr+r);
	a->gr = (sign_t*)MEMALIGNPOINT(a->xc+c);
	a->gc = (sign_t*)MEMALIGNPOINT(a->gr+r);
	
	if (a->sz< (byte1*)(a->gc+c)-(byte1*)a)  {PROGERROREXIT("something wrong with allocation of emxaut");}
	return a;
}


/**
 * This function simply prints the given automorphism a to the output fout with a prefix pref.
 * 
**/

void	emxaut_print(FILE *fout, emxaut *a, char *pref) {
	int	i,j;
	
	fprintf(fout,"%sPermutation %p of %dx%d:\n",pref,a,a->nr,a->nc);
	fprintf(fout,"%s row",pref);
	for (i=0; i<a->nr; i++)  fprintf(fout,"\t%d->%d ",i,a->pr[i]);
	fprintf(fout,"\tcol");
	for (i=0; i<a->nc; i++)  fprintf(fout,"\t%d->%d ",i,a->pc[i]);
	fprintf(fout,"\n%s rmul",pref);
	for (i=0; i<a->nr; i++)  fprintf(fout,"\t%s",pfield_pvalue(7,a->xr[i],a->gr[i]));
	fprintf(fout,"\tcmul");
	for (i=0; i<a->nc; i++)  fprintf(fout,"\t%s",pfield_pvalue(7,a->xc[i],a->gc[i]));
	fprintf(fout,"\n");
}


/**
 * This function allows to "re-use" once computed list of all automorphisms.
 * For this to work, one has first to call this function from estruct_automorphisms_ext() with the
 * generated list in al and the length in ll (used later to check correctness).
 * Then the user should call estruct_automorphisms_recycle() (a variant of this function)
 * to recycle the aut list instead of freeing it.
 * After that, the next call to estruct_automorphisms_ext() for the same matrix will return the
 * recycled list (really the same list, not a copy of it, so do not alter the list!).
**/

#if CACHEDAUTS>0

#ifndef	STR_MAXAUTCACHE
#define	STR_MAXAUTCACHE		2013
#endif
#define	STR_AUTCACHEIX(p)	HASHPOINTER((p),STR_MAXAUTCACHE)

emxaut**	struct_automorphisms_reuse(ematrix *e, int ll, emxaut **al) {
	emxaut		**a;
	static int	lsave[STR_MAXAUTCACHE], lgen[STR_MAXAUTCACHE], firstcall = 1;
	static emxaut	**save[STR_MAXAUTCACHE], **gen[STR_MAXAUTCACHE];
	static ematrix	*esave[STR_MAXAUTCACHE];
	int		i,j,lx;
	
	DEBUG(CURDLEV+1,"  call e = %p,  ll = %d,  al = %p;  i = %d\n",e,ll,al,e?STR_AUTCACHEIX(e):0);
	if (firstcall)  for (i=0; i<STR_MAXAUTCACHE; i++) {
		firstcall = 0;			/* first-call initialization */
		lsave[i] = lgen[i] = -1;
		save[i] = gen[i] = NULL;  esave[i] = NULL;
	}
	if (al?!*al:0)  {PROGERROREXIT("cannot handle empty aut list!");}
	
	if (ll>0) {		/* remembers the generated list for later check when recycling */
		gen[STR_AUTCACHEIX(e)] = al;
		lgen[STR_AUTCACHEIX(e)] = ll;
		
	} else if (al!=NULL) {	/* this part is called instead of freeing the aut list al */
		for (a=al; *a; a++) ;
		if (!e)  {PROGERROREXIT("must give the matrix when recycling auts!");}
		i = STR_AUTCACHEIX(e);
				/* (must conform to the last remembered list) */
		if (gen[i]==al) if (a-al!=lgen[i] || al[0]->nr!=ROWSM(e) || al[0]->nc!=COLSM(e))
			{PROGERROR("damaged list when calling \"recycle\", shoud not happen");}
		if (gen[i]!=al || a-al!=lgen[i]) {
			dispose_alist(al);
			DEBUG(CURDLEV-1," X not last gener list for %p at #%d recycling auts, %p!=%p || %d!=%d\n",e,i,gen[i],al,lgen[i],(int)(a-al));
		} else {
			if (esave[i])  dispose_ematrix(esave[i]);
			esave[i] = ematrix_copy(e);
			if (save[i])  dispose_alist(save[i]);
			save[i] = al;  lsave[i] = a-al;
		}
	} else if (e!=NULL) {	/* and here the function returns a previously recycled list */
		i = STR_AUTCACHEIX(e);
		if (esave[i])  if (ematrix_isequal(esave[i],e)) {
			lx = alist_getlength(save[i]);
			if (lx<=0)  {PROGERROR("cannot have empty list stored here");}
			DEBUG(CURDLEV,"  = returning recycled %d auts %p of the matrix %p.\n",lx,save[i],e);
			a = save[i];  save[i] = NULL;
			dispose_ematrix(esave[i]);  esave[i] = NULL;
			struct_automorphisms_reuse(e,lx,a);
			return a;	/* (must clear the list references before returning!) */
		}
	} else {
		firstcall = 1;	/* this NULL-call clears all cached lists from here */
		for (i=j=0; i<STR_MAXAUTCACHE; i++) {
			if (esave[i])  dispose_ematrix(esave[i]);
			if (save[i]) { dispose_alist(save[i]);  j++; }
		}
		if (j)  DEBUG(CURDLEV-1,"Freeing %d cached automorphism lists...\n",j);
	}
	return NULL;		/* (the recycled list is possibly returned above) */
}

#else

emxaut**	struct_automorphisms_reuse(ematrix *e, int ll, emxaut **al) {
	
	if (ll<0 && al!=NULL)
		dispose_alist(al);
	return NULL;
}

#endif

/**
 * 
 * 
**/


















/******************	Generating matrix automorphisms	*******************/
/**************************************************************************/


#undef CURDLEV
#define CURDLEV		7		/* (need <=DEBUGLEV+2 to perform paranoic tests) */



/**
 * This function generates all automorphisms (i.e. row and column permutations that,
 * possibly after rescaling, identically preserve all entries) of the matrix e.
 * The function works only for connected(!) matrices.
 * If un==1, then rescaling after permuting is allowed, unlike rescaling not allowed if un==0.
 * 
 * Each automorphism is stored in a structure emxaut keeping the row and column permutations
 * and the row/column scales.
 * To state that clearly, an entry Eij at (i,j) is mapped by an automorphism to an entry
 * Eij*sr[pr[i]]*sc[pc[j]] at (pr[i],pc[j]), where pr,pc are the row and column permutations
 * and sr,sc are the row and column scales (indexed by the original matrix indices).
 * Watch carefully in what direction the automorphisms are applied - the permutations are,
 * of course, closed under inversions, but the corresponding scales are not (they are inverted).
 * 
 * The automorphisms are returned in a (NULL terminated) list of pointers.
 * (Remember to free the list and the automorphisms later.)
 * The list includes the identity automorphism,
 * and there may be more than one automorphisms associated with one row permutation(!).
 * It is enough to "free(a)" to dispose an automorphism structure a.
 * There is a possibility of "recycling" the automorphisms, but this should be used with caution.
 * 
 * 
 * To speed-up the generating process, we use permutation- and scale-independent hash codes
 * for the rows and columns of the matrix.
 * These codes combine the numbers of zeros in the lines, and the numbers of zero-determinants
 * in the pairs of lines.
 * We first try all row-permutations that respect these hash codes, and then we try to fit
 * the columns for each such row-permutation (using the column hash-codes with an enhancement).
 * For the column-fitting, we require the matrix to be connected.
 * The column-fitting function is more general and it is also used in other structural functions.
 * 
 * This function is quite fast unless the matrix has too many symmetries.
 * If ch>0, then permutations are printed when they are found.
 * If ch<0, some speedups are selected in the code, like no extra paranoic tests.
 * There is a hard limit on the maximal matrix size MXSZ, but this is enough large for
 * any practical application.
**/

#define	MXSZ	(NUM_LONG_BITS-2)	/* (must fit into long bits!) */
#ifndef	STR_MAXAUTOMS
#define	STR_MAXAUTOMS	30000
#endif

emxaut**	struct_automorphisms_ext(int ch, ematrix *ee, int un) {
	emxaut		**aout;
	long		*ldata, hsr[MXSZ],hsc[MXSZ], *hsrr[MXSZ],*hscc[MXSZ];
	long		dsra[MXSZ+1], *dsr;
	int		i,k,l, ret,br, pr[MXSZ];
	ematrix		*e;
	
	if (ch<-1)  {PROGERROR("Do not call with ch==%d <-1",ch);}
	if (ROWSM(ee)>=MXSZ || COLSM(ee)>=MXSZ)	/* (not practical for larger matrices anyway) */
		{PROGERROREXIT("Not designed to deal with %d,%d > %d rows or columns",ROWSM(ee),COLSM(ee),MXSZ);}
	DEBUG(CURDLEV,"Computing %sautomorphisms for the matrix %p ...\n",un?"":"unscaled ",ee);
	if (un==1) {
		aout = struct_automorphisms_reuse(ee,-1,NULL);
		if (aout)  return aout;		/* when a copy of this aut list was cached before */
	}
	aout = NULL;
	e = ematrix_copy_to(ee,ROWSM(ee),COLSM(ee));	/* (it is safer to use a local copy - e may be a refering matrix) */
	
			/**
			 * First, we compute order-invariant (and scale invariant if un==1)
			 * hash values for each line and each pair of lines of e.
			 * These values are stored in the arrays hsr,hsc,hsrr,hscc.
			 * (If a row i can be mapped to j, then hsr[i]==hsr[j], and if rows
			 * {i,ii} can be mapped to {j,jj}, then hsrr[i][ii]==hsrr[j][jj].)
			 * Any automorphism must obviously respect these values.
			**/
	k = ROWSM(e)*ROWSM(e)+COLSM(e)*COLSM(e);
	ldata = MMALLOC(k*sizeof(ldata[0]));
	for (i=0; i<ROWSM(e); i++)
		hsrr[i] = ldata+ROWSM(e)*i;
	for (i=0; i<COLSM(e); i++)
		hscc[i] = ldata+ROWSM(e)*ROWSM(e)+COLSM(e)*i;
	ret = struct_autom_linehash(e,un,hsr,hsc,hsrr,hscc);	/* hash code - use ret ??? */
#ifndef	FASTPROG
	if (ch>0)  EMATDEBUG(CURDLEV+2,e,"\t\t");
	if (DEBUGLEV>=CURDLEV+2) {
		SDEBUG(CURDLEV+2,"\t\t- AUT row hash  ");  for (i=0; i<ROWSM(e); i++)  SDEBUG(CURDLEV+2,"%lX\t",hsr[i]);
		SDEBUG(CURDLEV+2,"\n\t\t- AUT col hash  ");  for (i=0; i<COLSM(e); i++)  SDEBUG(CURDLEV+2,"%lX\t",hsc[i]);
		SDEBUG(CURDLEV+2,"\n"); }
#endif
			/**
			 * Next, we cycle through all possible permutations permitted by
			 * the hash values of the rows.
			 * We sequentially choose the images of row 0,1,..., and we keep track
			 * of the image set in dsr[];  k keeps the number of current level.
			**/
	pr[0] = -1;  k = 0;
	dsr = dsra+1;  dsr[-1] = 0;
	while (k>=0) {
		for (i=pr[k]+1; i<ROWSM(e); i++)	/* next mapping (neighbour) */
			if (hsr[k]==hsr[i] && (dsr[k-1]&(1l<<i))==0)
				break;
		DEBUG(CURDLEV+3,"%*s neighbour found pr[%d]=%d%s\n",2*k,"",k,i,i>=ROWSM(e)?" \tXXX":"");
		pr[k] = i;
		if (i>=ROWSM(e)) {		/* done at this level, going back */
			--k;  continue;
		}
		dsr[k] = dsr[k-1]|(1l<<i);	/* list of "map-destination" neighbours */
			/**
			 * Whenever we choose an image of the row k above, we can check the
			 * pairs of rows with k for equality of the two-row hash.
			 * (This must always be possible if a permutation exists.)
			 * Finally, we move to the next level, or we ask to record the
			 * permutation if all is complete - one perm may produce more auts.
			 * Function estruct_autom_rowf() below is called for the row permutation.
			**/
		for (i=br=0; i<k; i++) {
			if (hsrr[i][k]!=hsrr[pr[i]][pr[k]])  br = i+1;
		}
		if (br) {			/* two-row hash failed for pr[k] */
			DEBUG(CURDLEV+3,"%*s  two-row %d,%d hash failed  XXX\n",2*k,"",k,br-1);
			continue;
		}
		if (k<ROWSM(e)-1) {		/* taking the next step if not finished */
			pr[++k] = -1;
		} else {			/* considering the row permutation found here */
			if (dsr[k]!=(1l<<ROWSM(e))-1)
				{PROGERROR("not a permutation: dsr=%lX, why???",dsr[k]);}
			aout = struct_autom_rowf(ch,e,un,pr,hsc,hscc,aout);
		}
	}
	l = alist_getlength(aout);
	DEBUG(CURDLEV,"   ... done with %d automorphisms for the matrix %p.\n",l,e);
	if (un==1)  struct_automorphisms_reuse(ee,l,aout);
	FREE(ldata);  dispose_ematrix(e);
	return aout;		/* the output array of permutations is returned, terminated by NULL */
}



/**
 * This is a supplementary function for use inside estruct_automorphisms().
 * The function computes an order-invariant (and scale-invariant if un==1) hash code
 * for the lines and line-pairs of the given matrix e.
 * The hash code is used when producing all (scaled) automorphisms of the matrix e.
 * These codes are stored in arrays hashr,hashc (lines) and hashrr,hashcc (line pairs),
 * the arryas must be allocated prior to calling this function.
 * 
 * The hash code takes all line entries if un==0, while it only counts zero entries if un==1.
 * The two-line hash counts zero 2x2 subdeterminants of these two (unordered) lines.
 * Then these codes are combined together in a magic (but correct) way...
**/

int	struct_autom_linehash(ematrix *e, int un,
			long hashr[], long hashc[], long *hashrr[], long *hashcc[]) {
	int	i,j, ii,jj, h, dif,odif;
	exp_t   xx;
	sign_t  gg;
	
	for (i=0; i<ROWSM(e); i++)  hashr[i] = 0;
	for (j=0; j<COLSM(e); j++)  hashc[j] = 0;
	for (i=0; i<ROWSM(e); i++)  for (ii=0; ii<ROWSM(e); ii++)  hashrr[i][ii] = 0;
	for (j=0; j<COLSM(e); j++)  for (jj=0; jj<COLSM(e); jj++)  hashcc[j][jj] = 0;

			/* we first compute the basic one-line hash, depending on un */
	if (un==0) {
		for (i=0; i<ROWSM(e); i++)  for (j=0; j<COLSM(e); j++) {
			h = SIGNM(e,i,j)? pfield_hashvalue(EXPM(e,i,j),SIGNM(e,i,j)): 0;
			hashr[i] += h%93+(h*h)%111;  hashc[j] += h%93+(h*h)%111;
		}
	} else {
		for (i=0; i<ROWSM(e); i++)  for (j=0; j<COLSM(e); j++)
			if (SIGNM(e,i,j)==0) {
				hashr[i]++;  hashc[j]++;
			}
	}
			/* the we compute the two-line hash - the number of zero subdeterminants */
	for (i=0; i<ROWSM(e); i++)  for (j=0; j<COLSM(e); j++)
		for (ii=i+1; ii<ROWSM(e); ii++)  for (jj=j+1; jj<COLSM(e); jj++) {
			ematrix_determinant_2x2(e,i,ii,j,jj,&xx,&gg);
			if (gg==0) {
				hashrr[i][ii]++;  hashrr[ii][i]++;
				hashcc[j][jj]++;  hashcc[jj][j]++;
			}
//****************** in fact, it is enough to consider determinants with consequent pairs of
//			lines here - to find just the skeleton of zero line-pairs (that form cliques)
//**************** maybe add here a computation of determinants of absolute-value entries ???
		}
	h = odif = -1;	/* finally, we combine the codes in a symmetric way together, several times */
	while (++h<ROWSM(e)) {
		dif = 0;
		for (i=0; i<ROWSM(e); i++)  for (ii=i+1; ii<ROWSM(e); ii++)
			if (hashr[i]!=hashr[ii])  dif++;
		for (j=0; j<COLSM(e); j++)  for (jj=j+1; jj<COLSM(e); jj++)
			if (hashc[j]!=hashc[jj])  dif++;
		if (dif<=odif)  break;	/* (we cycle as long as we are getting more difference) */
		odif = dif;
				/* one-line hash is symmetrically added to each two-line hash value */
		for (i=0; i<ROWSM(e); i++)  for (ii=0; ii<ROWSM(e); ii++)
			if (i!=ii)  hashrr[i][ii] += 101*(((h*h+2)*(hashr[i]+hashr[ii]))%11111);
		for (j=0; j<COLSM(e); j++)  for (jj=0; jj<COLSM(e); jj++)
			if (j!=jj)  hashcc[j][jj] += 101*(((h*h+2)*(hashc[j]+hashc[jj]))%11111);
				/* then the two-line hash is applied towards each line hash */
		for (i=0; i<ROWSM(e); i++)  for (ii=0; ii<ROWSM(e); ii++)
			if (i!=ii)  hashr[i] += 111*(((5*h+3)*hashrr[i][ii]*hashrr[i][ii])%10001);
		for (j=0; j<COLSM(e); j++)  for (jj=0; jj<COLSM(e); jj++)
			if (j!=jj)  hashc[j] += 111*(((5*h+3)*hashcc[j][jj]*hashcc[j][jj])%10001);
	}
	DEBUG(CURDLEV+2,"   line hash cycle reapplied %d-times...\n",h+1);
	return 1;	/* (may check whether the hash values allow a nontrivial automorphism) */
}


/**
 * This function is for internal use inside estruct_automorphisms():
 * It looks for all (also scaled if un==1) automorphisms of the matrix ee that permute
 * the rows by the given permutation pr[].
 * There are column and double-column hash codes given in hashc[] and hashcc[][] for use.
 * All automorphisms found here are stored in the static array yout[].
 * 
 * This function works only for connected matrices, but parallel lines are OK.
 * Notice that one row permutation pr[] may produce several automorphisms that differ
 * in row scales and column permutations.
 * If ch>0, then some matrices are printed, and paranoic debug tests are performed if ch>=0.
**/

emxaut**	struct_autom_rowf(int ch, ematrix *e, int un, int pr[],
					long hashc[], long *hashcc[], emxaut **aout) {
	emxaut		*a;
	void		*data;
	ematrix		*rep, *efch=NULL;
	int		i,j,r, pc[MXSZ], yold,ynew;
	exp_t		xr[MXSZ],xc[MXSZ];
	sign_t		gr[MXSZ],gc[MXSZ];

#ifndef	FASTPROG
	if (DEBUGLEV>=CURDLEV+3) {
		DEBUG(CURDLEV,"-- asking to test a row perm. %s ",un?"":"(unscaled!)");
		for (i=0; i<ROWSM(e); i++)  SDEBUG(CURDLEV,"%d->%d ",i,pr[i]);
		SDEBUG(CURDLEV,"\n");
	}
	if (ch>=0 && IFRANDDEBUGLESS(40))  for (i=0; i<ROWSM(e); i++)  for (j=0; j<ROWSM(e); j++)
		if ((i!=j && pr[i]==pr[j]) || pr[i]<0 || pr[i]>=ROWSM(e))
			{PROGERROR("pr[] is not a permutation (%d,%d), why???",i,j);}
#endif
	yold = alist_getlength(aout);	/* (counting the number of automorphisms recorded here) */
	for (i=0; i<ROWSM(e); i++)  pc[pr[i]] = i;
				/* refering matrix showing the permuted rows of e by pr[] */
	rep = ematrix_refer(e,-1,-1,0,COLSM(e));
	for (i=0; i<ROWSM(e); i++)
		ematrix_refadd_row(rep,pc[i]);	/* (pc[] is "inverse" of pr[] - just here) */
	
	r = struct_fitcols(1,rep,e,un,hashc,hashc,hashcc,hashcc, pc,xr,gr,xc,gc,&data);
	while (r>=0) {
			/**
			 * We actually call a separate function estruct_fitcols() to decide
			 * how the columns of the (row-permuted) matrix rep fit to columns
			 * of the original matrix e.
			 * The above call initializes all internal structures of estruct_fitcols().
			 * Then each subsequent call generates a new column mapping, that
			 * (after thorough debug testing) we record in an automorphism structure.
			 * (Data held by estruct_fitcols() are released at the end below.)
			**/
		r = struct_fitcols(0,rep,e,un,hashc,hashc,hashcc,hashcc, pc,xr,gr,xc,gc,&data);
		if (r<0)  break;			/* end of column mappings */
#ifndef	FASTPROG
		if (DEBUGLEV>=CURDLEV+3) {
			DEBUG(CURDLEV+3,"  > automorphism accepted\n");
			SDEBUG(CURDLEV,"\t\trow\t");  for (i=0; i<ROWSM(e); i++)  SDEBUG(CURDLEV,"%d->%d ",i,pr[i]);
			SDEBUG(CURDLEV,"\tcol\t");  for (j=0; j<COLSM(e); j++)  SDEBUG(CURDLEV,"%d->%d ",j,pc[j]);
			SDEBUG(CURDLEV,"\n");
		}if (DEBUGLEV>=CURDLEV+3 && un==1) {
			SDEBUG(CURDLEV,"\t\trmul\t");  for (i=0; i<ROWSM(e); i++)  SDEBUG(CURDLEV,"%s ",pfield_printvalue(8,xr[i],gr[i]));
			SDEBUG(CURDLEV,"\t\tcmul\t");  for (i=0; i<COLSM(e); i++)  SDEBUG(CURDLEV,"%s ",pfield_printvalue(8,xc[i],gc[i]));
			SDEBUG(CURDLEV,"\n");  if (ch>0)  EMATDEBUG(CURDLEV+4,e,"\t\t\t\t");
		}
		if (ch>=0 && IFRANDDEBUGLESS(88)) {		/* paranoic tests before accepting */
			for (i=0; i<COLSM(e); i++)  for (j=0; j<COLSM(e); j++)
				if ((i!=j && pc[i]==pc[j]) || pc[i]<0 || pc[i]>=COLSM(e))
					{PROGERROR("pc[] not a permutation (%d,%d), why???",i,j);}
			if (efch==NULL)  efch = new_ematrix(ROWSM(e),COLSM(e),ROWSM(e),COLSM(e));
			for (i=0; i<ROWSM(e); i++)  for (j=0; j<COLSM(e); j++)
				COPYEXSIGM(efch,pr[i],pc[j],e,i,j);
			if (un==1)
				for (i=0; i<ROWSM(e); i++)  for (j=0; j<COLSM(e); j++) {
					ematrix_multiply_entry(efch,i,j,1,xr[i],gr[i]);
					ematrix_multiply_entry(efch,i,j,1,xc[j],gc[j]);
				}
			if (!ematrix_isequal(e,efch)) {
				PROGERROR("the matrix is not the same after automorphism (aut above)!");
				EMATDEBUG(0,e,"\t");  EMATDEBUG(0,efch,"!\t"); }
		}
#endif
			/**
			 * Finally, it is time to record our newly found automorphism in the
			 * structure emxaut, which is stored into the array yout[].
			**/
		a = new_emxaut(ROWSM(e),COLSM(e));
		for (i=0; i<ROWSM(e); i++) {
			a->pr[i] = pr[i];
			if (un==1) { a->xr[i] = xr[i];  a->gr[i] = gr[i]; }
			else { pfield_setzeroexp(&(a->xr[i]));  a->gr[i] = 1; }
		}
		for (j=0; j<COLSM(e); j++) {
			a->pc[j] = pc[j];
			if (un==1) { a->xc[j] = xc[j];  a->gc[j] = gc[j]; }
			else { pfield_setzeroexp(&(a->xc[j]));  a->gc[j] = 1; }
		}
		aout = alist_append(aout,a);
	}
					/* destroy extra data used in estruct_fitcols() */
	struct_fitcols(-1,rep,e,un,hashc,hashc,hashcc,hashcc, pc,xr,gr,xc,gc,&data);
	dispose_ematrix(rep);
	ynew = alist_getlength(aout);
	if (ynew>STR_MAXAUTOMS)  {PROGERROREXIT("Too many output automorphisms >%d to handle",ynew);}
	if (efch!=NULL)  dispose_ematrix(efch);	/* (used in debug testing) */
	DEBUG(CURDLEV+3,"-> total %d automorphisms generated for the given row perm.\n",ynew-yold);
	return aout;
}











/******************	Mapping matrix columns	*******************/
/******************************************************************/


#undef CURDLEV
#define CURDLEV		7



/**
 * This function is called to find mappings of columns of the matrix efr to columns of
 * the matrix eto (scaled mappings if un==1) that respect rows of these matrices.
 * The first call must be with init==1, that initializes internal data structures
 * (stored within *pdata), and the last call must be with init==-1 to release the data.
 * Intermediate calls with init==0 look for all possible column mappings, one per call.
 * The function returns -1 if no more mappings are possible.
 * 
 * The matrices to map are in efr, eto - they need not have the same number of columns,
 * but they must have the same number of rows.
 * The arrays hscfr[],hscto[] give column hash codes - if a column j of efr can be mapped to
 * a column jj of eto, then hcfr[j]==hcto[jj].
 *  (These hash codes are copied and enhanced inside by adding the zero pattern of permuted rows.)
 * The arrays hsccfr[][],hsccto[][] give hash codes for (unordered) pairs of columns.
 * 
 * The arrays pcm,xr,gr,xc,gc are used to store the results of computation.
 * They must be allocated prior to calling this function.
 * pcm[] gets filled with the column permutation (mapping column j of efr to pcm[j] of eto),
 * and xr[],gr[] and xc[],gc[] are filled with the scale coefficients of rows and columns 
 * in the mapping (undefined if un==0) - as applied to permuted(!!!) columns of efr to get eto.
 * 
 * This function is used in struct_autom_rowf() as well as in struct_isminor_rowfix().
**/

/*#define	MXSZ	(NUM_LONG_BITS-2)	defined above... */

int	struct_fitcols(int init, ematrix *efr, ematrix *eto, int un,
			long hscfr[], long hscto[], long *hsccfr[], long *hsccto[],
			int pcm[], exp_t xr[], sign_t gr[], exp_t xc[], sign_t gc[], void **pdata) {
	
	long		zp, zzp[MXSZ], mtch[MXSZ], zf, *hcfr,*hcto, *pato, *ccfr, *rsc;
	ematrix		**escal;
	exp_t		xxc;
	sign_t		ggc;
	int		i,j,jj,k;
	static int	ilev = 0;	/* (just to check initialization level of this function) */
	
				/* prepares the data structures used here in computation */
	k = (COLSM(efr)+1)*sizeof(ematrix*) + (4*COLSM(efr)+1*COLSM(eto)+2)*sizeof(long);
	if (init==1)  *pdata = MMALLOC(k);
	hcfr = (long*)(*pdata);				/* (indexing  0...COLfr-1) */
	hcto = hcfr+COLSM(efr);				/* (indexing  0...COLto-1) */
	ccfr = hcto+COLSM(eto);				/* (indexing  0...COLfr-1) */
	pato = ccfr+COLSM(efr);  pato += 1;		/* (indexing -1...COLfr-1) - COLfr correct! */
	rsc = pato+COLSM(efr);  rsc += 1;		/* (indexing -1...COLfr-1) */
	escal = (ematrix**)(rsc+COLSM(efr));  escal += 1;	/* (indexing -1...COLfr-1) */
	if ((byte1*)(escal+COLSM(efr))-(byte1*)(*pdata)!=k)  {PROGERROREXIT("wrong memory allocation");}
	
			/**
			 * Local data for computation are prepared here (called with init==1):
			 * hcfr[],hcto[] are loaded with enhanced column hash codes (the
			 *  enhancement takes the row permutations into an account).
			 * ccfr[] is an order in which columns of efr may be taken to keep connectivity.
			 * escal[] keeps scaled copies of efr during the backtracking comp.
			 * rsc[] tells (in bitfield) which rows of escal[*] were scaled so far.
			 * pato[] is a bit pattern of columns of eto that are images of efr at level k.
			**/
	if (init==1) {
		if (ROWSM(eto)>=MXSZ || COLSM(eto)>=MXSZ)	/* (not practical for larger matrices anyway) */
			{PROGERROREXIT("not designed to deal with %d,%d > %d rows or columns",ROWSM(eto),COLSM(eto),MXSZ);}
		for (j=-1; j<COLSM(efr); j++)  escal[j] = NULL;
		ilev++;
		if (ROWSM(efr)!=ROWSM(eto)) {
			PROGERROR("cannot fit columns of different heights %d, %d",ROWSM(efr),ROWSM(eto));
			return -1; }
		if (COLSM(eto)<COLSM(efr))
			return -1;		/* cannot fit into less columns */
		
				/* we first "enhance" the given column hash to hcfr[],hcto[] by zero patterns */
		for (j=0; j<COLSM(efr); j++)  hcfr[j] = hscfr? hscfr[j]:0;
		for (j=0; j<COLSM(eto); j++)  hcto[j] = hscto? hscto[j]:0;
		for (j=0; j<COLSM(efr); j++) {
			for (i=0, zp=0; i<ROWSM(efr); i++)
				if (SIGNM(efr,i,j)!=0)  zp |= (1l<<i);
			hcfr[j] = 10001*hcfr[j] + zp%9999;
			zzp[j] = zp;		/* (used below to find connected order of cols) */
		}
		for (j=0; j<COLSM(eto); j++) {
			for (i=0, zp=0; i<ROWSM(eto); i++)
				if (SIGNM(eto,i,j)!=0)  zp |= (1l<<i);
			hcto[j] = 10001*hcto[j] + zp%9999;
		}
#ifndef	FASTPROG
		DEBUG(CURDLEV+3,"   -> mapping columns of %p to columns of %p, col hash:\n\t\t\t\t",efr,eto);
		if (DEBUGLEV>=CURDLEV+3) {
			for (i=0; i<COLSM(efr); i++)  SDEBUG(CURDLEV-1,"%lX  ",hcfr[i]);
			SDEBUG(CURDLEV-1,"\n\t\t\t\t");
			for (i=0; i<COLSM(eto); i++)  SDEBUG(CURDLEV-1,"%lX  ",hcto[i]);
			SDEBUG(CURDLEV-1,"\n");
		}
#endif
				/* we look whether the column codes allow a matching hcfr<->hcto : */
		for (j=0; j<COLSM(efr); j++)  for (mtch[j]= jj=0; jj<COLSM(eto); jj++)
			if (hcfr[j]==hcto[jj])  mtch[j] |= (1l<<jj);
		for (j=0; j<COLSM(efr); j++)
			if (mtch[j]==0)  return -1;	/* the column j has no image in the matrix */
		for (zp= i=0; i<1+COLSM(eto)/2; i++) {
			for (zp=k= j=0; j<COLSM(efr); j++)  if (i==0 || random()%(i+2)<i) {
				zp |= mtch[j];  k++;	/* union of the neighborhoods of the selected efr cols */
			}
			for (j=jj=0; jj<COLSM(eto); jj++)  j += ((zp&(1l<<jj))!=0);
			if (j<k)  return -1;		/* no matching - violates Hall's condition */
		}
				/* we find a connected order of columns of efr, stored in ccfr[] */
		ccfr[0] = 0;  k = 0;
		zf = (1l<<ccfr[0]);  zp = zzp[ccfr[0]];
		while (++k<COLSM(efr)) {
			for (j=0; j<COLSM(efr); j++)
				if ((zf&(1l<<j))==0 && (zp&zzp[j])!=0)  break;
			if (j>=COLSM(efr)) {
				PROGERROR("the base graph of matrix %p is not connected",efr);
				return -1; }
			ccfr[k] = j;
			zf |= (1l<<j);  zp |= zzp[j];
		}
		if (zf!=(1l<<COLSM(efr))-1)  {PROGERROREXIT("error computing connected order in the matrix");}
		
#ifndef	FASTPROG
		if (DEBUGLEV>=CURDLEV+3) {
			DEBUG(CURDLEV-1,"   -  connected col order: ");
			for (i=0; i<COLSM(efr); i++)  SDEBUG(CURDLEV-1,"%ld, ",ccfr[i]);
			SDEBUG(CURDLEV-1,"\n");
			EMATDEBUG(CURDLEV+4,efr,"\t\t\t<\t");  EMATDEBUG(CURDLEV+4,eto,"\t\t\t >\t");
		}
#endif
				/* remaining initializations for the backtracking process below */
		escal[-1] = ematrix_copy_to(efr,ROWSM(efr),COLSM(efr));
		pcm[ccfr[0]] = -1;  pato[-1] = rsc[-1] = 0;
		return 0;		/* (successfully initialized) */
	}
			/**
			 * This part is called only when the computation is finished (init==-1).
			 * All temporary matrices and data arrays are released.
			**/
	if (init==-1) {
		ilev--;
		for (j=-1; j<COLSM(efr); j++)
			if (escal[j]!=NULL)  dispose_ematrix(escal[j]);
		FREE(*pdata);
		return 0;
	}
	if (ilev<=0 || ilev>20)
		{PROGERROREXIT("wrong initialization of estruct_fitcols() ??? ilev=%d",ilev);}
	
			/**
			 * Here the heart of this function is.
			 * We look (in subsequent calls to this function) for all possible mappings
			 * of columns of efr to columns of eto, in order given by ccfr[].
			 * At level k we find a possible image j of the column ccfr[k], which
			 * we store in pcm[ccfr[k]].
			 * Then we test our choice against column and two-column hash codes.
			 * More is described below...
			 * (The meanings of backtracking variables are described above at init==1.)
			**/
	k = (pcm[ccfr[0]]==-1? 0:COLSM(efr)-1);	/* (the columns are ordered by ccfr[]-index !!!) */
	while (k>=0) {
		if (k>=COLSM(efr))  {PROGERROREXIT("cannot get here!!! %d",k);}
		for (j=pcm[ccfr[k]]+1; j<COLSM(eto); j++)
			if ((pato[k-1]&(1l<<j))==0 && hcfr[ccfr[k]]==hcto[j])
				break;		/* finding next suitable map of col ccfr[k] */
		pcm[ccfr[k]] = j;
		pato[k] = pato[k-1]|(1l<<j);
		if (j>=COLSM(eto)) {		/* no more choices at this level */
			--k;  continue;
		}				/* (zero-patten test was already hashed above) */
		if (hsccfr && hsccto) {
			for (i=0; i<k; i++)	/* two-column hash test */
				if (hsccfr[ccfr[i]][ccfr[k]]!=hsccto[pcm[ccfr[i]]][pcm[ccfr[k]]])
					break;
			if (i<k)  continue;	/* if column pairs do not match */
		}
		DEBUG(CURDLEV+4,"\t%*s. col map pass (k=%d) pcm[%ld]=%d\n",2*k,"",k,ccfr[k],pcm[ccfr[k]]);
		
			/**
			 * We make a level-k copy escal[k] of the matrix efr first.
			 * For the image pcm[ccfr[k]] (k>0), we find the first already scaled row
			 * that intersects this column in a non-zero (possible by connectivity),
			 * and we match the intersection entry by scaling the column in escal[k].
			 * If un==0, we skip the scaling phase.
			**/
		if (escal[k]!=NULL)  dispose_ematrix(escal[k]);
		escal[k] = ematrix_copy(escal[k-1]);	/* (escal[] are scaled copies of efr) */
		pfield_setzeroexp(&xxc);  ggc = 1;
		if (k>0 && un==1) {
			for (i=0; i<ROWSM(efr); i++)	/* finds an intersection with a scaled row */
				if ((rsc[k-1]&(1l<<i))!=0 && SIGNM(efr,i,ccfr[k])!=0)
					break;
			if (i>=ROWSM(efr))  {PROGERROREXIT("the base graph of matrix %p is not connected",efr);}
			pfield_mul(-1,EXPM(escal[k],i,ccfr[k]),SIGNM(escal[k],i,ccfr[k]),
					1,EXPM(eto,i,pcm[ccfr[k]]),SIGNM(eto,i,pcm[ccfr[k]]),
						&xxc,&ggc);
			if (ggc==0)  continue;		/* if zero does not fit in efr,eto */
			ematrix_multiply_col(escal[k],ccfr[k], 1,xxc,ggc);
#ifndef	FASTPROG
			if (!ematrix_isequal_entry(escal[k],i,ccfr[k],eto,i,pcm[ccfr[k]]))
				{PROGERROR("wrong multiplication of column");}
#endif
		}		/* the column scales must be stored against eto col indices!) */
		xc[pcm[ccfr[k]]] = xxc;  gc[pcm[ccfr[k]]] = ggc;
			/**
			 * After we have scaled the column pcm[ccfr[k]], we now use the nonzero
			 * entries of the column to determine the scale of yet unscaled rows,
			 * or to check equality on scaled rows (or on all rows if un==0).
			 * Remember that we must specially handle zero entries.
			 * 
			 * If this test passes OK, then we know the column pcm[ccfr[k]] fits,
			 * and we may proceed to the next level, or return the result.
			**/
		rsc[k] = rsc[k-1];
		for (i=0; i<ROWSM(efr); i++) {
						/* zero entries are not scalled, just compared */
			if (SIGNM(efr,i,ccfr[k])==0 && SIGNM(eto,i,pcm[ccfr[k]])==0)  continue;
			if (SIGNM(efr,i,ccfr[k])==0 || SIGNM(eto,i,pcm[ccfr[k]])==0)  break;
			
			if (un==1 && (rsc[k]&(1l<<i))==0) {	/* (yet unscaled row) */
				pfield_mul(-1,EXPM(escal[k],i,ccfr[k]),SIGNM(escal[k],i,ccfr[k]),
					1,EXPM(eto,i,pcm[ccfr[k]]),SIGNM(eto,i,pcm[ccfr[k]]),
						&xr[i],&gr[i]);
				ematrix_multiply_row(escal[k],i, 1,xr[i],gr[i]);
				rsc[k] |= (1l<<i);	/* another row is already scaled */
#ifndef	FASTPROG
				if (!ematrix_isequal_entry(escal[k],i,ccfr[k],eto,i,pcm[ccfr[k]]))
					{PROGERROR("wrong multiplication of row");}
#endif
			} else {
				if (!ematrix_isequal_entry(escal[k],i,ccfr[k],eto,i,pcm[ccfr[k]]))
					break;		/* if scaled-row entry does not fit */
			}
		}
		if (i<ROWSM(efr))  continue;		/* not passed the test on scaled rows */
		if (k<COLSM(efr)-1) {
			pcm[ccfr[++k]] = -1;		/* passed, go to the next level */
			continue;
		}
		if (un==1 && rsc[k]!=(1l<<ROWSM(efr))-1)
			{PROGERROR("not all rows were scaled when fitting columns (zero row?)");}
		return k;	/* successful return - the column mapping is in given variables */
	}
	DEBUG(CURDLEV+3,": fitting done with col maps.\n");
	return -1;		/* unsuccessful search - no more column mappings */
}



















/******************		*******************/



/**
 * 
**/












































