
/********************************************************************************
 *										*
 *				 M A C E K					*
 *		("Matroids Also Computed Efficiently" Kit)			*
 *										*
 * A set of tools and routines for computations with representable matroids.	*
 * Copyright (C) 2001,2002  Petr Hlineny.					*
 *										*
 *   This program is free software; you can redistribute it and/or modify	*
 *   it under the terms of the GNU General Public License as published by	*
 *   the Free Software Foundation; either version 2 of the License, or		*
 *   (at your option) any later version.					*
 *   You should have received a copy of the GNU General Public License		*
 *   along with this program; if not, write to the Free Software		*
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.	*
 * 										*
 * See the top-level "ABOUT" and "doc/ *" files for more information.		*
 *										*
 ********************************************************************************/














/**
 * Some theory for start... about 3-separations, fans, branch-width, and others:
 * (Read general description of "structural" functions on the top of ../include/struct.h.)
 * 
 * 
 * 
 * 
 * 
 * 
**/






#include "macek.h"
#include "str.h"












/******************	Checking for fans	*******************/
/******************************************************************/


//#undef CURDLEV
#define CURDLEV		7

static int	inrecur = 0;	/* (used for debug-testing below...) */
static FILE	*struct_printfan = NULL;
static char	*struct_printfanp = NULL;


/**
 * This function looks whether the matrix e has a "fan" of given length, moreover
 * ending with the line lo if tr>=0 (row for tr==0, column for tr==1),
 * and using only the lines of ec if given.
 * A fan of length d is a sequence of d element where consecutive triples switch between
 * being triangles and triads (parallel or serial pairs are not counted here).
 * 
 * If one asks about fans ending with the given line (tr>=0), then the matroids e and e without
 * that line must be 3-connected (!otherwise possibly incorrect answer!).
 * (These assumptions of 3-connectivity imply that the last added line must be the end of
 * a possible fan, and that the end triple must be a triangle (tr==1) or triad (tr==0).)
 * 
 * If tr>=0, then only fans ending with the given line lo (row for tr==0, column for tr==1)
 * are searched.
 * If lo==-1, then the last line is considered (determined row/col by tr).
 * If tr==-1, then the the whole matrix is searched without respect to the (last) line.
 * If fan==-1, then the length of the longest fan is returned (so all of them are searched).
 * If ec!=NULL, then only the lines of the refering matrix ec are considered in the search.
 * If ch<0, then the triple-row/triple-column tests are skipped, as they require proper
 * representability (which may not be correct yet in the generating process).
 * Hence ch<0 gives only a partial answer, a fan may be missed (this really happens in generating!).
 * 
 * The function returns 0 if there is NO such fan, and 1 if there is some;
 * or the length of the longest fan if fan==-1 was given.
**/

int	struct_hasfan_ext(int ch, ematrix *e, ematrix *ec, int tr, int lo, int fan) {
	ematrix		*ee,*eec;
	char		buf[500];
	int		i,k, st,tt1,tt2,tt,ttm=0, br, ret, lf,
			fn[102],fnm[102];
	
	if (ec)  if (!ISREFMAT(ec) || REFMAT(ec)!=e || ISTRANSPM(e)!=ISTRANSPM(ec))
		{ PROGERROR("When using ec, it must refer to e in the same transpose state"); ec = NULL; }
	if (ec && tr>=0) { PROGERROR("When using ec, do not use tr>=0"); tr = -1; }
	if (fan>=0 && (fan<=3 || fan>=100)) { PROGERROR("Not designed to deal with fans out of range 3< %d <100",fan); return 0; }
	if (lo<0)  lo = tr? COLSM(e)-1:ROWSM(e)-1;
#ifndef FASTPROG
	if (!inrecur)  DEBUG(CURDLEV+0,"looking for %d-fans in %p %dx%d (tr=%d, lo=%d%s)\n",fan,e,ROWSM(e),COLSM(e),tr,lo,ec?", ec":"");
	if (ch>0) { EMATDEBUGS(CURDLEV+1,e,"\t\tf\t");  if (ec) EMATDEBUGS(CURDLEV+1,ec,"\t\t\t-\t"); }
	if (ch>=0 && tr>=0 && IFRANDDEBUGLESS(55)) {
		ee = ematrix_copy(e);	/* debug-check for parallel and serial lines without lo */
		if (tr==0)  ematrix_remove_row(ee,lo);
		if (tr==1)  ematrix_remove_col(ee,lo);
		if (ematrix_parallel_any(ee)>=0)  {PROGERROR("wrong - serial/parallel pair detected without lo=%d",lo); EMATDEBUG(0,ee,"!!\t");}
		dispose_ematrix(ee);
	}
#endif
		/* initial preparation, depending on given conditions... */
	if (tr<0) {		/* no last line for the fan given - trying everything */
		st = 0;  tt1 = 0;  tt2 = 1;
	} else {		/* the last fan line is given, also determines tt by 3-connectivity */
		st = 1;  tt1 = tt2 = tr? 1:0;
		fn[0] = tr? ROWSM(e)+lo:lo;
		DEBUG(CURDLEV+2,".. hasfan pre-determined %s fn[0] = %d, tt=%d\n",tr?"col":"row",lo,tt1);
	}
		/**
		 * We cycle all choices of k lines and starting tt values (triad 0/triangle 1),
		 * beginning with a choice of fn[st], where st and tt1,tt2 depend on the input.
		 * One choice is a row for <ROWSM and a column +ROWSM otherwise.
		 * We record the longest fan length in lf, and the fan itself in fnm[].
		 * If we look only for a fan longer than the given value, we stop immediately
		 * after we find it, otherwise we search through all fans to find the longest one.
		 * As we generate our choice, we test the consecutive triples for being
		 * triangles/triads (by tt).
		**/
	ret = lf = 0;
	for (tt=tt1; tt<=tt2 && !ret; tt+=(ret?0:1)) {
	  k = st;  fn[k] = -1;
	  while (k>=st && !ret) {
			/* the value of fn[k] is the current choice - row for <ROWSM, column+ROWSM otherwise */
		if (++fn[k]>=ROWSM(e)+COLSM(e)) {
			--k;  continue;		/* (no more choices at this level) */
		}
		for (i=0; i<k; i++)  if (fn[i]==fn[k])  break;
		if (i<k)  continue;		/* (if fn[k] was already chosen previously) */
		if (ec) {
			if (fn[k]<ROWSM(e))  for (i=ROWSM(ec)-1; i>=0 && GETREFMROW(ec,i)!=fn[k]; i--) ;
			else  for (i=COLSM(ec)-1; i>=0 && GETREFMCOL(ec,i)!=fn[k]-ROWSM(e); i--) ;
			if (i<0)  continue;	/* (if fn[k] does not belong to ec if given) */
		}
		DEBUG(CURDLEV+2+inrecur,"%*s. hasfan(%d) trying choice %s fn[%d] = %d (%d) for %s\n",2*k,"",fan,
				fn[k]<ROWSM(e)?"row":"col",k,fn[k]<ROWSM(e)?fn[k]:fn[k]-ROWSM(e),fn[k]<ROWSM(e)?ROWSID(e,fn[k]):COLSID(e,fn[k]-ROWSM(e)),(k+tt)%2?"triang":"triad");
		
		if (k>=2) {	/* looking for a triangle ((k+tt)%2==1) or triad ((k+tt)%2==0) at this level */
			br = struct_hasfan_triax(ch,e,(k+tt)%2,fn[k-2],fn[k-1],fn[k],1);
			if (br<=0)  continue;	/* if there was no triaxx, then cycle other choices */
		}
		if (++k>lf) {			/* records the longest fan found so far */
			lf = k;  ttm = tt;
			for (i=0; i<k; i++)  fnm[i] = fn[i];
			DEBUG(CURDLEV+2+inrecur,"-%*s. hasfan(%d) found longer = %d\n",2*k,"",fan,lf);
		}
		if (k<fan || fan<0)  fn[k] = -1;	/* to the next choice of line, or */
		else  ret = 1;				/* fan of given length found here */
		if (k>100)  {PROGERROREXIT("Out of index range looking for a fan");}
	}}
	if (fan<0)  ret = lf>=3? lf:0;
	else  ret = ret? 1:0;	/* return value indicates existence of a fan, or the longest fan */
	
		/* printing the fan if requested or for debug purposes... */
	if (!inrecur && ret && ((ch>=0 && CURDLEV-1<=DEBUGLEV) || struct_printfan)) {
		sprintf(buf,"fan: (%s+..) ",ttm?"triangle":"triad");
		for (i=0; i<lf; i++) {
			buf[400] = 0;  k = strlen(buf);
			snprintf(buf+k,50,"%s%d(%d), ",fnm[i]<ROWSM(e)?"r":"c",
				fnm[i]<ROWSM(e)?fnm[i]:fnm[i]-ROWSM(e),fnm[i]<ROWSM(e)?ROWSID(e,fnm[i]):COLSID(e,fnm[i]-ROWSM(e)));
		}
		if (struct_printfan)  fprintf(struct_printfan,"%s  %s\n",struct_printfanp,buf);
		struct_printfan = NULL;
	}
#ifndef FASTPROG		/* final paranoic tests... */
	DEBUG(CURDLEV-1+(!ret)+(!!inrecur),"- hasfan(%d) test for %p (l %d, tr=%d%s) found max %d - %s\n",fan,e,lo,tr,ec?", ec":"",lf,ret?"HAS":"no fan");
	DEBUG(CURDLEV-1,"  -> %s\n",buf);
	if (ret>0 && tr>=0 && IFRANDDEBUG(222) && !inrecur && !ec)
	  if ((fnm[lf-1]>=ROWSM(e))==((lf+tt)%2==0)) {
		inrecur = 1;		/* (debug-testing the symmetric situation - if the other end of the fan fits) */
		DEBUG(CURDLEV-0," ... hasfan test checking the reverse order recursively now ...\n");
		if (!struct_hasfan_ext(ch,e,NULL,(fnm[lf-1]>=ROWSM(e)),(fnm[lf-1]>=ROWSM(e)?fnm[lf-1]-ROWSM(e):fnm[lf-1]),fan))
			{PROGERROR("The result must be symmetric! ret=%d",ret); EMATDEBUG(0,e,"e !!\t");}
		/* (if the recur-call fails due to lack of 3-connectivity, then there is something wrong anyway) */
		inrecur = 0;
	}
	if (tr<0 && IFRANDDEBUG(222) && !inrecur) {
		inrecur = 1;		/* (debug-testing the full-test for diff basis) */
		DEBUG(CURDLEV-0," ... hasfan test checking another basis recursively now ...\n");
		for (ee=ematrix_copy(e),eec=ec, tt=0; tt<5; tt++) {
			i = random()%ROWSM(ee);  k = random()%COLSM(ee);
			if (SIGNM(ee,i,k)==0)  continue;
			if (ec!=NULL) {	/* (not to interfere with the matrix ec if given) */
				for (tt1=ROWSM(ec)-1; tt1>=0 && GETREFMROW(ec,tt1)!=i; tt1--) ;
				for (tt2=COLSM(ec)-1; tt2>=0 && GETREFMCOL(ec,tt2)!=k; tt2--) ;
				if (tt1>=0 && tt2>=0)  ematrix_pivot(ee,i,k);
			} else  ematrix_pivot(ee,i,k);
		}
		if (ec) { eec = ematrix_refer_all(ec);  ematrix_rerefer(eec,ee); }
		else if (random()%2==1)  ematrix_transpose(ee);
		if (ret!=struct_hasfan_ext(ch,ee,eec,-1,-1,fan))  {PROGERROR("The result must be the same after pivoting (transpose)! ret=%d",ret);}
		inrecur = 0;  if (ec) dispose_ematrix(eec);  dispose_ematrix(ee);
	}
#endif
	ee = eec = NULL;
	return ret;
}


/**
 * This function is for use in estruct_hasfan_ext() and estruct_hastail_ext() only...
 * It tests a triangle (trt==1) or triad (trt==0) on the given lines of e, taking lX for a row
 * when <ROWSM(e), and for a column otherwise.
 * If nps==1, then only proper trangles and triads are allowed, no parallel or series pairs.
 * If ch<0, then the triple-row/triple-column tests are skipped, as they require proper
 * representability (which may not be correct yet in the generating process).
 * The return value is 1 for a triaxx, 0 for no one, and -1 for parallel/series pair.
**/

int	struct_hasfan_ppair(ematrix *e, int l1, int l2, int t1, int t2) {
	switch (t1+2*t2) {
	case 0:	return 0;
	case 1:	return ematrix_isparallel_rc(e,l2,l1);
	case 2:	return ematrix_isparallel_rc(e,l1,l2);
	case 3:	return ematrix_isparallel_cc(e,l1,l2);
	}
	return 0;
}

int	struct_hasfan_triax(int ch, ematrix *e, int trt, int l1, int l2, int l3, int nps) {
	int	r,t1,t2,t3, ret;
	
	trt = (trt?1:0);  r = ROWSM(e);
	if (!trt)  ematrix_transpose(e);
	t1 = t2 = t3 = !trt;	/* deciding row/column and the test to perform... */
	if (l1>=r) { t1 = trt;  l1 -= r; }
	if (l2>=r) { t2 = trt;  l2 -= r; }
	if (l3>=r) { t3 = trt;  l3 -= r; }
	
	ret = 0;		/* rejecting parallel pairs in the triple if nps is given */
	if (nps)  if (struct_hasfan_ppair(e,l1,l2,t1,t2) ||
			struct_hasfan_ppair(e,l1,l3,t1,t3) || struct_hasfan_ppair(e,l3,l2,t3,t2))
		ret = -1;
				/* testing for the triangle (triad is now transposed): */
	if (ret>=0)  switch (t1+2*t2+4*t3) {
	case 1:	ret = ematrix_istriangle_rrc(e,l2,l3,l1);
		break;
	case 2:	ret = ematrix_istriangle_rrc(e,l1,l3,l2);
		break;
	case 4:	ret = ematrix_istriangle_rrc(e,l1,l2,l3);
		break;
	case 3:	ret = ematrix_istriangle_rcc(e,l3,l1,l2);
		break;
	case 5:	ret = ematrix_istriangle_rcc(e,l2,l1,l3);
		break;
	case 6:	ret = ematrix_istriangle_rcc(e,l1,l2,l3);
		break;
	case 7:	ret = (ch>=0? ematrix_istriangle_ccc(e,l1,l2,l3):0);
		/* (the matrix may not be representable for ch<0, so cannot call _ccc !) */
		break;
	case 0:	ret = 0;	/* (no triangle on three rows) */
		break;
	default:PROGERROR("Cannot get here!");	ret = -1;  break;
	}
	if (!trt)  ematrix_transpose(e);
	DEBUG(CURDLEV+1," Test returning %s %s.\n",trt?"triangle":"triad",ret>0?"YES":(ret<0?"parallel":"NO"));
	return ret;	/* (the above tests return 1 for triaxx, and 0 or -1 for no one) */
}


/**
 * 
 * request to print the fan found here.......
**/

void	struct_hasfan_printit(FILE *fp, char *pref) {
	struct_printfan = fp;
	struct_printfanp = pref?pref:"";
}

















/******************	Checking for tails	*******************/
/******************************************************************/


#undef CURDLEV
#define CURDLEV		4


/**
 * This function computes the full closure of the given submatrix re in the matrix e.
 * A full closure is defined by a sequence of closures and coclosures applied to the submatrix,
 * until this process stops.
 * The full-closure of re is returned in a new refering matrix to e.
 * 
 * ........ maybe store also the sequence of matrices as the closures are generated ???
**/

ematrix*	struct_fclosure_ext(int ch, ematrix *e, ematrix *re) {
	ematrix		*ec1,*ec2;
	int		i,om;
	
	if (ISREFMAT(e) || !ISREFMAT(re) || REFMAT(re)!=e)  {PROGERROREXIT("call only for a refering matrix re -> e");}
	if (ISTRANSPM(e)!=ISTRANSPM(re))  {PROGERROR("call only for matrices e,re of the same transpose");}
	DEBUG(CURDLEV+1,"looking for a full closure of re: %p %dx%d  in e: %p %dx%d\n",re,ROWSM(re),COLSM(re),e,ROWSM(e),COLSM(e));
	
	i = om = 0;
	ec1 = re;		/* a simple loop switching between closures and co-closures */
	while (i++<2 || om<ROWSM(ec1)+COLSM(ec1)) {
		ec2 = ematrix_closure_ext(e,ec1,i%2);
		om = ROWSM(ec1)+COLSM(ec1);
		if (ec1!=re)  dispose_ematrix(ec1);
		ec1 = ec2;
	}
#ifndef FASTPROG
	if (ch>0) { EMATDEBUG(CURDLEV+2,e,"\t\t\t"); EMATDEBUGS(CURDLEV+2,re,"\t\t\t fc\t"); }
	else if (CURDLEV+1<=DEBUGLEV) {
		DEBUG(CURDLEV+1," given re:\t");  for (i=0; i<ROWSM(re); i++)  SDEBUG(CURDLEV+1,"(%d) ",ROWSID(re,i));
		SDEBUG(CURDLEV+1," x  ");  for (i=0; i<COLSM(re); i++)  SDEBUG(CURDLEV+1,"(%d) ",COLSID(re,i));
		SDEBUG(CURDLEV+1,"\n");
	}
	if (ch>0)  EMATDEBUGS(CURDLEV+2,ec2,"\t\t\t->\t");
	else if (CURDLEV+1<=DEBUGLEV) {	
		DEBUG(CURDLEV+1,"-> full cl:\t");  for (i=0; i<ROWSM(ec2); i++)  SDEBUG(CURDLEV+1,"(%d) ",ROWSID(ec2,i));
		SDEBUG(CURDLEV+1," x  ");  for (i=0; i<COLSM(ec2); i++)  SDEBUG(CURDLEV+1,"(%d) ",COLSID(ec2,i));
		SDEBUG(CURDLEV+1,"\n");
	}
#endif
	return ec2;
}



/**
 * This function looks whether the matrix e has a "width-3 tail" of the given length.
 * Moreover, if tr>=0 is given, then it assumes that the searched tail did not exist
 * before the last line lo (row for tr==0, column for tr==1) was added to the matrix.
 * A "tail of width 3" is a piece of the matroid having path-width 3, and it is actually
 * obtained by a full closure of a triangle or a triad.
 * See struct_fclosure_ext() above.
 * The matrix must be properly represented and 3-connected when calling this function!
 * 
 * If tr==-1, then the the whole matrix is searched without respect to the (last) line lo.
 * If tail==-1, then the length of the longest tail is returned.
 * If lo==-1, then the last line is considered (determined row/col by tr) as the added one.
 * If txl!=NULL, then only the list of starting submatrices from txl is considered when
 * making full closures.
 * 
 * The value of tmod allows to request modifications to the tail size for particular tail
 * types (size up or down), as defined by the macros TMOD_xxx (see in the header).
 * This tail type must be self-dual.
 * Precisely speaking, when a sub-tail of the particular type is found inside the whole tail
 * (or if the whole tail is of that particular type for TMOD_xxxONLY flag),
 * then the value of TMOD_xxx modifier is ADDed to the length of the particular-type tail, 
 * and the resulting tail length is adjusted accordingly up.
 * For negative modifier value, if the whole tail length plus mod falls below the particular
 * type length, then the resulting tail length is adjusted accordingly down.
 * See more inside struct_hastail3_core() ...
 * 
 * The function returns 0 if there is NO such tail, and 1 if there is some;
 * or the length of the longest tail if tail==-1 is given (subject to modifications by tmod).
**/

int	struct_hastail3_ext(int ch, ematrix *e, int tr, int lo, ematrix **txl, int tail, int tmod) {
	ematrix		*ee, *re;
	int		i,k, x,y,z, tl, ret=0;
	
	if (tail>=0 && tail<3)  {PROGERROR("What is a tail of size %d ???",tail);}
	if (tr>=0 || txl!=NULL)  {PROGERROR("Not implemented yet");  return lo=0;}
	if (lo<0)  lo = tr? COLSM(e)-1:ROWSM(e)-1;
#ifndef FASTPROG
	if (ch>=0 && IFRANDDEBUGLESS(222))
		if (ematrix_inpfield_rand(e)<0)  {PROGERROR("The matrix must be properly represented here!");}
	if (!inrecur)  DEBUG(CURDLEV,"looking for tail3 of len %d (tmod %X) in %p %dx%d (tr=%d, lo=%d%s)\n",tail,tmod,e,ROWSM(e),COLSM(e),tr,lo,txl?", txl":"");
	if (ch>0)  EMATDEBUG(CURDLEV+1,e,"\t\t-..\t");
#endif
	ee = ematrix_copy(e);	/* (works on a copy of the matrix, but does not change it) */
	
		/**
		 * A tail of width 3 must end with a triangle or a triad, so we try to look
		 * for all such triples in three loops (x,y,z).
		 * We use the above function struct_hasfan_triax() to test for them.
		 * If a triangle or a triad is found, we compute its full closure in the matrix,
		 * which gives us the whole tail.
		 * Then we call struct_hastail3_core() to compare the size of the tail
		 * with the given parameter (using tmod as well), or to find the longest tail.
		**/
	ret = tl = 0;
	for (x=0; x<ROWSM(ee)+COLSM(ee) && !ret; x++)
	  for (y=x+1; y<ROWSM(ee)+COLSM(ee) && !ret; y++)
		for (z=y+1; z<ROWSM(ee)+COLSM(ee) && !ret; z++) {
			//********* how this computation should be when not 3-connected???
			if (!struct_hasfan_triax(0,ee,0,x,y,z,0) && !struct_hasfan_triax(0,ee,1,x,y,z,0))
				continue;	/* (only triples that are triangles or triads are interesting) */
			re = ematrix_refer_empty(ee);
			ematrix_refadd_rc(re,x<ROWSM(ee)?x:-1,x<ROWSM(ee)?-1:x-ROWSM(ee));
			ematrix_refadd_rc(re,y<ROWSM(ee)?y:-1,y<ROWSM(ee)?-1:y-ROWSM(ee));
			ematrix_refadd_rc(re,z<ROWSM(ee)?z:-1,z<ROWSM(ee)?-1:z-ROWSM(ee));
			
			k = struct_hastail3_core(ch,ee,re,tail,tmod);
			if (k>tl)  tl = k;	/* (the longest tail found so far) */
			dispose_ematrix(re);
			if (tail>0 && tl>=tail)  ret = 1;
		}
	//******* for txl!=NULL, use the list for triangles and triads, also consider the last line...
	
	if (tail<0)  ret = tl>=3? tl:0;
	else  ret = ret? 1:0;	/* return value indicates existence of a tail, or the longest tail */
	i = 0;
#ifndef FASTPROG
	DEBUG(CURDLEV-1+(!ret)+(!!inrecur),"- hastail(%d) test for %p (l %d, tr=%d, tmod %X) found max %d - %s\n",tail,e,lo,tr,tmod,tl,ret?"HAS":"no tail");
	if (tr<0 && txl==NULL && ch>=0) if (IFRANDDEBUG(111) && !inrecur) {
		inrecur = 1;		/* (debug-testing the full-test for diff basis) */
		DEBUG(CURDLEV-0," ... hastail3 test checking another basis recursively now ...\n");
		for (x=0; x<5; x++) {
			i = random()%ROWSM(ee);  y = random()%COLSM(ee);
			if (SIGNM(ee,i,y)!=0)  ematrix_pivot(ee,i,y);
		}
		ematrix_transpose(ee);
		if (ret!=struct_hastail3_ext(ch,ee,-1,-1,NULL,tail,tmod))  {PROGERROR("The result must be the same! ret=%d",ret);}
		ematrix_transpose(ee);  inrecur = 0;
	}		/* (ee is changed here !!!) */
#endif
	dispose_ematrix(ee);	/* (ee is a copy of e used in the computation) */
	return ret;
}


/**
 * The computation of one tail is performed here, including tmod modifications for special tail types.
 * The tail length is returned, after tmod modifications (see TMOD... in frame.h).
 * 
 * For use only in struct_hastail3_ext()...
**/

int	struct_hastail3_core(int ch, ematrix *ee, ematrix *re, int tail, int tmod) {
	ematrix		*rf;
	int		i, lc,ln,lnn,ll, tm;
	
	rf = struct_fclosure_ch(DECRCH(ch),ee,re);
	lc = ln = lnn = ROWSM(rf)+COLSM(rf);
	ll = 0;
	tm = ISTMOD_FAN(tmod);		/* modify the length if the tail is a fan (or close to) */
	if ((tm!=0 && tail<0) || (ln<tail && ln+tm>=tail) || (ln>=tail && ln+tm<tail)) {
		//printlev -= 1;
		ll = struct_hasfan_part(ch>1?ch-1:0,ee,rf,-1);	/* (ch<0 has spec. meaning in hasfan!) */
		//printlev += 1;
		if (ll==ln || !ISTMOD_FANONLY(tmod)) {
			if (tm>0 && ll+tm>ln)  lnn = ll+tm;
			if (tm<0 && ll>ln+tm)  lnn = ln+tm;
		}
	}
	//******** other tail-length modifications similarly... - but against the original lc !!!
	ln = lnn;
	i = 0;
#ifndef FASTPROG
	if (lc!=ln)  DEBUG(CURDLEV+1,"  - tail3 tmod-changed (%X) length %d -> %d (ll=%d)\n",tmod,lc,ln,ll);
	if (tail>0 && ln>=tail && !inrecur) {	/* if a sufficiently long tail is found... */
		DEBUG(CURDLEV,"  - tail3 of length %d>=%d (tmod %X) found here\n",ROWSM(rf)+COLSM(rf),tail,tmod);
		if (ch>0 && CURDLEV<=DEBUGLEV)  EMATDEBUGS(CURDLEV,rf,"\t\t-->\t");
		else if (CURDLEV-1<=DEBUGLEV && ch>=0) {
			DEBUG(CURDLEV-1,"  --> tail3 %d~%d: ",lc,ln);  for (i=0; i<ROWSM(rf); i++)  SDEBUG(CURDLEV-1,"(%d) ",ROWSID(rf,i));
			SDEBUG(CURDLEV-1," x  ");  for (i=0; i<COLSM(rf); i++)  SDEBUG(CURDLEV-1,"(%d) ",COLSID(rf,i));
			SDEBUG(CURDLEV-1,"\n");	}
	}
#endif
	dispose_ematrix(rf);
	return ln;
}






















/******************	Testing Branch-width 3	*******************/
/******************************************************************/


#undef CURDLEV
#define CURDLEV		6


/**
 * This function computes a branch-width-3 decomposition of the given matroid in ee
 * if one exists.
 * (So far, the decomposition is only implicit in the process, not returned back...)
 * The function requires the matroid to be 3-connected !!!
 * 
 * We use the fact that any 3-separating set of size >=4 can be displayed in some
 * bw-3 decomposition, even when considering an already partitioned matroid.
 * On the other hand, if a bw-3 decomposition exists, then there always is 3-separating
 * subset of <=6 elements, or <=4 blocks if using bigger than 1-element parts.
 * Hence we just look for such possible 3-separating set, which must itself be branched.
 * If we find one, we make its union as a new block in the matroid.
 * If we succeed all the way to just two blocks at the end, we have a bw-3 decomposition.
 * Otherwise, the matroid has no such decomposition.
 * (See that one cannot use triangles or triads here, only 3-sep sets of >=4 elements!)
 * 
 * The return value is 1 for branch-width 3, and 0 for no one.
 * 
**/
//******** make the algorithm nicer and faster........
//******** handle not 3-connected matroids as well (decompose first)...............
//******** how to give the decomposition back ???


int	struct_hasbwidth3_ext(int ch, ematrix *ee) {
	ematrix		*e,*rr, **plist, **pol[10], **x;
	int		i,j,k, ret;
	
	if (!ee)  {PROGERROR("must give ee here"); return 1;}
	DEBUG(CURDLEV,"Testing branch-width 3 for the %dx%d matrix %p...\n",ROWSM(ee),COLSM(ee),ee);
	if (ROWSM(ee)+COLSM(ee)<=6) {
		DEBUG(CURDLEV,"- all matroids on %d<=6 elements have branch-width <=3.\n",ROWSM(ee)+COLSM(ee));
		return 1;
	}
			/**
			 * Here we prepare a list of refering matrices, one for each line of e.
			 * (This is a partition of the matroid elements into size-1 sets,
			 * which are then joined to bigger sets across suitable 3-separations.)
			 * We work with a copy of the given matrix.
			**/
	e = ematrix_copy(ee);
	plist = NULL;
	for (i=0; i<ROWSM(e)+COLSM(e); i++) {
		rr = ematrix_refer_empty(e);
		if (i<ROWSM(e))  ematrix_refadd_row(rr,i);
		else  ematrix_refadd_col(rr,i-ROWSM(e));
		plist = alist_append(plist,rr);
	}
	for (i=0; i<(int)(sizeof(pol)/sizeof(pol[0])); i++)  pol[i] = NULL;
	
			/**
			 * As long as there are more than 2 sets in the partition plist,
			 * we cycle this part of the program.
			 * The function struct_hasbwidth3_cycled() looks for a suitable
			 * 3-separation obtained as a union of k (k=2..6) sets of plist.
			 * (It is important to start looking from smaller k to bigger!)
			 * The union is returned as a refering matrix rr, or NULL when finished.
			**/
	while (alist_getlength(plist)>2) {
#ifndef FASTPROG
		DEBUG(CURDLEV+1,"- testing list of length %d: ",alist_getlength(plist));
		for (x=plist; *x; x++)  SDEBUG(CURDLEV+1,",%d",ROWSM(*x)+COLSM(*x));
		SDEBUG(CURDLEV+1,"\n");
#endif
		for (k=2, rr=NULL; k<=6 && !rr; k++)
			rr = struct_hasbwidth3_cycled(e,plist,pol,k);
		if (!rr)  break;
		--k;
			/**
			 * Here we replace the parts of the suitable 3-separation found above
			 * by their union rr (the parts are referred by pol[0..k-1] in plist).
			 * This is correct since a "suitable" 3-separation rr may always
			 * be displayed in some bw-3 decomposition respecting the partition.
			 * (So we never have to backtrack here...)
			**/
		DEBUG(CURDLEV+1,"suitable 3-separation found - replacing %d parts by one of size %d..\n",k,ROWSM(rr)+COLSM(rr));
		if (ch>0)  EMATDEBUG(CURDLEV+2,rr,"\t\t");
		for (i=k-1; i>=0; i--) {
			j = pol[i]-plist;
			if (j<0 || j>=ROWSM(e)+COLSM(e))  {PROGERROR("wrong pol[%d]=%p -%p value!",i,pol[i],plist);}
			dispose_ematrix(*pol[i]);
			alist_delete(pol[i]);	/* (must delete from the last in the list!!!) */
		}
		plist = alist_append(plist,rr);
	}
	ret = (alist_getlength(plist)<=2? 1:0);
#ifndef FASTPROG
	if (!inrecur)  DEBUG(CURDLEV,"- testing branch-width 3 in %p found %s.\n",ee,ret?"YES":"NO");
	if (IFRANDDEBUG(222) && !inrecur) {
		rr = ematrix_copydual(ee);
		for (k=0; k<4; k++) {	/* extra debug testing with pivoted dual matrix */
			i = RANDOM()%ROWSM(rr);  j = RANDOM()%COLSM(rr);
			if (SIGNM(rr,i,j)!=0)  ematrix_pivot(rr,i,j);
		}
		DEBUG(CURDLEV+1,"- recursive debug testing...\n");
		inrecur = 1;  printlev--;
		k = struct_hasbwidth3_ext(0,rr);
		printlev++;  inrecur = 0;
		if (k!=ret)  {PROGERROR("wrong result when recursive calling r%d o%d !",k,ret);
				EMATDEBUG(0,rr," !r!\t"); EMATDEBUG(0,e," !o!\t");}
	}
#endif
	x = NULL;	/* finally, we dispose the temporary matrices used here: */
	dispose_alist_mats(plist);
	dispose_ematrix(e);
	return ret;
}


/**
 * This part is used in struct_hasbwidth3_..() above to find a "suitable" 3-separation
 * in the partition plist of the matrix e.
 * It is known that any 3-separation of >=4 elements in a 3-connected partitioned matroid 
 * can be displayed in some bw 3-decomposition respecting the partitioning.
 * So we are going to find such a separation (by trying all k-tuples, k=2...6),
 * test that the selected (small) side is branched,
 * and return it back via a refering submatrix; or NULL if nothing found.
 * (Notice that we do not have to look at the other side of the selected separation
 * since if it is of size only 3 - the problematic case, we are already finished with bw-3.)
**/

ematrix*	struct_hasbwidth3_cycled(ematrix *e, ematrix **plist, ematrix **pol[], int k) {
	ematrix		*pun[10];
	int		i,j;
	
	if (k>9 || k<=1)  {PROGERROR("wrong value k=%d",k);}
	for (i=0; i<10; i++)  pun[i] = NULL;
	pol[0] = plist-1;
	i = 0;
	while (i>=0) {
		pol[i]++;
		if (!*pol[i]) { --i;  continue; }
		DEBUG(CURDLEV+2,"%*s cycling at %d; list of %d (%d, %d, %d, %d, %d, %d)\n",2*i,"",i+1,
			alist_getlength(plist),(int)(pol[0]-plist),(int)(pol[1]-plist),(int)(pol[2]-plist),(int)(pol[3]-plist),(int)(pol[4]-plist),(int)(pol[5]-plist));
		
			/**
			 * The given array pol[] keeps references to the parts of plist at each level.
			 * The array pun[] then stores the union of the parts from pol[].
			 * We cycle all choice of pol[i], 0<=i<=k-1 by a backtracking.
			 * Then we try whether the union pun[k-1] is 3-separating, and
			 * whether it is a suitable 3-separation by struct_hasbwidth3_separe().
			 * (That includes a test whether pun[k-1] itself is branched.)
			 * If so, we finish the computation here, and return pun[k-1] (and pol[]).
			**/
		if (i==0) {
			pun[i] = *pol[i];
		} else {
			if (pun[i])  dispose_ematrix(pun[i]);
			pun[i] = ematrix_union(pun[i-1],*pol[i]);
		}
		if (i<k-1) {
			++i;  pol[i] = pol[i-1];
		} else {
			j = ematrix_whatsep(e,pun[k-1]);
			if (j>=3)  continue;
			j = struct_hasbwidth3_separe(e,pun[k-1],pol,k);
			if (j)  break;
		}
	}
	//DEBUG(CURDLEV+1,"ending with k=%d  i=%d  pun[k-1]=%p\n",k,i,pun[k-1]);
	if (i>=k)  {PROGERROR("cannot get here! i=%d>=%d",i,k);}
			/* finally, we dispose the temporary matrices used here (not pun[0]!!!): */
	for (j=1; j<k-1 || (j<k&&i<0); j++)
		if (pun[j])  dispose_ematrix(pun[j]);
	return (i<0?NULL: pun[k-1]);
}


/**
 * Here we test whether the given union rr of parts pol[i], 0<=i<=k-1 is "suitable"
 * to be displayed in some bw 3-decomposition of the (partitioned) matroid.
 * (It is known that any 3-separation of >=4 elements in a 3-connected partitioned matroid 
 * can be displayed in some bw 3-decomposition respecting the partitioning.)
 * 
 * This function assumes that rr is obtained as the 3-separating union of the smallest
 * possible number of parts of the partition (2...6 parts guaranteed).
 * The k parts of the union are given in pol[0..k-1].
 * The return value is 1 for suitable, 0 otherwise.
 * 
 * struct_hasbwidth3_lsep3() is supplementary for a quick separation value of a union.
**/

int	struct_hasbwidth3_lsep3(ematrix *e, ematrix *r1, ematrix *r2, ematrix *r3) {
	int	i;
	
	r1 = ematrix_union(r1,r2);  r2 = ematrix_union(r1,r3);
	i = ematrix_whatsep(e,r2);
	dispose_ematrix(r1);  dispose_ematrix(r2);
	return i;
}

int	struct_hasbwidth3_separe(ematrix *e, ematrix *rr, ematrix **pol[], int k) {
	int		i,j,n, a[5],*b, sz;
	static int	comb[][6] = { {1,2,3,4,5,6},{1,2,4,3,5,6},{1,2,5,3,4,6},{1,3,4,2,5,6},{1,3,5,2,4,6},
				      {1,4,5,2,3,6},{2,3,4,1,5,6},{2,3,5,1,4,6},{2,4,5,1,3,6},{3,4,5,1,2,6} };
	
	if (k>6 || k<=1)  {PROGERROR("wrong value k=%d",k);}
	if (ematrix_whatsep(e,rr)>=3)  {PROGERROR("wrong separation given");}
	sz = ROWSM(rr)+COLSM(rr);
	if (sz<k || sz<2)  {PROGERROR("wrong value sz=%d < %d=k",sz,k);}
	for (i=n=0; i<k; i++)  if (ROWSM(*pol[i])+COLSM(*pol[i])>1)  n++;
			/* values that are not accepted since a smaller one must exist: */
	if (n>2 || (n==2 && k>2) || (n>=1 && k>4) || (sz<4))  return 0;
	
			/**
			 * If there are at most 3 parts with one big, then the separation
			 * is branched, similarly if there are exactly 4 single elements.
			 * If there are four parts with one big and three singles, then the
			 * singles must form a 3-separation to be branched.
			**/
	if (k<=3 && sz>=4)  return 1;
	if (k==4 && sz==4)  return 1;
	if (k==4 && sz>4) {
		for (i=j=0; i<4; i++)
			if (ROWSM(*pol[i])+COLSM(*pol[i])==1)  a[j++] = i;
		if (j!=3)  {PROGERROR("something wrong happened to j=%d",j);}
		j = struct_hasbwidth3_lsep3(e,*pol[a[0]],*pol[a[1]],*pol[a[2]]);
		return (j<3);
	}
	if (n>0 || k<5)  {PROGERROR("cannot get here! %d %d",n,k);}
	
			/**
			 * We get here if 5 or 6 single elements are given in the separation.
			 * In such case, there must be one (5) or two (6) 3-separating
			 * triples in order to get branched union.
			 * This is tested here over all 10 3-element subsets...
			**/
	for (i=0; i<10; i++) {
		b = comb[i];
		j = struct_hasbwidth3_lsep3(e,*pol[b[0]-1],*pol[b[1]-1],*pol[b[2]-1]);
		if (j>=3)  continue;
		if (k<=5)  return 1;
		j = struct_hasbwidth3_lsep3(e,*pol[b[3]-1],*pol[b[4]-1],*pol[b[5]-1]);
		if (j<3)  return 1;
	}
	return 0;	/* if no suitable case was found above */
}





















/******************		*******************/
/**************************************************************************/


#undef CURDLEV
#define CURDLEV		8



































